package eu.jonahbauer.json;

import eu.jonahbauer.json.exceptions.JsonConversionException;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

/**
 * Java representation of a JSON array.
 * @param elements the array's elements
 */
@SuppressWarnings("unused")
public record JsonArray(
        @NotNull List<? extends @Nullable JsonValue> elements
) implements JsonValue, List<@Nullable JsonValue> {
    public static final @NotNull JsonArray EMPTY = new JsonArray(List.of());

    public JsonArray {
        Objects.requireNonNull(elements, "elements");
        elements = Util.defensiveCopy(elements);
    }

    public static @NotNull JsonArray of(@Nullable JsonValue @NotNull... elements) {
        return new JsonArray(Arrays.stream(elements).toList());
    }

    //<editor-fold desc="valueOf(...)" defaultstate="collapsed">
    /**
     * Creates a JSON array from the given list, using {@link JsonValue#valueOf(Object)} to convert the objects to
     * {@link JsonValue}s.
     * @param list a list of objects
     * @return a new JSON array
     * @throws JsonConversionException if the given list or its entries cannot be converted to json
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(@Nullable List<?> list) {
        if (list == null) return null;
        if (list instanceof JsonArray json) return json;
        return new JsonArray(list.stream().map(JsonValue::valueOf).toList());
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonValue#valueOf(Object)} to convert the objects to
     * {@link JsonValue}s.
     * @param array an array of objects
     * @return a new JSON array
     * @throws JsonConversionException if the given array or its entries cannot be converted to json
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(@Nullable Object @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).map(JsonValue::valueOf).toList());
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonBoolean#valueOf(boolean)} to convert the
     * booleans to {@link JsonBoolean}s.
     * @param array an array of booleans
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(boolean @Nullable... array) {
        if (array == null) return null;
        var list = new ArrayList<JsonValue>();
        for (var b : array) {
            list.add(JsonBoolean.valueOf(b));
        }
        return new JsonArray(Util.trusted(Collections.unmodifiableList(list)));
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(int)} to convert the
     * bytes to {@link JsonNumber}s.
     * @param array an array of bytes
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(byte @Nullable... array) {
        if (array == null) return null;
        var list = new ArrayList<JsonValue>();
        for (var b : array) {
            list.add(JsonNumber.valueOf(b));
        }
        return new JsonArray(Util.trusted(Collections.unmodifiableList(list)));
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(int)} to convert the
     * shorts to {@link JsonNumber}s.
     * @param array an array of shorts
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(short @Nullable... array) {
        if (array == null) return null;
        var list = new ArrayList<JsonValue>();
        for (var s : array) {
            list.add(JsonNumber.valueOf(s));
        }
        return new JsonArray(Util.trusted(Collections.unmodifiableList(list)));
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonString#valueOf(char)} to convert the
     * chars to {@link JsonString}s.
     * @param array an array of chars
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(char @Nullable... array) {
        if (array == null) return null;
        var list = new ArrayList<JsonValue>();
        for (var c : array) {
            list.add(JsonString.valueOf(c));
        }
        return new JsonArray(Util.trusted(Collections.unmodifiableList(list)));
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(int)} to convert the
     * ints to {@link JsonNumber}s.
     * @param array an array of ints
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(int @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).mapToObj(JsonNumber::valueOf).toList());
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(long)} to convert the
     * longs to {@link JsonNumber}s.
     * @param array an array of longs
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(long @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).mapToObj(JsonNumber::valueOf).toList());
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(double)} to convert the
     * floats to {@link JsonNumber}s.
     * @param array an array of floats
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(float @Nullable... array) {
        if (array == null) return null;
        var list = new ArrayList<JsonValue>();
        for (var f : array) {
            list.add(JsonNumber.valueOf(f));
        }
        return new JsonArray(Util.trusted(Collections.unmodifiableList(list)));
    }

    /**
     * Creates a new JSON array from the given array, using {@link JsonNumber#valueOf(double)} to convert the
     * doubles to {@link JsonNumber}s.
     * @param array an array of doubles
     * @return a new JSON array
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonArray valueOf(double @Nullable... array) {
        return array == null ? null : new JsonArray(Arrays.stream(array).mapToObj(JsonNumber::valueOf).toList());
    }
    //</editor-fold>

    //<editor-fold desc="List" defaultstate="collapsed">
    @Override
    public int size() {
        return elements.size();
    }

    @Override
    public boolean isEmpty() {
        return elements.isEmpty();
    }

    @Override
    public boolean contains(Object o) {
        return elements.contains(o);
    }

    @Override
    @SuppressWarnings("unchecked") // elements is immutable
    public @NotNull Iterator<@Nullable JsonValue> iterator() {
        return (Iterator<JsonValue>) elements.iterator();
    }

    @Override
    public Object @NotNull[] toArray() {
        return elements.toArray();
    }

    @Override
    public <T> T @NotNull[] toArray(T @NotNull[] a) {
        return elements.toArray(a);
    }

    @Override
    public boolean add(@Nullable JsonValue value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean remove(Object o) {
        throw new UnsupportedOperationException();
    }

    @Override
    @SuppressWarnings("SlowListContainsAll")
    public boolean containsAll(@NotNull Collection<?> c) {
        return elements.containsAll(c);
    }

    @Override
    public boolean addAll(@NotNull Collection<? extends @Nullable JsonValue> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean addAll(int index, @NotNull Collection<? extends @Nullable JsonValue> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean removeAll(@NotNull Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public boolean retainAll(@NotNull Collection<?> c) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue get(int index) {
        return elements.get(index);
    }

    @Override
    public @Nullable JsonValue set(int index, @Nullable JsonValue element) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void add(int index, @Nullable JsonValue element) {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue remove(int index) {
        throw new UnsupportedOperationException();
    }

    @Override
    public int indexOf(Object o) {
        return elements.indexOf(o);
    }

    @Override
    public int lastIndexOf(Object o) {
        return elements.lastIndexOf(o);
    }

    @Override
    @SuppressWarnings("unchecked") // elements is immutable
    public @NotNull ListIterator<@Nullable JsonValue> listIterator() {
        return (ListIterator<JsonValue>) elements.listIterator();
    }

    @Override
    @SuppressWarnings("unchecked") // elements is immutable
    public @NotNull ListIterator<@Nullable JsonValue> listIterator(int index) {
        return (ListIterator<JsonValue>) elements.listIterator(index);
    }

    @Override
    public @NotNull JsonArray subList(int fromIndex, int toIndex) {
        return new JsonArray(elements.subList(fromIndex, toIndex));
    }

    @Override
    public @NotNull JsonArray reversed() {
        return new JsonArray(Util.trusted(elements.reversed()));
    }

    //</editor-fold>

    //<editor-fold desc="get...(int index)" defaultstate="collapsed">
    /**
     * {@return the element at the specified index in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified index is not an instance of {@link JsonString}
     * @throws NullPointerException if the element at the specified index is {@code null}
     * @throws IndexOutOfBoundsException if the index is less than zero or greater than or equals to the size of the array
     */
    public @NotNull JsonString getString(int index) {
        return get(index, JsonString.class);
    }

    /**
     * {@return the element at the specified index in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified index is not an instance of {@link JsonNumber}
     * @throws NullPointerException if the element at the specified index is {@code null}
     * @throws IndexOutOfBoundsException if the index is less than zero or greater than or equals to the size of the array
     */
    public @NotNull JsonNumber getNumber(int index) {
        return get(index, JsonNumber.class);
    }

    /**
     * {@return the element at the specified index in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified index is not an instance of {@link JsonBoolean}
     * @throws NullPointerException if the element at the specified index is {@code null}
     * @throws IndexOutOfBoundsException if the index is less than zero or greater than or equals to the size of the array
     */
    public @NotNull JsonBoolean getBoolean(int index) {
        return get(index, JsonBoolean.class);
    }

    /**
     * {@return the element at the specified index in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified index is not an instance of {@link JsonArray}
     * @throws NullPointerException if the element at the specified index is {@code null}
     * @throws IndexOutOfBoundsException if the index is less than zero or greater than or equals to the size of the array
     */
    public @NotNull JsonArray getArray(int index) {
        return get(index, JsonArray.class);
    }

    /**
     * {@return the element at the specified index in this array}
     * @param index index of the element to return
     * @throws ClassCastException if the element at the specified index is not an instance of {@link JsonObject}
     * @throws NullPointerException if the element at the specified index is {@code null}
     * @throws IndexOutOfBoundsException if the index is less than zero or greater than or equals to the size of the array
     */
    public @NotNull JsonObject getObject(int index) {
        return get(index, JsonObject.class);
    }

    private <T extends JsonValue> @NotNull T get(int index, @NotNull Class<T> type) {
        var value = elements().get(index);
        if (value == null) {
            throw new NullPointerException("Value at index" + index + " is null.");
        } else {
            return type.cast(value);
        }
    }
    //</editor-fold>

    @Override
    public @NotNull String toString() {
        var joiner = new StringJoiner(",", "[", "]");
        for (JsonValue element : elements) {
            joiner.add(JsonValue.toJsonString(element));
        }
        return joiner.toString();
    }

    @Override
    public @NotNull String toPrettyJsonString(int depth) {
        if (size() < 2) return toJsonString();

        var entries = new ArrayList<String>(elements.size());
        var length = 0L;
        var composite = false;
        for (var element : elements) {
            var string = JsonValue.toPrettyJsonString(element, depth + Util.PRETTY_PRINT_INDENT.length());
            entries.add(string);

            composite |= element instanceof JsonObject || element instanceof JsonArray;
            length += string.length() + 2; // +2 to account for ", " and "[]"
        }

        if (!composite && depth + length <= Util.PRETTY_PRINT_ARRAY_MULTILINE_THRESHOLD) {
            return "[" + String.join(", ", entries) + "]";
        } else {
            var out = new StringJoiner(",\n", "[\n", "\n]");
            entries.forEach(e -> out.add(indent(e)));
            return out.toString();
        }
    }

    private static @NotNull String indent(@NotNull String string) {
        if (string.isEmpty()) return "";

        StringBuilder out = new StringBuilder();

        Iterator<String> it = string.lines().iterator();
        while (it.hasNext()) {
            out.append(Util.PRETTY_PRINT_INDENT).append(it.next());
            if (it.hasNext()) out.append('\n');
        }

        return out.toString();
    }
}
