package eu.jonahbauer.json;

import eu.jonahbauer.json.tokenizer.token.JsonToken;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/**
 * Java representation of a JSON boolean. Note, that JSON does not distinguish between integers and floating point
 * numbers and therefore all numbers are stored as {@code double}.
 */
public record JsonNumber(double value) implements JsonValue, JsonToken {
    public static final @NotNull JsonNumber ZERO = new JsonNumber(0);
    public static final @NotNull JsonNumber ONE = new JsonNumber(1);

    public JsonNumber {
        if (!Double.isFinite(value)) throw new IllegalArgumentException("value must be finite");
    }

    /**
     * Converts the given int to a JSON number.
     * @param i an int
     * @return a new JSON number
     */
    public static @NotNull JsonNumber valueOf(int i) {
        return new JsonNumber(i);
    }

    /**
     * Converts the given long to a JSON number.
     * @param l a long
     * @return a new JSON number
     * @throws IllegalArgumentException if conversion from {@code long} to {@code double} is lossy
     */
    public static @NotNull JsonNumber valueOf(long l) {
        if ((long) (double) l != l) {
            throw new IllegalArgumentException("lossy conversion from long to double for value " + l);
        }
        return new JsonNumber(l);
    }

    /**
     * Converts the given double to a JSON number.
     * @param d a double
     * @return a new JSON number
     * @throws IllegalArgumentException if {@code d} is not {@linkplain Double#isFinite(double) finite}
     */
    public static @NotNull JsonNumber valueOf(double d) {
        return new JsonNumber(d);
    }

    /**
     * Converts the given number to a JSON number using {@link Number#doubleValue()}. The conversion to a JSON number
     * might be lossy.
     * @param number a number
     * @return a new JSON number
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonNumber valueOf(@Nullable Number number) {
        return number == null ? null : new JsonNumber(number.doubleValue());
    }

    @Override
    public @NotNull String toString() {
        return (long) value == value ? Long.toString((long) value) : Double.toString(value);
    }
}
