package eu.jonahbauer.json;

import eu.jonahbauer.json.exceptions.JsonConversionException;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;

/**
 * Java representation of a JSON object.
 * @param entries the object's entries
 */
@SuppressWarnings("unused")
public record JsonObject(@NotNull Map<@NotNull String, @Nullable JsonValue> entries) implements JsonValue, Map<@NotNull String, @Nullable JsonValue> {
    public static final @NotNull JsonObject EMPTY = new JsonObject(Map.of());

    public JsonObject {
        Objects.requireNonNull(entries, "entries");
        entries = Util.defensiveCopy(entries);
    }

    /**
     * Returns a JSON object containing one mapping.
     * @param k1 the key
     * @param v1 the value
     * @return a new JSON object containing the specified mapping
     */
    public static @NotNull JsonObject of(
            @NotNull String k1, @Nullable JsonValue v1
    ) {
        var map = new LinkedHashMap<String, JsonValue>();
        map.put(Objects.requireNonNull(k1), v1);
        return new JsonObject(Util.trusted(Collections.unmodifiableSequencedMap(map)));
    }

    /**
     * Returns a JSON object containing two mapping.
     * @param k1 the first key
     * @param v1 the first value
     * @param k2 the second key
     * @param v2 the second value
     * @return a new JSON object containing the specified mappings
     */
    public static @NotNull JsonObject of(
            @NotNull String k1, @Nullable JsonValue v1,
            @NotNull String k2, @Nullable JsonValue v2
    ) {
        var map = new LinkedHashMap<String, JsonValue>();
        map.put(Objects.requireNonNull(k1), v1);
        map.put(Objects.requireNonNull(k2), v2);
        return new JsonObject(Util.trusted(Collections.unmodifiableSequencedMap(map)));
    }

    /**
     * Returns a JSON object containing three mapping.
     * @param k1 the first key
     * @param v1 the first value
     * @param k2 the second key
     * @param v2 the second value
     * @param k3 the third key
     * @param v3 the third value
     * @return a new JSON object containing the specified mappings
     */
    public static @NotNull JsonObject of(
            @NotNull String k1, @Nullable JsonValue v1,
            @NotNull String k2, @Nullable JsonValue v2,
            @NotNull String k3, @Nullable JsonValue v3
    ) {
        var map = new LinkedHashMap<String, JsonValue>();
        map.put(Objects.requireNonNull(k1), v1);
        map.put(Objects.requireNonNull(k2), v2);
        map.put(Objects.requireNonNull(k3), v3);
        return new JsonObject(Util.trusted(Collections.unmodifiableSequencedMap(map)));
    }

    /**
     * Returns a JSON object containing four mapping.
     * @param k1 the first key
     * @param v1 the first value
     * @param k2 the second key
     * @param v2 the second value
     * @param k3 the third key
     * @param v3 the third value
     * @param k4 the fourth key
     * @param v4 the fourth value
     * @return a new JSON object containing the specified mappings
     */
    public static @NotNull JsonObject of(
            @NotNull String k1, @Nullable JsonValue v1,
            @NotNull String k2, @Nullable JsonValue v2,
            @NotNull String k3, @Nullable JsonValue v3,
            @NotNull String k4, @Nullable JsonValue v4
    ) {
        var map = new LinkedHashMap<String, JsonValue>();
        map.put(Objects.requireNonNull(k1), v1);
        map.put(Objects.requireNonNull(k2), v2);
        map.put(Objects.requireNonNull(k3), v3);
        map.put(Objects.requireNonNull(k4), v4);
        return new JsonObject(Util.trusted(Collections.unmodifiableSequencedMap(map)));
    }

    /**
     * Creates a JSON object from the given map, using {@link JsonValue#valueOf(Object)} to convert the values to
     * {@link JsonValue}s. The keys are expected to be instances of {@link CharSequence}.
     * @param map a map
     * @return a new JSON object
     * @throws JsonConversionException if the given map or its entries cannot be converted to json
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonObject valueOf(@Nullable Map<?, ?> map) {
        if (map == null) return null;
        if (map instanceof JsonObject json) return json;
        
        var out = new LinkedHashMap<String, JsonValue>();
        map.forEach((key, value) -> out.put(keyOf(key), JsonValue.valueOf(value)));
        return new JsonObject(out);
    }

    private static @NotNull String keyOf(@Nullable Object object) {
        return switch (object) {
            case CharSequence chars -> chars.toString();
            case null -> throw new JsonConversionException("cannot use null as json key");
            default -> throw new JsonConversionException("cannot convert object of type " + object.getClass() + " to json key");
        };
    }

    //<editor-fold desc="Map" defaultstate="collapsed">
    @Override
    public int size() {
        return entries.size();
    }

    @Override
    public boolean isEmpty() {
        return entries.isEmpty();
    }

    @Override
    public boolean containsKey(Object key) {
        Objects.requireNonNull(key, "key");
        return entries.containsKey(key);
    }

    @Override
    public boolean containsValue(Object value) {
        return entries.containsValue(value);
    }

    @Override
    public @Nullable JsonValue get(Object key) {
        Objects.requireNonNull(key, "key");
        return entries.get(key);
    }

    @Override
    public @Nullable JsonValue put(@NotNull String key, @Nullable JsonValue value) {
        throw new UnsupportedOperationException();
    }

    @Override
    public @Nullable JsonValue remove(Object key) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void putAll(@NotNull Map<? extends @NotNull String, ? extends @Nullable JsonValue> m) {
        throw new UnsupportedOperationException();
    }

    @Override
    public void clear() {
        throw new UnsupportedOperationException();
    }

    @Override
    public @NotNull Set<@NotNull String> keySet() {
        return entries.keySet();
    }

    @Override
    public @NotNull Collection<@Nullable JsonValue> values() {
        return entries.values();
    }

    @Override
    public @NotNull Set<Entry<@NotNull String, @Nullable JsonValue>> entrySet() {
        return entries.entrySet();
    }
    //</editor-fold>

    //<editor-fold desc="get...(String key)" defaultstate="collapsed">
    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is not an instance of {@link JsonString}
     * @throws NullPointerException if there is no mapping for the key or the key is mapped to {@code null}
     */
    public @NotNull JsonString getString(@NotNull String key) {
        return get(key, JsonString.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is not an instance of {@link JsonNumber}
     * @throws NullPointerException if there is no mapping for the key or the key is mapped to {@code null}
     */
    public @NotNull JsonNumber getNumber(@NotNull String key) {
        return get(key, JsonNumber.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is not an instance of {@link JsonBoolean}
     * @throws NullPointerException if there is no mapping for the key or the key is mapped to {@code null}
     */
    public @NotNull JsonBoolean getBoolean(@NotNull String key) {
        return get(key, JsonBoolean.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is not an instance of {@link JsonArray}
     * @throws NullPointerException if there is no mapping for the key or the key is mapped to {@code null}
     */
    public @NotNull JsonArray getArray(@NotNull String key) {
        return get(key, JsonArray.class);
    }

    /**
     * {@return the value to which the specified key is mapped, or <code>null</code> if this object contains no mapping for the key}
     * @throws ClassCastException if the value is not an instance of {@link JsonObject}
     * @throws NullPointerException if there is no mapping for the key or the key is mapped to {@code null}
     */
    public @NotNull JsonObject getObject(@NotNull String key) {
        return get(key, JsonObject.class);
    }

    private <T extends JsonValue> @NotNull T get(@NotNull String key, @NotNull Class<T> type) {
        var value = entries().get(key);
        if (value == null) {
            if (entries().containsKey(key)) {
                throw new NullPointerException("Key " + key + " is mapped to null.");
            } else {
                throw new NullPointerException("No mapping for key " + key);
            }
        } else {
            return type.cast(value);
        }
    }
    //</editor-fold>

    //<editor-fold desc="get...OrDefault(String key, ...)" defaultstate="collapsed">
    // getStringOrDefault(String, JsonString) omitted because JsonString implements CharSequence

    /**
     * Returns the value to which the specified key is mapped, or 
     * {@link JsonString#valueOf(CharSequence) JsonString.valueOf(defaultValue)} if this object contains no mapping for
     * the key or the key is mapped to a value that is not an instance of {@link JsonString}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonString getStringOrDefault(@NotNull String key, @Nullable CharSequence defaultValue) {
        return getOrDefault(key, defaultValue, JsonString.class, JsonString::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     * for the key or the key is mapped to a value that is not an instance of {@link JsonNumber}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonNumber getNumberOrDefault(@NotNull String key, @Nullable JsonNumber defaultValue) {
        return getOrDefault(key, defaultValue, JsonNumber.class, Function.identity());
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonNumber#valueOf(Number) JsonNumber.valueOf(defaultValue)} if this object contains no mapping for the
     * key or the key is mapped to a value that is not an instance of {@link JsonNumber}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonNumber getNumberOrDefault(@NotNull String key, @Nullable Number defaultValue) {
        return getOrDefault(key, defaultValue, JsonNumber.class, JsonNumber::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or {@code defaultValue} if this object contains no mapping
     * for the key or the key is mapped to a value that is not an instance of {@link JsonBoolean}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonBoolean getBooleanOrDefault(@NotNull String key, @Nullable JsonBoolean defaultValue) {
        return getOrDefault(key, defaultValue, JsonBoolean.class, Function.identity());
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonBoolean#valueOf(Boolean) JsonBoolean.valueOf(defaultValue)} if this object contains no mapping
     * for the key or the key is mapped to a value that is not an instance of {@link JsonBoolean}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonBoolean getBooleanOrDefault(@NotNull String key, @Nullable Boolean defaultValue) {
        return getOrDefault(key, defaultValue, JsonBoolean.class, JsonBoolean::valueOf);
    }
    
    // getArrayOrDefault(String, JsonArray) omitted because JsonArray implements List<?>

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonArray#valueOf(List) JsonArray.valueOf(defaultValue)} if this object contains no mapping for
     * the key or the key is mapped to a value that is not an instance of {@link JsonArray}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonArray getArrayOrDefault(@NotNull String key, @Nullable List<?> defaultValue) {
        return getOrDefault(key, defaultValue, JsonArray.class, JsonArray::valueOf);
    }

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonArray#valueOf(Object...) JsonArray.valueOf(defaultValue)} if this object contains no mapping for the
     * key or the key is mapped to a value that is not an instance of {@link JsonArray}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonArray getArrayOrDefault(@NotNull String key, @Nullable Object @Nullable... defaultValue) {
        return getOrDefault(key, defaultValue, JsonArray.class, JsonArray::valueOf);
    }

    // getObjectOrDefault(String, JsonObject) omitted because JsonObject implements Map<?, ?>

    /**
     * Returns the value to which the specified key is mapped, or
     * {@link JsonObject#valueOf(Map) JsonObject.valueOf(defaultValue)} if this object contains no mapping for the key
     * or the key is mapped to a value that is not an instance of {@link JsonObject}.
     * @return the value to which the specified key is mapped falling back to the default value
     */
    public @Nullable JsonObject getObjectOrDefault(@NotNull String key, @Nullable Map<?, ?> defaultValue) {
        return getOrDefault(key, defaultValue, JsonObject.class, JsonObject::valueOf);
    }
    
    private <T extends JsonValue, S> @Nullable T getOrDefault(@NotNull String key, @Nullable S defaultValue, @NotNull Class<T> type, @NotNull Function<S, T> converter) {
        var value = entries().get(key);
        if (type.isInstance(value)) {
            return type.cast(value);
        } else {
            return converter.apply(defaultValue);
        }
    }
    //</editor-fold>

    @Override
    public @NotNull String toString() {
        StringJoiner out = new StringJoiner(",", "{", "}");
        entries.forEach((key, value) -> out.add(JsonString.quote(key) + ":" + JsonValue.toJsonString(value)));
        return out.toString();
    }

    @Override
    public @NotNull String toPrettyJsonString(int depth) {
        StringJoiner out = new StringJoiner(",\n" + Util.PRETTY_PRINT_INDENT, "{\n" + Util.PRETTY_PRINT_INDENT, "\n}");
        entries.forEach((key, value) -> {
            var keyString = JsonString.quote(key);
            var valueIdent = depth + Util.PRETTY_PRINT_INDENT.length() + keyString.length() + 3; // +3 for ": " and ","
            var valueString = JsonValue.toPrettyJsonString(value, valueIdent);
            out.add(keyString + ": " + indent(valueString));
        });
        return out.toString();
    }

    private static @NotNull String indent(@NotNull String string) {
        if (string.isEmpty()) return "";

        StringBuilder out = new StringBuilder();

        Iterator<String> it = string.lines().iterator();
        out.append(it.next());

        while (it.hasNext()) {
            String next = it.next();
            out.append('\n').append(Util.PRETTY_PRINT_INDENT).append(next);
        }

        return out.toString();
    }
}
