package eu.jonahbauer.json;

import eu.jonahbauer.json.tokenizer.token.JsonToken;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.stream.IntStream;

@SuppressWarnings("unused")
public record JsonString(@NotNull String value) implements JsonValue, JsonToken, CharSequence {
    public static final @NotNull JsonString EMPTY = new JsonString("");

    public JsonString {
        Objects.requireNonNull(value, "value");
    }

    /**
     * Converts the given character to a JSON string.
     * @param chr a character
     * @return a new JSON string
     */
    public static @NotNull JsonString valueOf(char chr) {
        return new JsonString(String.valueOf(chr));
    }

    /**
     * Converts the given character to a JSON string.
     * @param chr a character
     * @return a new JSON string
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonString valueOf(@Nullable Character chr) {
        return chr == null ? null : new JsonString(String.valueOf(chr));
    }

    /**
     * Converts the given string to a JSON string.
     * @param chars a string
     * @return a new JSON string
     */
    @Contract("null -> null; !null -> !null")
    public static @Nullable JsonString valueOf(@Nullable CharSequence chars) {
        return switch (chars) {
            case JsonString json -> json;
            case null -> null;
            default -> {
                @SuppressWarnings("java:S2259")
                var result = new JsonString(chars.toString());
                yield result;
            }
        };
    }

    /**
     * Quotes the given {@link String} and encodes all invalid characters as an escape sequence.
     * @param value a string value
     * @return a JSON representation of the given {@code value}
     */
    public static @NotNull String quote(@NotNull String value) {
        StringBuilder out = new StringBuilder(value.length() + 2);
        out.append('"');
        for (int i = 0, length = value.length(); i < length; i++) {
            char chr = value.charAt(i);
            switch (chr) {
                case '"' -> out.append("\\\"");
                case '\\' -> out.append("\\\\");
                case '\b' -> out.append("\\b");
                case '\f' -> out.append("\\f");
                case '\n' -> out.append("\\n");
                case '\r' -> out.append("\\r");
                case '\t' -> out.append("\\t");
                default -> {
                    if (chr < 32) {
                        out.append("\\u%04x".formatted((int) chr));
                    } else {
                        out.append(chr);
                    }
                }
            }
        }
        out.append('"');
        return out.toString();
    }

    //<editor-fold desc="CharSequence" defaultstate="collapsed">
    @Override
    public int length() {
        return value.length();
    }

    @Override
    public char charAt(int index) {
        return value.charAt(index);
    }

    @Override
    public @NotNull JsonString subSequence(int start, int end) {
        return new JsonString(value.substring(start, end));
    }

    @Override
    public boolean isEmpty() {
        return value.isEmpty();
    }

    @Override
    public @NotNull IntStream chars() {
        return value.chars();
    }

    @Override
    public @NotNull IntStream codePoints() {
        return value.codePoints();
    }

    @Override
    public @NotNull String toString() {
        return value;
    }
    //</editor-fold>

    @Override
    public @NotNull String toJsonString() {
        return quote(value());
    }
}
