package eu.jonahbauer.json;

import eu.jonahbauer.json.exceptions.JsonConversionException;
import eu.jonahbauer.json.exceptions.JsonParserException;
import eu.jonahbauer.json.parser.JsonParser;
import org.jetbrains.annotations.Contract;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.Serializable;
import java.util.List;
import java.util.Map;
import java.util.Objects;

public sealed interface JsonValue extends Serializable permits JsonObject, JsonArray, JsonBoolean, JsonNumber, JsonString {

    /**
     * {@return the JSON representation of this value}
     */
    default @NotNull String toJsonString() {
        return toString();
    }

    /**
     * {@return a prettified JSON representation of this value}
     */
    default @NotNull String toPrettyJsonString() {
        return toPrettyJsonString(0);
    }

    /**
     * {@return a prettified JSON representation of this value}
     * @param depth indicates the depth at which this value appears (no formal definition given)
     */
    default @NotNull String toPrettyJsonString(int depth) {
        return toJsonString();
    }

    /**
     * {@return the JSON representation of the given value}
     * @param value a {@link JsonValue} (possibly {@code null})
     */
    static @NotNull String toJsonString(@Nullable JsonValue value) {
        return value == null ? "null" : value.toJsonString();
    }

    /**
     * {@return the prettified JSON representation of the given value}
     * @param value a {@link JsonValue} (possibly {@code null})
     */
    static @NotNull String toPrettyJsonString(@Nullable JsonValue value) {
        return toPrettyJsonString(value, 0);
    }

    /**
     * {@return the prettified JSON representation of the given value}
     * @param value a {@link JsonValue} (possibly {@code null})
     * @param depth indicates the depth at which the value appears (no formal definition given)
     */
    static @NotNull String toPrettyJsonString(@Nullable JsonValue value, int depth) {
        return value == null ? "null" : value.toPrettyJsonString(depth);
    }

    /**
     * Parses a string as a JSON. The string must adhere to the <a href="https://www.json.org">JSON grammar</a>.
     * @param json a string
     * @return the result of parsing the string as a JSON
     * @throws JsonParserException if the string cannot be parsed as JSON
     */
    static @Nullable JsonValue parse(@NotNull String json) {
        var parser = new JsonParser(json);
        return parser.parse();
    }

    /**
     * Parses a string template as a JSON. The string template must adhere to the
     * <a href="https://www.json.org">standard JSON grammar</a> with the following additional rules:
     * <pre>{@code
     * value
     *    placeholder
     *
     * member
     *    ws placeholder ws ':' element
     *
     * characters
     *     placeholder characters
     * }</pre>
     * where <code>placeholder</code> represents an object embedded in the string template.
     * <p>
     * A placeholder will be converted to JSON depending on the context it appears in.
     * When used as a value, {@link JsonValue#valueOf(Object)} will be used.
     * When used as an object key, only instances or {@link CharSequence} are allowed, which will be converted using
     * {@link CharSequence#toString()}.
     * When used in a string, {@link Objects#toString(Object)} will be used.
     *
     * @param json a string template
     * @return the result of parsing the string template as a JSON
     * @throws JsonParserException if the string template cannot be parsed as JSON
     * @throws JsonConversionException if a placeholder cannot be converted to JSON
     */
    @SuppressWarnings("preview")
    static @Nullable JsonValue parse(@NotNull StringTemplate json) {
        var parser = new JsonParser(json);
        return parser.parse();
    }

    /**
     * Converts an object to a JSON value:
     * <ul>
     *     <li>{@link JsonValue}s and {@code null} are returned as is</li>
     *     <li>{@link Boolean}s are converted with {@link JsonBoolean#valueOf(Boolean)}</li>
     *     <li>{@link Number}s are converted with {@link JsonNumber#valueOf(Number)}</li>
     *     <li>{@link Character}s are converted with {@link JsonString#valueOf(Character)}</li>
     *     <li>{@link CharSequence}s are converted with {@link JsonString#valueOf(CharSequence)}</li>
     *     <li>{@link List}s are converted with {@link JsonArray#valueOf(List)}</li>
     *     <li>arrays are converted with the respective overload of {@link JsonArray#valueOf(Object...)}</li>
     *     <li>{@link Map}s are converted with {@link JsonObject#valueOf(Map)}</li>
     * </ul>
     * @param object an object
     * @return a {@link JsonValue} representing the object
     * @throws JsonConversionException if the object cannot be converted to a {@link JsonValue}
     */
    @Contract("null -> null; !null -> !null")
    static @Nullable JsonValue valueOf(@Nullable Object object) {
        return switch (object) {
            case JsonValue json -> json;
            case Boolean bool -> JsonBoolean.valueOf(bool);
            case Number number -> JsonNumber.valueOf(number);
            case Character chr -> JsonString.valueOf(String.valueOf(chr));
            case CharSequence chars -> JsonString.valueOf(chars);
            case List<?> list -> JsonArray.valueOf(list);
            case Object[] array -> JsonArray.valueOf(array);
            case boolean[] array -> JsonArray.valueOf(array);
            case byte[] array -> JsonArray.valueOf(array);
            case short[] array -> JsonArray.valueOf(array);
            case char[] array -> JsonArray.valueOf(array);
            case int[] array -> JsonArray.valueOf(array);
            case long[] array -> JsonArray.valueOf(array);
            case float[] array -> JsonArray.valueOf(array);
            case double[] array -> JsonArray.valueOf(array);
            case Map<?, ?> map -> JsonObject.valueOf(map);
            case null -> null;
            default -> throw new JsonConversionException("cannot convert object of type " + object.getClass() + " to json value");
        };
    }
}
