package eu.jonahbauer.json;

import lombok.experimental.Delegate;
import lombok.experimental.UtilityClass;
import org.jetbrains.annotations.NotNull;

import java.util.*;

@UtilityClass
class Util {
    static final @NotNull String PRETTY_PRINT_INDENT = "    ";
    static final int PRETTY_PRINT_ARRAY_MULTILINE_THRESHOLD = 80;

    private static final @NotNull Class<?> LIST_1_2 = List.of(1).getClass();
    private static final @NotNull Class<?> LIST_N = List.of(1, 2, 3).getClass();
    private static final @NotNull Class<?> SUB_LIST = List.of(1, 2, 3).subList(0, 2).getClass();
    private static final @NotNull Class<?> MAP_1 = Map.of(1, 2).getClass();
    private static final @NotNull Class<?> MAP_N = Map.of(1, 2, 3, 4).getClass();

    /**
     * Performs a defensive copy of the given list unless it is known to be an immutable list.
     */
    public static <T> @NotNull List<T> defensiveCopy(@NotNull List<T> list) {
        if (LIST_1_2.isInstance(list) || LIST_N.isInstance(list) || SUB_LIST.isInstance(list)) {
            return list;
        } else if (list instanceof TrustedList<T>(var delegate)) {
            return delegate;
        } else {
            return list.stream().toList();
        }
    }

    /**
     * Performs a defensive copy of the given map unless it is known to be an immutable map.
     * If the given map implements {@link SequencedMap} as will the returned one.
     */
    public static <K, V> @NotNull Map<K, V> defensiveCopy(@NotNull Map<@NotNull K, V> map) {
        if (MAP_1.isInstance(map) || MAP_N.isInstance(map)) {
            return map;
        } else if (map instanceof TrustedMap<K, V>(var delegate)) {
            return delegate;
        } else if (map instanceof SequencedMap<K,V> sequenced) {
            var out = new LinkedHashMap<K, V>();
            sequenced.forEach((key, value) -> out.put(Objects.requireNonNull(key, "key"), value));
            return Collections.unmodifiableSequencedMap(out);
        } else {
            var out = new HashMap<K, V>();
            map.forEach((key, value) -> out.put(Objects.requireNonNull(key, "key"), value));
            return Collections.unmodifiableMap(out);
        }
    }

    /**
     * Returns a list that will be assumed immutable when given to {@link #defensiveCopy(List)}.
     * {@return a list that will be assumed immutable}
     */
    public static <T> @NotNull List<T> trusted(@NotNull List<T> list) {
        return new TrustedList<>(list);
    }

    /**
     * Returns a map that will be assumed immutable when given to {@link #defensiveCopy(Map)}.
     * {@return a map that will be assumed immutable}
     */
    public static <K, V> @NotNull Map<K, V> trusted(@NotNull Map<@NotNull K, V> map) {
        return new TrustedMap<>(map);
    }

    private record TrustedList<T>(@Delegate @NotNull List<T> delegate) implements List<T> {}

    private record TrustedMap<K, V>(@Delegate @NotNull Map<K, V> delegate) implements Map<K, V> {}
}
