package eu.jonahbauer.json.parser;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.exceptions.JsonConversionException;
import eu.jonahbauer.json.exceptions.JsonParserException;
import eu.jonahbauer.json.exceptions.JsonTokenizerException;
import eu.jonahbauer.json.tokenizer.JsonTokenizer;
import eu.jonahbauer.json.tokenizer.JsonTokenizerImpl;
import eu.jonahbauer.json.tokenizer.token.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.io.Reader;
import java.io.UncheckedIOException;
import java.util.*;

public final class JsonParser {
    private static final String TT_JSON_STRING = "JSON_STRING";
    private static final String TT_JSON_NUMBER = "JSON_NUMBER";
    private static final String TT_JSON_STRING_TEMPLATE = "JSON_STRING_TEMPLATE";
    private static final String TT_JSON_PLACEHOLDER = "JSON_PLACEHOLDER";

    private final @NotNull JsonTokenizer tokenizer;

    @SuppressWarnings("preview")
    public JsonParser(@NotNull StringTemplate json) {
        this(new JsonTokenizerImpl(json));
    }

    public JsonParser(@NotNull String json) {
        this(new JsonTokenizerImpl(json));
    }

    public JsonParser(@NotNull Reader reader) {
        this(new JsonTokenizerImpl(reader));
    }

    public JsonParser(@NotNull JsonTokenizer tokenizer) {
        this.tokenizer = Objects.requireNonNull(tokenizer);
    }

    /**
     * Parses the input as a JSON value.
     * @return the JSON value
     * @throws JsonParserException if the input could not be parsed as JSON
     * @throws UncheckedIOException if an I/O error occurs
     */
    public @Nullable JsonValue parse() throws JsonParserException {
        try {
            var out = parse0();
            if (tokenizer.next() != null) {
                throw new JsonParserException(tokenizer.getFragment(), tokenizer.getLineNumber(), tokenizer.getColumnNumber(), "unexpected trailing entries");
            }
            return out;
        } catch (IOException ex) {
            throw new UncheckedIOException(ex);
        } catch (JsonTokenizerException ex) {
            throw new JsonParserException(ex.getFragment(), ex.getLineNumber(), ex.getColumnNumber(), ex.getMessage(), ex);
        }
    }

    private @Nullable JsonValue parse0() throws IOException {
        // Use a stack of Context objects to store the objects and arrays which are currently being parsed.
        // This avoids the risk of stack overflow which would exist when having one nested calls to parse0.
        var stack = new ArrayDeque<@NotNull Context>();

        while (true) {
            JsonValue value;

            var token = tokenizer.next();
            if (!stack.isEmpty() && token == stack.peek().end()) {
                value = stack.pop().build();
            } else {
                // consume value separator
                if (stack.peek() instanceof Context context && !context.isEmpty()) {
                    if (token != JsonPunctuation.VALUE_SEPARATOR) {
                        throw unexpectedToken(token, context.end(), JsonPunctuation.VALUE_SEPARATOR);
                    }
                    token = tokenizer.next();
                }

                // read object key and consume name separator
                if (stack.peek() instanceof ObjectContext object) {
                    var key = parseJsonObjectKey(object, token);
                    if ((token = tokenizer.next()) != JsonPunctuation.NAME_SEPARATOR) {
                        throw unexpectedToken(token, JsonPunctuation.NAME_SEPARATOR);
                    }
                    token = tokenizer.next();
                    object.setKey(key);
                }

                if (token == JsonPunctuation.BEGIN_OBJECT) {
                    stack.push(new ObjectContext());
                    continue;
                } else if (token == JsonPunctuation.BEGIN_ARRAY) {
                    stack.push(new ArrayContext());
                    continue;
                } else {
                    // read json value
                    value = switch (token) {
                        case JsonPunctuation.BEGIN_ARRAY, JsonPunctuation.BEGIN_OBJECT -> throw new AssertionError();
                        case JsonNull.NULL -> null;
                        case JsonValue v -> v;
                        case JsonStringTemplate template -> template.asString();
                        case JsonPlaceholder(var object) -> JsonValue.valueOf(object);
                        case JsonPunctuation _ -> throw unexpectedStartOfValue(token);
                        case null -> throw unexpectedEndOfFile();
                    };
                }
            }

            if (stack.peek() instanceof Context context) {
                context.add(value);
            } else {
                return value;
            }
        }
    }

    private @NotNull String parseJsonObjectKey(@NotNull ObjectContext context, @Nullable JsonToken token) {
        return switch (token) {
            case JsonString string -> string.value();
            case JsonStringTemplate template -> template.asString().value();
            case JsonPlaceholder(var object) -> switch (object) {
                case CharSequence chars -> chars.toString();
                case null -> throw new JsonConversionException("cannot use null as json object key");
                default -> throw new JsonConversionException("cannot convert object of type " + object.getClass() + " to json object key");
            };
            case null -> throw unexpectedEndOfFile();
            default -> throw unexpectedStartOfKey(context, token);
        };
    }

    private @NotNull JsonParserException unexpectedStartOfKey(@NotNull ObjectContext context, @Nullable JsonToken token) {
        var expected = new ArrayList<>(4);
        expected.add(TT_JSON_STRING);

        if (context.isEmpty()) {
            // if the context is not empty, we have already consumed a VALUE_SEPARATOR(,) and standard JSON does not
            // allow trailing commas
            expected.add(JsonPunctuation.END_OBJECT);
        }

        if (tokenizer.getFragment() != 0) {
            // when the input is a StringTemplate, additional tokens are allowed
            expected.add(TT_JSON_STRING_TEMPLATE);
            expected.add(TT_JSON_PLACEHOLDER);
        }

        return unexpectedToken(token, expected.toArray());
    }

    private @NotNull JsonParserException unexpectedStartOfValue(@Nullable JsonToken token) {
        var expected = new ArrayList<>(9);
        expected.addAll(Arrays.asList(
                JsonPunctuation.BEGIN_OBJECT, JsonPunctuation.BEGIN_ARRAY, JsonBoolean.TRUE, JsonBoolean.FALSE,
                JsonNull.NULL, TT_JSON_STRING, TT_JSON_NUMBER
        ));
        if (tokenizer.getFragment() != 0) {
            expected.add(TT_JSON_STRING_TEMPLATE);
            expected.add(TT_JSON_PLACEHOLDER);
        }
        throw unexpectedToken(token, expected);
    }

    private @NotNull JsonParserException unexpectedEndOfFile() {
        return new JsonParserException(tokenizer.getFragment(), tokenizer.getLineNumber(), tokenizer.getColumnNumber(), "unexpected end of file");
    }

    private @NotNull JsonParserException unexpectedToken(@Nullable JsonToken token, @NotNull Object @NotNull... expected) {
        if (expected.length == 1) {
            throw new JsonParserException(
                    tokenizer.getFragment(), tokenizer.getLineNumber(), tokenizer.getColumnNumber(),
                    "unexpected token: " + token + " (expected " + expected[0] + ")"
            );
        } else {
            throw new JsonParserException(
                    tokenizer.getFragment(), tokenizer.getLineNumber(), tokenizer.getColumnNumber(),
                    "unexpected token: " + token + " (expected one of " + Arrays.toString(expected) + ")"
            );
        }
    }

    /**
     * Represents an JSON object or array that is currently being parsed.
     */
    private sealed interface Context {
        boolean isEmpty();

        /**
         * Adds a nested JSON value to this context
         * @param value the value
         */
        void add(@Nullable JsonValue value);


        /**
         * Builds and returns the JSON value represented by this context.
         * @return a JSON value
         */
        @NotNull JsonValue build();

        /**
         * {@return the token that indicates the end of this context}
         */
        @NotNull JsonPunctuation end();
    }

    /**
     * Represents a JSON object that is currently being parsed.
     */
    private static final class ObjectContext implements Context {
        private final @NotNull SequencedMap<@NotNull String, @Nullable JsonValue> map = new LinkedHashMap<>();

        private @Nullable String key;

        public void setKey(@Nullable String key) {
            if (this.key != null) throw new IllegalStateException();
            this.key = key;
        }

        @Override
        public void add(@Nullable JsonValue value) {
            if (key == null) throw new IllegalStateException();
            map.put(key, value);
            key = null;
        }

        @Override
        public boolean isEmpty() {
            return map.isEmpty();
        }

        @Override
        public @NotNull JsonValue build() {
            if (key != null) throw new IllegalStateException();
            return new JsonObject(map);
        }

        @Override
        public @NotNull JsonPunctuation end() {
            return JsonPunctuation.END_OBJECT;
        }
    }

    /**
     * Represents a JSON array that is currently being parsed.
     */
    private static final class ArrayContext implements Context {
        private final @NotNull List<@Nullable JsonValue> list = new ArrayList<>();

        @Override
        public void add(@Nullable JsonValue value) {
            list.add(value);
        }

        @Override
        public boolean isEmpty() {
            return list.isEmpty();
        }

        @Override
        public @NotNull JsonArray build() {
            return new JsonArray(list);
        }

        @Override
        public @NotNull JsonPunctuation end() {
            return JsonPunctuation.END_ARRAY;
        }
    }
}
