package eu.jonahbauer.json.tokenizer;

import eu.jonahbauer.json.tokenizer.token.JsonToken;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.Range;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public interface JsonTokenizer extends Iterable<JsonToken> {

    /**
     * {@return the next token or <code>null</code> if the end of the input has been reached}
     * @throws IOException if an I/O error occurs
     */
    @Nullable JsonToken next() throws IOException;

    /**
     * {@return the fragment at which the previously read token appeared}
     */
    @Range(from = 0, to = Integer.MAX_VALUE) int getFragment();

    /**
     * {@return the line at which the previously read token appeared}
     */
    @Range(from = 1, to = Integer.MAX_VALUE) int getLineNumber();

    /**
     * {@return the column at which the previously read token appeared}
     */
    @Range(from = 1, to = Integer.MAX_VALUE) int getColumnNumber();

    /**
     * {@return an iterator over the tokens} The {@link Iterator#next()} and {@link Iterator#hasNext()} methods
     * may throw an {@link UncheckedIOException} if an I/O error occurs.
     */
    @Override
    default @NotNull Iterator<@NotNull JsonToken> iterator() {
        class JsonTokenizerIterator implements Iterator<@NotNull JsonToken> {
            private @Nullable JsonToken next;
            private boolean isValid = false;

            private void ensureValid() {
                if (isValid) return;

                try {
                    next = JsonTokenizer.this.next();
                    isValid = true;
                } catch (IOException ex) {
                    throw new UncheckedIOException(ex);
                }
            }

            @Override
            public boolean hasNext() {
                ensureValid();
                return next != null;
            }

            @Override
            public @NotNull JsonToken next() {
                ensureValid();
                if (next == null) {
                    throw new NoSuchElementException();
                }

                var token = next;
                next = null;
                isValid = false;
                return token;
            }
        }

        return new JsonTokenizerIterator();
    }

    /**
     * {@return a stream of tokens} When an I/O error occurs, the corresponding {@link IOException} is wrapped in an
     * {@link UncheckedIOException} exception.
     */
    default @NotNull Stream<@NotNull JsonToken> stream() {
        return StreamSupport.stream(this.spliterator(), false);
    }
}
