package eu.jonahbauer.json.tokenizer;

import eu.jonahbauer.json.exceptions.JsonTokenizerException;
import eu.jonahbauer.json.tokenizer.reader.PushbackReader;
import eu.jonahbauer.json.tokenizer.reader.PushbackReaderImpl;
import eu.jonahbauer.json.tokenizer.token.*;
import lombok.Getter;
import eu.jonahbauer.json.JsonBoolean;
import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonString;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.*;
import java.util.ArrayList;
import java.util.regex.Pattern;

public final class JsonTokenizerImpl implements JsonTokenizer {
    private static final Pattern NUMBER_PATTERN = Pattern.compile("-?(?:0|[1-9]\\d*)(?:\\.\\d+)?([eE][+-]?\\d+)?");
    private static final JsonPunctuation[] PUNCTUATION;

    static {
        PUNCTUATION = new JsonPunctuation[128];
        for (var punctuation : JsonPunctuation.values()) {
            PUNCTUATION[punctuation.value()] = punctuation;
        }
    }

    private final @NotNull PushbackReader reader;
    private final int[] buffer = new int[4];

    @Getter
    private int fragment;
    @Getter
    private int lineNumber;
    @Getter
    private int columnNumber;

    public JsonTokenizerImpl(@NotNull String json) {
        this(new StringReader(json));
    }

    @SuppressWarnings("preview")
    public JsonTokenizerImpl(@NotNull StringTemplate template) {
        this.reader = new PushbackReaderImpl(template);
    }

    public JsonTokenizerImpl(@NotNull Reader reader) {
        this.reader = new PushbackReaderImpl(reader);
    }

    @Override
    public @Nullable JsonToken next() throws IOException {
        int chr;

        do {
            chr = reader.read();
        } while (isWhitespace(chr));

        fragment = reader.getFragment();
        lineNumber = reader.getLineNumber();
        columnNumber = reader.getColumnNumber();

        if (chr == PushbackReader.EOF) return null;
        if (chr == PushbackReader.PLACEHOLDER) return new JsonPlaceholder(reader.getObject());
        if (chr < 128 && PUNCTUATION[chr] != null) return PUNCTUATION[chr];
        if (chr == '"') return nextString();

        reader.pushback();
        String token = nextToken();
        return switch (token) {
            case "true" -> JsonBoolean.TRUE;
            case "false" -> JsonBoolean.FALSE;
            case "null" -> JsonNull.NULL;
            case String number when isNumberLiteral(number) -> {
                try {
                    yield new JsonNumber(Double.parseDouble(number));
                } catch (IllegalArgumentException ex) {
                    throw error("invalid number literal: " + number, ex);
                }
            }
            default -> throw error("invalid token: " + token);
        };
    }

    private boolean isWhitespace(int chr) {
        return chr == ' ' || chr == '\t' || chr == '\r' || chr == '\n';
    }

    private boolean isPunctuation(int chr) {
        return chr == '{' || chr == '}' || chr == '[' || chr == ']' || chr == ',' || chr == ':' || chr == '"';
    }

    private boolean isNumberLiteral(@NotNull String token) {
        return NUMBER_PATTERN.matcher(token).matches();
    }

    private @NotNull String nextToken() throws IOException {
        StringBuilder out = new StringBuilder();
        while (true) {
            int chr = reader.read();
            if (isWhitespace(chr) || isPunctuation(chr)) {
                reader.pushback();
                return out.toString();
            } else if (chr == -1) {
                return out.toString();
            } else {
                out.append((char) chr);
            }
        }
    }

    private @NotNull JsonToken nextString() throws IOException {
        var fragments = new ArrayList<String>();
        var values = new ArrayList<>();

        var current = new StringBuilder();
        while (true) {
            int chr = reader.read();
            if (chr == '"') {
                if (fragments.isEmpty()) {
                    return new JsonString(current.toString());
                } else {
                    fragments.add(current.toString());
                    @SuppressWarnings("preview")
                    var result = new JsonStringTemplate(StringTemplate.of(fragments, values));
                    return result;
                }
            } else if (chr == '\\') {
                chr = reader.read();
                switch (chr) {
                    case PushbackReader.EOF -> throw error("incomplete escape sequence in string literal");
                    case '"' -> current.append('"');
                    case '\\' -> current.append('\\');
                    case '/' -> current.append('/');
                    case 'b' -> current.append('\b');
                    case 'f' -> current.append('\f');
                    case 'n' -> current.append('\n');
                    case 'r' -> current.append('\r');
                    case 't' -> current.append('\t');
                    case 'u' -> {
                        buffer[0] = reader.read();
                        buffer[1] = reader.read();
                        buffer[2] = reader.read();
                        buffer[3] = reader.read();

                        if (buffer[0] < 0 || buffer[1] < 0 || buffer[2] < 0 || buffer[3] < 0) {
                            throw error("incomplete escape sequence in string literal");
                        }

                        int code = 0;
                        for (int i = 0; i < 4; i++) {
                            code *= 16;
                            char esc = (char) buffer[i];
                            if ('0' <= esc && esc <= '9') {
                                code += esc - '0';
                            } else if ('a' <= esc && esc <= 'f') {
                                code += 11 + (esc - 'a');
                            } else if ('A' <= esc && esc <= 'F') {
                                code += 11 + (esc - 'A');
                            } else {
                                throw error("invalid character " + toString(buffer[i]) + " in escape sequence");
                            }
                        }
                        current.append((char) code);
                    }
                    default -> throw error("invalid character " + toString(chr) + " in escape sequence");
                }

            } else if (chr == PushbackReader.EOF) {
                throw error("unclosed string literal");
            } else if (chr == PushbackReader.PLACEHOLDER) {
                fragments.add(current.toString());
                current.setLength(0);
                values.add(reader.getObject());
            } else if (chr < 32) {
                throw error("unescaped control character in string literal");
            } else {
                current.append((char) chr);
            }
        }
    }

    private @NotNull JsonTokenizerException error(@NotNull String message) {
        return new JsonTokenizerException(getFragment(), getLineNumber(), getColumnNumber(), message);
    }

    private @NotNull JsonTokenizerException error(@NotNull String message, @NotNull Throwable cause) {
        return new JsonTokenizerException(getFragment(), getLineNumber(), getColumnNumber(), message, cause);
    }

    private static @NotNull String toString(int chr) {
        return switch (chr) {
            case PushbackReader.EOF -> "EOF";
            case '\t' -> "'\\t'";
            case '\r' -> "'\\r'";
            case '\n' -> "'\\n'";
            case '\b' -> "'\\b'";
            case '\f' -> "'\\f'";
            default -> {
                if (Character.isISOControl((char) chr)) {
                    yield "0x" + Integer.toHexString(chr);
                } else {
                    yield "'" + (char) chr + "'";
                }
            }
        };
    }

}
