package eu.jonahbauer.json.tokenizer.reader;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.*;

public class PushbackReaderImpl implements PushbackReader {
    private @Nullable Input input;
    private @Nullable Reader reader;

    private int current = Integer.MIN_VALUE;
    private Object object;

    private int fragment;
    private int lineNumber;
    private int columnNumber;

    private boolean pushback;

    public PushbackReaderImpl(@NotNull Reader reader) {
        this.input = new Input(List.of(reader).iterator(), Collections.emptyIterator());
        this.fragment = -1; // fragment will be incremented to 0 when the first character is read
    }

    @SuppressWarnings("preview")
    public PushbackReaderImpl(@NotNull StringTemplate template) {
        this.input = new Input(
                template.fragments().stream().<Reader>map(StringReader::new).iterator(),
                template.values().iterator()
        );
        this.fragment = 0; // fragment will be incremented to 1 when the first character is read
    }

    @Override
    public int getFragment() {
        if (current == Integer.MIN_VALUE) {
            throw new IllegalStateException("No character has been read so far.");
        }
        return fragment;
    }

    @Override
    public int getLineNumber() {
        if (current == Integer.MIN_VALUE) {
            throw new IllegalStateException("No character has been read so far.");
        }
        return lineNumber;
    }

    @Override
    public int getColumnNumber() {
        if (current == Integer.MIN_VALUE) {
            throw new IllegalStateException("No character has been read so far.");
        }
        return columnNumber;
    }

    @Override
    public void pushback() {
        if (current == Integer.MIN_VALUE) {
            throw new IllegalStateException("No character has been read so far.");
        } else if (pushback) {
            throw new IllegalStateException("Cannot push back more than one character at a time.");
        }
        pushback = true;
    }

    @Override
    public int read() throws IOException {
        if (pushback) {
            pushback = false;
            return current;
        }

        var input = this.input;
        if (input == null) {
            throw new IOException("closed");
        }

        var reader = this.reader;
        if (reader == null && input.readers().hasNext()) {
            this.reader = reader = input.readers().next();
            fragment++;
            lineNumber = 1;
            columnNumber = 0;
        } else if (reader == null) {
            return EOF;
        }

        int result = reader.read();

        if (current == '\n' && result != '\n' || current == '\r' && result != '\n' && result != '\r') {
            lineNumber++;
            columnNumber = 1;
        } else {
            columnNumber++;
        }

        if (result == EOF) {
            reader.close();
            this.reader = null;

            if (input.values().hasNext()) {
                this.object = input.values().next();
                return this.current = PLACEHOLDER;
            }
        }

        this.object = null;
        return this.current = result;
    }

    @Override
    public @Nullable Object getObject() {
        if (current != PLACEHOLDER) {
            throw new IllegalStateException("Currently not at a placeholder.");
        }
        return object;
    }

    @Override
    public void close() throws IOException {
        var input = this.input;
        var reader = this.reader;
        this.input = null;
        this.reader = null;
        close(Arrays.asList(input, reader).iterator());
    }

    private static void close(@NotNull Iterator<? extends @Nullable AutoCloseable> closeables) throws ClosingException {
        ClosingException exception = null;
        while (closeables.hasNext()) {
            var closeable = closeables.next();
            if (closeable == null) continue;
            try {
                closeable.close();
            } catch (Throwable t) {
                if (exception == null) exception = new ClosingException();
                exception.addSuppressed(t);
            }
        }
        if (exception != null) throw exception;
    }

    private record Input(@NotNull Iterator<@NotNull Reader> readers, @NotNull Iterator<Object> values) implements AutoCloseable {

        @Override
        public void close() throws ClosingException {
            PushbackReaderImpl.close(readers);
        }
    }

    private static final class ClosingException extends IOException { }
}
