package eu.jonahbauer.json.tokenizer;

import eu.jonahbauer.json.JsonBoolean;
import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonString;
import eu.jonahbauer.json.exceptions.JsonTokenizerException;
import eu.jonahbauer.json.tokenizer.token.*;
import org.jetbrains.annotations.NotNull;
import org.junit.jupiter.api.Test;

import java.util.List;

import static java.lang.StringTemplate.RAW;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;

class JsonTokenizerImplTest {

    @Test
    void simpleTokens() {
        test("{[]},:", JsonPunctuation.BEGIN_OBJECT, JsonPunctuation.BEGIN_ARRAY, JsonPunctuation.END_ARRAY, JsonPunctuation.END_OBJECT, JsonPunctuation.VALUE_SEPARATOR, JsonPunctuation.NAME_SEPARATOR);
    }

    @Test
    void literals() {
        test("true", JsonBoolean.TRUE);
        test("false", JsonBoolean.FALSE);
        test("null", JsonNull.NULL);
        test("null true false", JsonNull.NULL, JsonBoolean.TRUE, JsonBoolean.FALSE);
    }

    @Test
    void misspelledLiterals() {
        assertThrows(JsonTokenizerException.class, () -> test("tru"));
        assertThrows(JsonTokenizerException.class, () -> test("fal"));
        assertThrows(JsonTokenizerException.class, () -> test("nu"));
        assertThrows(JsonTokenizerException.class, () -> test("TRUE"));
        assertThrows(JsonTokenizerException.class, () -> test("FALSE"));
        assertThrows(JsonTokenizerException.class, () -> test("NULL"));
    }

    @Test
    void stringWithoutEscapes() {
        test("\"foobar\"\"baz\"", new JsonString("foobar"), new JsonString("baz"));
    }

    @Test
    void stringWithSimpleEscapes() {
        test("\"\\b\\t\\f\\r\\n\\/\\\\\\\"\"", new JsonString("\b\t\f\r\n/\\\""));
    }

    @Test
    void stringWithUnicodeEscapes() {
        test("\"\\u0041\\u0042\\u0043\"", new JsonString("ABC"));
    }

    @Test
    void stringWithInvalidEscapeSequence() {
        assertThrows(JsonTokenizerException.class, () -> test("\"\\x\""));
        assertThrows(JsonTokenizerException.class, () -> test("\"\\uxxxx\""));
        assertThrows(JsonTokenizerException.class, () -> test("\"\\uaa\""));
    }

    @Test
    void stringWithControlCharacters() {
        assertThrows(JsonTokenizerException.class, () -> test("\"\u0010\""));
    }

    @Test
    void stringWithoutTrailingQuotes() {
        assertThrows(JsonTokenizerException.class, () -> test("\"hello world"));
    }

    @Test
    void number() {
        test("-0", new JsonNumber(-0.0));
        test("0", new JsonNumber(0.0));
        test("123", new JsonNumber(123.0));
        test("123.456", new JsonNumber(123.456));
        test("-123", new JsonNumber(-123.0));
        test("-123.456", new JsonNumber(-123.456));
        test("1E10", new JsonNumber(1E10));
        test("1E+10", new JsonNumber(1E+10));
        test("1E-10", new JsonNumber(1E-10));
        test("123.456E10", new JsonNumber(123.456E10));
        test("123.456E+10", new JsonNumber(123.456E+10));
        test("123.456E-10", new JsonNumber(123.456E-10));
    }

    @Test
    void numberWithLeadingZero() {
        assertThrows(JsonTokenizerException.class, () -> test("00"));
        assertThrows(JsonTokenizerException.class, () -> test("-00"));
    }

    @Test
    void numberWithEmptyIntegralPart() {
        assertThrows(JsonTokenizerException.class, () -> test(".0"));
    }

    @Test
    void numberWithEmptyFractionPart() {
        assertThrows(JsonTokenizerException.class, () -> test("0."));
    }

    @Test
    void numberWithEmptyExponentPart() {
        assertThrows(JsonTokenizerException.class, () -> test("0.0E"));
    }

    @Test
    void numberWithGrouping() {
        assertThrows(JsonTokenizerException.class, () -> test("1_234"));
        test("1,234", new JsonNumber(1), JsonPunctuation.VALUE_SEPARATOR, new JsonNumber(234));
    }

    @Test
    @SuppressWarnings("preview")
    void stringTemplate() {
        var name = "World";
        test(RAW."\"Hello, \{name}!\"", new JsonStringTemplate(RAW."Hello, \{name}!"));
    }

    @Test
    @SuppressWarnings("preview")
    void placeholder() {
        var name = "World";
        test(RAW."\{name}", new JsonPlaceholder(name));
    }

    private void test(@NotNull String json, @NotNull JsonToken @NotNull... expected) {
        var tokenizer = new JsonTokenizerImpl(json);
        var actual = tokenizer.stream().toList();
        assertEquals(List.of(expected), actual);
    }

    @SuppressWarnings("preview")
    private void test(@NotNull StringTemplate json, @NotNull JsonToken @NotNull... expected) {
        var tokenizer = new JsonTokenizerImpl(json);
        var actual = tokenizer.stream().toList();
        assertEquals(List.of(expected), actual);
    }
}
