package eu.jonahbauer.json.query;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.exceptions.JsonParserException;
import eu.jonahbauer.json.query.impl.Generator;
import eu.jonahbauer.json.query.parser.ast.*;
import eu.jonahbauer.json.query.util.Util;
import lombok.experimental.UtilityClass;
import org.intellij.lang.annotations.MagicConstant;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.function.*;
import java.util.regex.MatchResult;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.*;

@UtilityClass
@SuppressWarnings("preview")
public class JsonMath {
    private static final @NotNull Comparator<@Nullable JsonValue> COMPARATOR = JsonMath::compare;

    //<editor-fold desc="simple arithmetic operations" defaultstate="collapsed">
    public static @Nullable JsonValue add(@Nullable JsonValue first, @Nullable JsonValue second) {
        if (first instanceof JsonNumber(double x) && second instanceof JsonNumber(double y)) {
            return new JsonNumber(x + y);
        } else if (first instanceof JsonArray x && second instanceof JsonArray y) {
            var out = new ArrayList<JsonValue>(x.size() + y.size());
            out.addAll(x);
            out.addAll(y);
            return new JsonArray(out);
        } else if (first instanceof JsonString(var x) && second instanceof JsonString(var y)) {
            return new JsonString(x + y);
        } else if (first instanceof JsonObject x && second instanceof JsonObject y) {
            var out = new HashMap<>(x);
            out.putAll(y);
            return new JsonObject(out);
        } else if (first == null) {
            return second;
        } else if (second == null) {
            return first;
        } else {
            throw error(first, second, "cannot be added");
        }
    }

    public static @NotNull JsonValue sub(@Nullable JsonValue first, @Nullable JsonValue second) {
        if (first instanceof JsonNumber(double x) && second instanceof JsonNumber(double y)) {
            return new JsonNumber(x - y);
        } else if (first instanceof JsonArray x && second instanceof JsonArray y) {
            var out = new ArrayList<>(x);
            out.removeAll(y);
            return new JsonArray(out);
        } else {
            throw error(first, second, "cannot be subtracted");
        }
    }

    public static @Nullable JsonValue mul(@Nullable JsonValue first, @Nullable JsonValue second) {
        if (first instanceof JsonNumber(double x) && second instanceof JsonNumber(double y)) {
            return new JsonNumber(x * y);
        } else if (first instanceof JsonNumber(double n) && second instanceof JsonString(var str)) {
            if (n == 0) return JsonString.EMPTY;
            if (n < 0 || Math.ceil(n) > Integer.MAX_VALUE) return null;
            return new JsonString(str.repeat((int) Math.ceil(n)));
        } else if (first instanceof JsonString(var str) && second instanceof JsonNumber(var n)) {
            if (n == 0) return JsonString.EMPTY;
            if (n < 0 || Math.ceil(n) > Integer.MAX_VALUE) return null;
            return new JsonString(str.repeat((int) Math.ceil(n)));
        } else if (first instanceof JsonObject x && second instanceof JsonObject y) {
            var out = new HashMap<>(x);
            y.forEach((key, value) -> out.merge(key, value, (a, b) ->
                a instanceof JsonObject && b instanceof JsonObject ? mul(a, b) : b
            ));
            return new JsonObject(out);
        } else {
            throw error(first, second, "cannot be multiplied");
        }
    }

    public static @NotNull JsonValue div(@Nullable JsonValue first, @Nullable JsonValue second) {
        if (first instanceof JsonNumber(double x) && second instanceof JsonNumber(double y)) {
            if (y == 0) throw error(first, second, "cannot be divided because the divisor is zero");
            return new JsonNumber(x / y);
        } else if (first instanceof JsonString(var x) && second instanceof JsonString(var y)) {
            return new JsonArray(Arrays.stream(x.split(Pattern.quote(y), -1)).map(JsonValue::valueOf).toList());
        } else {
            throw error(first, second, "cannot be divided");
        }
    }

    public static @NotNull JsonNumber mod(@Nullable JsonValue first, @Nullable JsonValue second) {
        if (first instanceof JsonNumber(double x) && second instanceof JsonNumber(double y)) {
            if (y == 0) throw error(first, second, "cannot be divided (remainder) because the divisor is zero");
            return new JsonNumber(x % y);
        } else {
            throw error(first, second, "cannot be divided (remainder)");
        }
    }

    public static @NotNull JsonNumber neg(@Nullable JsonValue value) {
        if (value instanceof JsonNumber(double x)) {
            return new JsonNumber(- x);
        } else {
            throw error(value, "cannot be negated");
        }
    }
    //</editor-fold>

    //<editor-fold desc="boolean operations" defaultstate="collapsed">
    public static @NotNull JsonBoolean and(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(isTruthy(first) && isTruthy(second));
    }

    public static @NotNull JsonBoolean or(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(isTruthy(first) || isTruthy(second));
    }

    public static @NotNull JsonBoolean not(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isFalsy(value));
    }

    public static boolean isTruthy(@Nullable JsonValue value) {
        return value != null && value != JsonBoolean.FALSE;
    }

    public static boolean isFalsy(@Nullable JsonValue value) {
        return !isTruthy(value);
    }
    //</editor-fold>

    //<editor-fold desc="comparisons" defaultstate="collapsed">
    public static @NotNull JsonBoolean eq(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(Objects.equals(first, second));
    }

    public static @NotNull JsonBoolean neq(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(!Objects.equals(first, second));
    }

    public static @NotNull JsonBoolean lt(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(compare(first, second) < 0);
    }

    public static @NotNull JsonBoolean gt(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(compare(first, second) > 0);
    }

    public static @NotNull JsonBoolean leq(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(compare(first, second) <= 0);
    }

    public static @NotNull JsonBoolean geq(@Nullable JsonValue first, @Nullable JsonValue second) {
        return JsonBoolean.valueOf(compare(first, second) >= 0);
    }

    public static @Nullable JsonValue abs(@Nullable JsonValue value) {
        return compare(value, JsonNumber.ZERO) < 0 ? neg(value) : value;
    }

    public static @NotNull Comparator<@Nullable JsonValue> comparator() {
        return COMPARATOR;
    }

    public static int compare(@Nullable JsonValue first, @Nullable JsonValue second) {
        int firstType = compareByType(first);
        int secondType = compareByType(second);
        if (firstType != secondType) return Integer.compare(firstType, secondType);

        if (first instanceof JsonNumber x && second instanceof JsonNumber y) {
            return compareNumbers(x, y);
        } else if (first instanceof JsonString x && second instanceof JsonString y) {
            return compareStrings(x, y);
        } else if (first instanceof JsonArray x && second instanceof JsonArray y) {
            return compareLexicographically(x, y, COMPARATOR);
        } else if (first instanceof JsonObject x && second instanceof JsonObject y) {
            return compareObjects(x, y);
        } else {
            // order of booleans and null has already been established via compareByType
            throw new AssertionError();
        }
    }

    private static int compareByType(@Nullable JsonValue value) {
        return switch (value) {
            case null -> 0;
            case JsonBoolean.FALSE -> 1;
            case JsonBoolean.TRUE -> 2;
            case JsonNumber _ -> 3;
            case JsonString _ -> 4;
            case JsonArray _ -> 5;
            case JsonObject _ -> 6;
        };
    }

    private static int compareNumbers(@NotNull JsonNumber first, @NotNull JsonNumber second) {
        return Double.compare(first.value(), second.value());
    }

    private static int compareStrings(@NotNull JsonString first, @NotNull JsonString second) {
        return first.value().compareTo(second.value());
    }

    private static int compareObjects(@NotNull JsonObject first, @NotNull JsonObject second) {
        // compare sorted key sets lexicographically
        var x1 = new TreeSet<>(first.keySet());
        var x2 = new TreeSet<>(second.keySet());
        var d = compareLexicographically(x1, x2, Comparator.naturalOrder());
        if (d != 0) return d;

        // compare each value order of sorted keys
        for (var key : x1) {
            var v1 = first.get(key);
            var v2 = second.get(key);
            var dk = compare(v1, v2);
            if (dk != 0) return dk;
        }
        return 0;
    }

    private static <T> int compareLexicographically(@NotNull Iterable<T> first, @NotNull Iterable<T> second, @NotNull Comparator<T> comparator) {
        var it1 = first.iterator();
        var it2 = second.iterator();
        while (it1.hasNext() && it2.hasNext()) {
            var d = comparator.compare(it1.next(), it2.next());
            if (d != 0) return d;
        }
        if (it1.hasNext()) return 1;
        if (it2.hasNext()) return -1;
        return 0;
    }
    //</editor-fold>

    //<editor-fold desc="error handling" defaultstate="collapsed">
    public static @NotNull Generator<@Nullable JsonValue> error(@Nullable JsonValue value) {
        if (value != null) throw new JsonQueryUserException(value);
        return Generator.empty();
    }

    public static @NotNull Generator<@Nullable JsonValue> halt() {
        throw new JsonQueryHaltException(null, 0);
    }

    public static @NotNull Generator<@Nullable JsonValue> halt(@Nullable JsonValue value) {
        throw new JsonQueryHaltException(value, 5);
    }

    public static @NotNull Generator<@Nullable JsonValue> halt(@Nullable JsonValue value, @Nullable JsonValue code) {
        throw new JsonQueryHaltException(value, switch (code) {
            case JsonNumber(var number) -> (int) number;
            case null, default -> throw new JsonQueryException(STR."\{Util.type(value)} (\{Util.value(value)}) halt_error/1: number required");
        });
    }
    //</editor-fold>

    //<editor-fold desc="stream operations" defaultstate="collapsed">
    public static @NotNull Generator<@NotNull JsonValue> range(@Nullable JsonValue upto) {
        return range(new JsonNumber(0), upto);
    }

    public static @NotNull Generator<@NotNull JsonValue> range(@Nullable JsonValue from, @Nullable JsonValue upto) {
        var min = switch (from) {
            case JsonNumber(var number) -> number;
            case null, default -> throw new JsonQueryException("Range bounds must be numeric.");
        };
        var max = switch (upto) {
            case JsonNumber(var number) -> number;
            case null, default -> throw new JsonQueryException("Range bounds must be numeric.");
        };
        return Generator.from(IntStream.range((int) Math.ceil(min), (int) Math.ceil(max)).boxed()).map(JsonNumber::new);
    }

    public static @NotNull Generator<@Nullable JsonValue> range(@Nullable JsonValue from, @Nullable JsonValue upto, @Nullable JsonValue by) {
        if (compare(by, JsonNumber.ZERO) > 0) {
            return Generator.from(Stream.iterate(from, i -> compare(i, upto) < 0, i -> add(i, by)));
        } else if (compare(by, JsonNumber.ZERO) < 0) {
            return Generator.from(Stream.iterate(from, i -> compare(i, upto) > 0, i -> add(i, by)));
        } else {
            return Generator.empty();
        }
    }

    public static @NotNull Generator<@Nullable JsonValue> limit(@NotNull Generator<@Nullable JsonValue> stream, @Nullable JsonValue limit) {
        if (compare(limit, JsonNumber.ZERO) > 0) {
            return new Generator<>() {
                private @Nullable JsonValue i = limit;

                @Override
                public boolean hasNext() {
                    return compare(i, JsonNumber.ZERO) > 0;
                }

                @Override
                public JsonValue next() {
                    if (compare(i, JsonNumber.ZERO) <= 0) throw new NoSuchElementException();
                    i = sub(i, JsonNumber.ONE);
                    return stream.next();
                }
            };
        } else if (Objects.equals(limit, JsonNumber.ZERO)) {
            return Generator.empty();
        } else {
            return stream;
        }
    }

    public static @NotNull Generator<@Nullable JsonValue> first(@NotNull Generator<@Nullable JsonValue> stream) {
        return stream.limit(1);
    }

    public static @NotNull Generator<@Nullable JsonValue> last(@NotNull Generator<@Nullable JsonValue> stream) {
        return new Generator<>() {
            @Override
            public boolean hasNext() {
                return stream.hasNext();
            }

            @Override
            public JsonValue next() throws NoSuchElementException {
                var out = stream.next();
                while (stream.hasNext()) {
                    out = stream.next();
                }
                return out;
            }
        };
    }

    public static @NotNull Generator<@Nullable JsonValue> nth(@NotNull Generator<@Nullable JsonValue> stream, @Nullable JsonValue limit) {
        if (compare(limit, JsonNumber.ZERO) < 0) throw new JsonQueryException("nth doesn't support negative indices");
        return Generator.of(() -> {
            for (var i = sub(add(limit, JsonNumber.ONE), JsonNumber.ONE); compare(i, JsonNumber.ZERO) > 0; i = sub(i, JsonNumber.ONE)) {
                stream.next();
            }
            return stream.next();
        });
    }

    public static @NotNull Generator<@Nullable JsonValue> select(@Nullable JsonValue value, @NotNull JQFilter filter) {
        return filter.apply(value).mapMulti((result, consumer) -> {
            if (isTruthy(result)) consumer.accept(value);
        });
    }
    //</editor-fold>

    //<editor-fold desc="iterable operations" defaultstate="collapsed">
    public static @Nullable JsonValue index(@Nullable JsonValue value, @Nullable JsonValue index) {
        return switch (index) {
            case JsonNumber(var number) -> switch (value) {
                case JsonArray array when isInt(number) && 0 <= number && number < array.size() -> array.get((int) number);
                case JsonArray array when isInt(number) && -array.size() <= number && number < 0 -> array.get(array.size() + (int) number);
                case JsonArray _ -> null;
                case null -> null;
                default -> throw indexError(value, index);
            };
            case JsonString(var string) -> switch (value) {
                case JsonObject object -> object.get(string);
                case null -> null;
                default -> throw indexError(value, index);
            };
            case JsonArray indexArray -> switch (value) {
                case JsonArray _ when indexArray.isEmpty() -> JsonArray.EMPTY;
                case JsonArray array -> {
                    var out = new ArrayList<JsonValue>();
                    var indexLength = indexArray.size();
                    var length = array.size();
                    outer: for (int i = 0; i < length - indexLength; i++) {
                        for (int j = 0; j < indexLength; j++) {
                            if (!Objects.equals(array.get(i + j), indexArray.get(j))) continue outer;
                        }
                        out.add(new JsonNumber(i));
                    }
                    yield new JsonArray(out);
                }
                case null, default -> throw indexError(value, index);
            };
            case null, default -> throw indexError(value, index);
        };
    }

    private static boolean isInt(double value) {
        return ((int) value) == value;
    }

    private static @NotNull JsonQueryException indexError(@Nullable JsonValue value, @Nullable JsonValue index) {
        var valueString = Util.type(value);
        var indexString = Util.type(index) + (index instanceof JsonString(var string) ? " " + JsonString.quote(string) : "");
        return new JsonQueryException(STR."Cannot index \{valueString} with \{indexString}.");
    }

    public static @NotNull JsonArray keys(@Nullable JsonValue value) {
        return JsonArray.valueOf(switch (value) {
            case JsonArray array -> IntStream.range(0, array.size()).mapToObj(JsonNumber::new).toList();
            case JsonObject object -> new TreeSet<>(object.keySet()).stream().map(JsonString::new).toList();
            case null, default -> throw error(value, "has no keys");
        });
    }

    public static @NotNull JsonArray keysUnsorted(@Nullable JsonValue value) {
        return JsonArray.valueOf(switch (value) {
            case JsonArray array -> IntStream.range(0, array.size()).mapToObj(JsonNumber::new);
            case JsonObject object -> object.keySet().stream().map(JsonString::new);
            case null, default -> throw error(value, "has no keys");
        });
    }

    public static @NotNull JsonBoolean has(@Nullable JsonValue value, @Nullable JsonValue index) {
        return JsonBoolean.valueOf((boolean) switch (index) {
            case JsonNumber(var number) -> switch (value) {
                case JsonArray array -> 0 <= number && (int) number < array.size();
                case null -> false;
                default -> throw new JsonQueryException(STR."Cannot check whether \{Util.type(value)} has a number key.");
            };
            case JsonString(var string) -> switch (value) {
                case JsonObject object -> object.containsKey(string);
                case null -> false;
                default -> throw new JsonQueryException(STR."Cannot check whether \{Util.type(value)} has a string key.");
            };
            case null, default -> throw new JsonQueryException(STR."Cannot check whether \{Util.type(value)} has a \{Util.type(index)} key.");
        });
    }

    public static @NotNull JsonBoolean in(@Nullable JsonValue index, @Nullable JsonValue value) {
        return has(value, index);
    }

    public static @Nullable JsonValue first(@Nullable JsonValue value) {
        return index(value, JsonNumber.ZERO);
    }

    public static @Nullable JsonValue last(@Nullable JsonValue value) {
        return index(value, new JsonNumber(-1));
    }

    public static @NotNull Generator<@Nullable JsonValue> values(@Nullable JsonValue value) {
        return values(value, false);
    }

    public static @NotNull Generator<@Nullable JsonValue> values(@Nullable JsonValue value, boolean optional) {
        return switch (value) {
            case JsonArray array -> Generator.from(array);
            case JsonObject object -> Generator.from(object.values());
            case null, default -> {
                if (optional) yield Generator.empty();
                throw new JsonQueryException(STR."Cannot iterate over \{Util.type(value)} (\{Util.value(value)}).");
            }
        };
    }

    public static @NotNull JsonArray map(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return new JsonArray(values(value).flatMap(expression).toList());
    }

    public static @NotNull JsonValue mapValues(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array -> new JsonArray(Generator.from(array).map(expression).flatMap(s -> s.limit(1)).toList());
            case JsonObject object -> {
                var out = new LinkedHashMap<String, JsonValue>();
                object.forEach((key, v) -> expression.apply(v).limit(1).forEach(newValue -> out.put(key, newValue)));
                yield new JsonObject(out);
            }
            case null, default -> throw new JsonQueryException(STR."Cannot iterate over \{Util.type(value)} (\{Util.value(value)}).");
        };
    }

    public static @NotNull JsonBoolean any(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(values(value).anyMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonBoolean any(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return JsonBoolean.valueOf(values(value).flatMap(expression).anyMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonBoolean any(@NotNull Generator<@Nullable JsonValue> values, @NotNull JQFilter expression) {
        return JsonBoolean.valueOf(values.flatMap(expression).anyMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonBoolean all(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(values(value).allMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonBoolean all(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return JsonBoolean.valueOf(values(value).flatMap(expression).allMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonBoolean all(@NotNull Generator<@Nullable JsonValue> values, @NotNull JQFilter expression) {
        return JsonBoolean.valueOf(values.flatMap(expression).allMatch(JsonMath::isTruthy));
    }

    public static @NotNull JsonArray flatten(@Nullable JsonValue value) {
        return new JsonArray(values(value).flatMap(JsonMath::flatten0).toList());
    }

    public static @NotNull JsonArray flatten(@Nullable JsonValue value, @Nullable JsonValue depth) {
        return switch (depth) {
            case JsonNumber(var number) when number < 0 -> throw new JsonQueryException("flatten depth must not be negative");
            case JsonNumber(var number) -> new JsonArray(values(value).flatMap(v -> flatten0(v, (int) number)).toList());
            case null, default -> throw new JsonQueryException("flatten depth must be a number");
        };
    }

    private static @NotNull Generator<@Nullable JsonValue> flatten0(@Nullable JsonValue value) {
        return flatten0(value, -1);
    }

    private static @NotNull Generator<@Nullable JsonValue> flatten0(@Nullable JsonValue value, int depth) {
        return switch (value) {
            case JsonArray array when depth == 0 -> Generator.of(array);
            case JsonArray array -> {
                var d = depth < 0 ? depth : depth - 1;
                yield Generator.from(array).flatMap(v -> flatten0(v, d));
            }
            case null, default -> Generator.of(value);
        };
    }

    public static @NotNull JsonArray sort(@Nullable JsonValue value) {
        return switch (value) {
            case JsonArray array -> {
                var list = new ArrayList<>(array);
                list.sort(comparator());
                yield new JsonArray(list);
            }
            case null, default -> throw error(value, "cannot be sorted, as it is not an array");
        };
    }

    public static @NotNull JsonArray sort(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array -> new JsonArray(array.stream()
                    .map(v -> new SortEntry(v, expression.apply(v).toList()))
                    .sorted(SortEntry.COMPARATOR)
                    .map(SortEntry::value)
                    .toList()
            );
            case null, default -> throw error(value, "cannot be sorted, as it is not an array");
        };
    }

    public static @Nullable JsonValue min(@Nullable JsonValue value) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> null;
            case JsonArray array -> Collections.min(array, COMPARATOR);
            case null, default -> throw error(value, value, "cannot be iterated over");
        };
    }

    public static @Nullable JsonValue min(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> null;
            case JsonArray array -> {
                var it = array.iterator();
                var minValue = it.next();
                var minKey = expression.apply(minValue).toList();

                while (it.hasNext()) {
                    var nextValue = it.next();
                    var nextKey = expression.apply(nextValue).toList();
                    if (compareLexicographically(nextKey, minKey, COMPARATOR) < 0) {
                        minValue = nextValue;
                        minKey = nextKey;
                    }
                }

                yield minValue;
            }
            case null, default -> throw error(value, value, "cannot be iterated over");
        };
    }

    public static @Nullable JsonValue max(@Nullable JsonValue value) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> null;
            case JsonArray array -> Collections.max(array, COMPARATOR);
            case null, default -> throw error(value, "cannot be sorted as it is not an array");
        };
    }

    public static @Nullable JsonValue max(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> null;
            case JsonArray array -> {
                var it = array.iterator();
                var maxValue = it.next();
                var maxKey = expression.apply(maxValue).toList();

                while (it.hasNext()) {
                    var nextValue = it.next();
                    var nextKey = expression.apply(nextValue).toList();
                    if (compareLexicographically(nextKey, maxKey, COMPARATOR) > 0) {
                        maxValue = nextValue;
                        maxKey = nextKey;
                    }
                }

                yield maxValue;
            }
            case null, default -> throw error(value, value, "cannot be iterated over");
        };
    }

    public static @NotNull JsonArray unique(@Nullable JsonValue value) {
        return switch (value) {
            case JsonArray array when array.size() <= 1 -> array;
            case JsonArray array -> {
                var set = new TreeSet<>(COMPARATOR);
                set.addAll(array);
                yield  new JsonArray(set.stream().toList());
            }
            case null, default -> throw error(value, "cannot be sorted as it is not an array");
        };
    }

    public static @NotNull JsonArray unique(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> array;
            case JsonArray array -> {
                var map = new TreeMap<Iterable<JsonValue>, JsonValue>(SortEntry.LEXICOGRAPHIC_COMPARATOR);
                array.forEach(v -> map.putIfAbsent(expression.apply(v).toList(), v));
                yield  new JsonArray(map.values().stream().toList());
            }
            case null, default -> throw error(value, "cannot be sorted as it is not an array");
        };
    }

    public static @NotNull JsonArray group(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return switch (value) {
            case JsonArray array when array.isEmpty() -> array;
            case JsonArray array -> {
                var map = new TreeMap<Iterable<JsonValue>, List<JsonValue>>(SortEntry.LEXICOGRAPHIC_COMPARATOR);
                array.forEach(v -> {
                    var key = expression.apply(v).toList();
                    map.computeIfAbsent(key, _ -> new ArrayList<>()).add(v);
                });
                yield new JsonArray(map.values().stream().map(JsonArray::new).toList());
            }
            case null, default -> throw error(value, "cannot be sorted as it is not an array");
        };
    }

    public static @NotNull JsonArray reverse(@Nullable JsonValue value) {
        return switch (value) {
            case JsonArray array -> array.reversed();
            case null, default -> throw error(value, "cannot be reversed, as it is not an array");
        };
    }

    public static @NotNull JsonBoolean contains(@Nullable JsonValue container, @Nullable JsonValue content) {
        var out = contains0(container, content);
        if (out == null) throw error(container, content, "cannot have their containment checked");
        return out;
    }

    private static @Nullable JsonBoolean contains0(@Nullable JsonValue container, @Nullable JsonValue content) {
        if (container instanceof JsonArray acontainer && content instanceof JsonArray acontent) {
            outer: for (var e1 : acontent) {
                for (var e2 : acontainer) {
                   if (contains0(e2, e1) == JsonBoolean.TRUE) continue outer;
                }
                return JsonBoolean.FALSE;
            }
            return JsonBoolean.TRUE;
        } else if (container instanceof JsonString(var scontainer) && content instanceof JsonString(var scontent)) {
            return JsonBoolean.valueOf(scontainer.contains(scontent));
        } else if (container instanceof JsonObject ocontainer && content instanceof JsonObject ocontent) {
            for (var entry : ocontent.entrySet()) {
                if (contains0(ocontainer.get(entry.getKey()), entry.getValue()) != JsonBoolean.TRUE) {
                    return JsonBoolean.FALSE;
                }
            }
            return JsonBoolean.TRUE;
        } else {
            return null;
        }
    }

    public static @Nullable JsonValue indices(@Nullable JsonValue container, @Nullable JsonValue content) {
        if (container instanceof JsonArray acontainer && content instanceof JsonArray acontent) {
            return index(acontainer, acontent);
        } else if (container instanceof JsonArray acontainer) {
            return index(acontainer, JsonArray.of(content));
        } else if (container instanceof JsonString(var scontainer) && content instanceof JsonString(var scontent)) {
            var out = new ArrayList<JsonNumber>();
            var idx = -1;
            while (true) {
                idx = scontainer.indexOf(scontent, idx + 1);
                if (idx == -1) break;
                out.add(new JsonNumber(idx));
            }
            return new JsonArray(out);
        } else {
            return index(container, content);
        }
    }

    public static @Nullable JsonValue firstindex(@Nullable JsonValue container, @Nullable JsonValue content) {
        return index(indices(container, content), JsonNumber.ZERO);
    }

    public static @Nullable JsonValue lastindex(@Nullable JsonValue container, @Nullable JsonValue content) {
        return index(indices(container, content), new JsonNumber(-1));
    }

    public static @NotNull JsonBoolean inside(@Nullable JsonValue content, @Nullable JsonValue container) {
        return contains(container, content);
    }

    public static @NotNull Generator<@NotNull JsonArray> combinations(@Nullable JsonValue value) {
        if (length0(value) == 0) return Generator.of(JsonArray.EMPTY);
        if (!(value instanceof JsonArray array)) throw indexError(value, JsonNumber.ZERO);

        return values(array.getFirst())
                .flatMap(head -> combinations(array.subList(1, array.size())).map(tail -> {
                    var out = new ArrayList<JsonValue>(tail.size() + 1);
                    out.add(head);
                    out.addAll(tail);
                    return new JsonArray(out);
                }));
    }

    public static @NotNull Generator<@NotNull JsonArray> combinations(@Nullable JsonValue value, @Nullable JsonValue n) {
        if (!(n instanceof JsonNumber(var number))) throw new JsonQueryException("Range bounds must be numeric.");
        return combinations(new JsonArray(Collections.nCopies((int) Math.ceil(number), value)));
    }

    public static @NotNull JsonNumber bsearch(@Nullable JsonValue container, @Nullable JsonValue value) {
        if (length0(container) == 0) return new JsonNumber(-1);
        if (container instanceof JsonArray array) {
            return new JsonNumber(Collections.binarySearch(array, value, COMPARATOR));
        } else {
            throw indexError(container, JsonNumber.ZERO);
        }
    }

    public static @NotNull JsonArray transpose(@Nullable JsonValue value) {
        if (!(value instanceof JsonArray array)) throw indexError(value, JsonNumber.ZERO);

        var length = (int) array.stream().mapToDouble(JsonMath::length0).max().orElse(0.0);
        var out = new ArrayList<List<JsonValue>>(length);
        for (int i = 0; i < length; i++) out.add(new ArrayList<>());

        for (var element : array) {
            if (!(element instanceof JsonArray subarray)) throw indexError(element, JsonNumber.ZERO);
            int i = 0;
            for (; i < subarray.size(); i++) out.get(i).add(subarray.get(i));
            for (; i < length; i ++) out.get(i).add(null);
        }

        return new JsonArray(out.stream().map(JsonArray::new).toList());
    }

    private static final JsonString KEY = new JsonString("key");
    private static final JsonString VALUE = new JsonString("value");

    public static @NotNull JsonArray toEntries(@Nullable JsonValue value) {
        return new JsonArray((switch (value) {
            case JsonArray array -> IntStream.range(0, array.size())
                    .mapToObj(i -> JsonObject.of(
                            "key", new JsonNumber(i),
                            "value", array.get(i)
                    ));
            case JsonObject object -> object.entrySet().stream()
                    .map(entry -> JsonObject.of(
                            "key", new JsonString(entry.getKey()),
                            "value", entry.getValue()
                    ));
            case null, default -> throw error(value, "has no keys");
        }).toList());
    }

    public static @NotNull JsonObject fromEntries(@Nullable JsonValue value) {
        var out = new LinkedHashMap<String, JsonValue>();
        values(value).forEach(entry -> {
            var key = index(entry, KEY);
            var val = index(entry, VALUE);
            if (!(key instanceof JsonString(var k))) throw new JsonQueryException(STR."Cannot use \{Util.type(key)} (\{Util.value(key)}) as object key.");
            out.put(k, val);
        });
        return new JsonObject(out);
    }

    public static @NotNull JsonObject withEntries(@Nullable JsonValue value, @NotNull JQFilter expression) {
        return fromEntries(map(toEntries(value), expression));
    }

    private record SortEntry(@Nullable JsonValue value, @NotNull Iterable<@Nullable JsonValue> key) {
        private static final Comparator<Iterable<JsonValue>> LEXICOGRAPHIC_COMPARATOR = (a, b) -> compareLexicographically(a, b, JsonMath.COMPARATOR);
        private static final Comparator<SortEntry> COMPARATOR = Comparator.comparing(SortEntry::key, LEXICOGRAPHIC_COMPARATOR);
    }
    //</editor-fold>

    //<editor-fold desc="filters" defaultstate="collapsed">
    public static @NotNull JsonBoolean isArray(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isArray0(value));
    }

    public static @NotNull JsonBoolean isObject(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isObject0(value));
    }

    public static @NotNull JsonBoolean isIterable(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isIterable0(value));
    }

    public static @NotNull JsonBoolean isBoolean(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isBoolean0(value));
    }

    public static @NotNull JsonBoolean isNumber(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isNumber0(value));
    }

    public static @NotNull JsonBoolean isNormal(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isNormal0(value));
    }

    public static @NotNull JsonBoolean isFinite(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isFinite0(value));
    }

    public static @NotNull JsonBoolean isInfinite(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isInfinite0(value));
    }

    public static @NotNull JsonBoolean isNan(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isNan0(value));
    }

    public static @NotNull JsonBoolean isString(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isString0(value));
    }

    public static @NotNull JsonBoolean isNull(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isNull0(value));
    }

    public static @NotNull JsonBoolean isValue(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isValue0(value));
    }

    public static @NotNull JsonBoolean isScalar(@Nullable JsonValue value) {
        return JsonBoolean.valueOf(isScalar0(value));
    }

    public static @NotNull JsonBoolean isEmpty(@NotNull Generator<?> stream) {
        return JsonBoolean.valueOf(isEmpty0(stream));
    }

    public static boolean isArray0(@Nullable JsonValue value) {
        return value instanceof JsonArray;
    }

    public static boolean isObject0(@Nullable JsonValue value) {
        return value instanceof JsonObject;
    }

    public static boolean isIterable0(@Nullable JsonValue value) {
        return isArray0(value) || isObject0(value);
    }

    public static boolean isBoolean0(@Nullable JsonValue value) {
        return value instanceof JsonBoolean;
    }

    public static boolean isNumber0(@Nullable JsonValue value) {
        return value instanceof JsonNumber;
    }

    public static boolean isNormal0(@Nullable JsonValue value) {
        return value instanceof JsonNumber(var number) && Math.abs(number) >= Double.MIN_NORMAL;
    }

    public static boolean isFinite0(@Nullable JsonValue value) {
        return value instanceof JsonNumber(var number) && Double.isFinite(number);
    }

    public static boolean isInfinite0(@Nullable JsonValue value) {
        return value instanceof JsonNumber(var number) && Double.isInfinite(number);
    }

    public static boolean isNan0(@Nullable JsonValue value) {
        return value instanceof JsonNumber(var number) && Double.isNaN(number);
    }

    public static boolean isString0(@Nullable JsonValue value) {
        return value instanceof JsonString;
    }

    public static boolean isNull0(@Nullable JsonValue value) {
        return value == null;
    }

    public static boolean isValue0(@Nullable JsonValue value) {
        return value != null;
    }

    public static boolean isScalar0(@Nullable JsonValue value) {
        return !isIterable0(value);
    }

    public static boolean isEmpty0(@NotNull Generator<?> stream) {
        return !stream.hasNext();
    }
    //</editor-fold>

    //<editor-fold desc="string operations" defaultstate="collapsed">
    private static final Pattern TRIM = Pattern.compile("^\\s+|\\s+$");
    public static @NotNull JsonString trim(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> new JsonString(TRIM.matcher(string).replaceAll(""));
            case null, default -> throw error(value, "only strings can be trimmed");
        };
    }

    private static final Pattern LTRIM = Pattern.compile("^\\s+");
    public static @NotNull JsonString ltrim(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> new JsonString(LTRIM.matcher(string).replaceAll(""));
            case null, default -> throw error(value, "only strings can be trimmed");
        };
    }

    private static final Pattern RTRIM = Pattern.compile("\\s+$");
    public static @NotNull JsonString rtrim(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> new JsonString(RTRIM.matcher(string).replaceAll(""));
            case null, default -> throw error(value, "only strings can be trimmed");
        };
    }

    public static @NotNull JsonString ltrimstr(@Nullable JsonValue value, @Nullable JsonValue prefix) {
        if (!(value instanceof JsonString(var string))) throw error(value, "only strings can be trimmed");
        if (!(prefix instanceof JsonString(var prefixString))) throw error(value, "prefix must be string");
        return string.startsWith(prefixString) ? new JsonString(string.substring(prefixString.length())) : (JsonString) value;
    }

    public static @NotNull JsonString rtrimstr(@Nullable JsonValue value, @Nullable JsonValue suffix) {
        if (!(value instanceof JsonString(var string))) throw error(value, "only strings can be trimmed");
        if (!(suffix instanceof JsonString(var suffixString))) throw error(value, "suffix must be string");
        return string.endsWith(suffixString) ? new JsonString(string.substring(0, string.length() - suffixString.length())) : (JsonString) value;
    }

    public static @NotNull JsonArray split(@Nullable JsonValue input, @Nullable JsonValue separator) {
        if (input instanceof JsonString(var inputString) && separator instanceof JsonString(var separatorString)) {
            return new JsonArray(Arrays.stream(inputString.split(Pattern.quote(separatorString), -1)).map(JsonString::new).toList());
        } else {
            throw new JsonQueryException("split input and separator must be strings.");
        }
    }

    public static @NotNull JsonString join(@Nullable JsonValue input, @Nullable JsonValue separator) {
        var it = values(input).map(value -> switch (value) {
            case JsonString string -> string;
            case JsonNumber _, JsonBoolean _ -> tostring(value);
            case null, default -> value;
        });

        if (!it.hasNext()) return JsonString.EMPTY;

        var out = JsonString.EMPTY;
        while (it.hasNext()) {
            var next = it.next();
            if (next == null) continue;
            out = (JsonString) add(out, next);
            if (it.hasNext()) out = (JsonString) add(out, separator);
        }
        return Objects.requireNonNull(out);
    }

    public static @NotNull JsonArray explode(@Nullable JsonValue input) {
        return switch (input) {
            case JsonString(var string) -> JsonArray.valueOf(string.codePoints().toArray());
            case null, default -> throw new JsonQueryException("explode input must be a string");
        };
    }

    public static @NotNull JsonString implode(@Nullable JsonValue input) {
        return switch (input) {
            case JsonArray array -> {
                var codepoints = array.stream().mapToInt(value -> switch (value) {
                    case JsonNumber(var number) when isInt(number) && 0 <= number && number <= 0x10FFFF -> (int) number;
                    case null, default -> throw new JsonQueryException("implode input must be an array of unicode codepoints");
                });
                yield toString(codepoints);
            }
            case null, default -> throw new JsonQueryException("implode input must be an array");
        };
    }

    public static @NotNull JsonString asciiUpcase(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> toString(string.codePoints().map(i -> 'a' <= i && i <= 'z' ? i + ('A' - 'a') : i));
            case null, default -> throw new JsonQueryException("explode input must be a string");
        };
    }

    public static @NotNull JsonString asciiDowncase(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> toString(string.codePoints().map(i -> 'A' <= i && i <= 'Z' ? i + ('a' - 'A') : i));
            case null, default -> throw new JsonQueryException("explode input must be a string");
        };
    }

    private static @NotNull JsonString toString(@NotNull IntStream codepoints) {
        return new JsonString(codepoints.collect(StringBuilder::new, StringBuilder::appendCodePoint, StringBuilder::append).toString());
    }

    public static @NotNull JsonNumber utf8ByteLength(@Nullable JsonValue value) {
        return new JsonNumber(switch (value) {
            case JsonString(var string) -> string.codePoints().map(code -> {
                if (code <= 0x7F) return 1;
                if (code <= 0x7FF) return 2;
                if (code <= 0xFFFF) return 3;
                return 4;
            }).sum();
            case null, default -> throw error(value, "only strings have UTF-8 byte length");
        });
    }

    public static @NotNull JsonBoolean startswith(@Nullable JsonValue expression, @Nullable JsonValue prefix) {
        if (expression instanceof JsonString(var string) && prefix instanceof JsonString(var string2)) {
            return JsonBoolean.valueOf(string.startsWith(string2));
        } else {
            throw new JsonQueryException("startswith() requires string inputs");
        }
    }

    public static @NotNull JsonBoolean endswith(@Nullable JsonValue expression, @Nullable JsonValue prefix) {
        if (expression instanceof JsonString(var string) && prefix instanceof JsonString(var string2)) {
            return JsonBoolean.valueOf(string.endsWith(string2));
        } else {
            throw new JsonQueryException("endswith() requires string inputs");
        }
    }
    //</editor-fold>

    //<editor-fold desc="regex" defaultstate="collapsed">
    public static @NotNull Generator<@NotNull JsonValue> match(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        var input = regexInput(context);
        return match(context, args, false).flatMap(Function.identity()).map(result -> match0(input, result));
    }

    private static @NotNull Generator<Generator<@NotNull MatchResult>> match(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args, boolean global
    ) {
        var input = regexInput(context);
        return switch (args.size()) {
            case 1 -> args.getFirst().evaluate(context).map(arg -> switch (arg) {
                case JsonString pattern -> match(input, pattern, null, global);
                case JsonArray array -> switch (array.size()) {
                    case 0 -> throw new JsonQueryException("array not a string or array");
                    case 1 -> match(input, array.getFirst(), null, global);
                    default -> match(input, array.get(0), array.get(1), global);
                };
                case null, default -> throw error(arg, "not a string or array");
            });
            case 2 -> Util.cross(args, context).map(list -> match(input, list.get(0), list.get(1), global));
            default -> throw new IllegalArgumentException();
        };
    }

    private static @NotNull Generator<@NotNull MatchResult> match(
            @NotNull String input, @Nullable JsonValue pattern, @Nullable JsonValue flags, boolean global
    ) {
        if (!(pattern instanceof JsonString(var patternString))) throw error(pattern, "is not a string");

        var skipEmpty = false;
        @MagicConstant(flagsFromClass = Pattern.class)
        var patternFlags = Pattern.MULTILINE;

        switch (flags) {
            case JsonString(var flagsString) -> {
                for (int i = 0; i < flagsString.length(); i++) {
                    switch (flagsString.charAt(i)) {
                        case 'g' -> global = true;
                        case 'i' -> patternFlags |= Pattern.CASE_INSENSITIVE;
                        case 'm' -> patternFlags |= Pattern.DOTALL;
                        case 'n' -> skipEmpty = true;
                        case 'p' -> patternFlags = patternFlags & ~Pattern.MULTILINE | Pattern.DOTALL;
                        case 's' -> patternFlags &= ~Pattern.MULTILINE;
                        case 'l' -> throw new UnsupportedOperationException("unsupported flag l");
                        case 'x' -> patternFlags |= Pattern.COMMENTS;
                        default -> throw new JsonQueryException(flagsString + " is not a valid modifier string.");
                    }
                }
            }
            case null -> {}
            default -> throw error(flags, "is not a string");
        }

        var compiledPattern = Pattern.compile(patternString, patternFlags);
        var out = matchesAsStream(compiledPattern.matcher(input));
        if (skipEmpty) out = out.filter(Predicate.not(result -> result.group().isEmpty()));
        if (!global) out = out.limit(1);
        return out;
    }

    private static @NotNull Generator<@NotNull MatchResult> matchesAsStream(@NotNull Matcher matcher) {
        return new Generator<>() {
            private boolean valid = false;

            @Override
            public boolean hasNext() {
                if (valid) return true;
                return valid = matcher.find();
            }

            @Override
            public MatchResult next() {
                if (!valid) valid = matcher.find();
                if (!valid) throw new NoSuchElementException();
                valid = false;
                return matcher.toMatchResult();
            }
        };
    }

    private static @NotNull JsonValue match0(@NotNull String input, @NotNull MatchResult result) {
        var groupNames = new HashMap<Integer, String>();
        result.namedGroups().forEach((name, index) -> groupNames.put(index, name));

        var groups = new ArrayList<JsonObject>();
        var groupCount = result.groupCount();
        for (int i = 1; i <= groupCount; i++) {
            var start = result.start(i);
            groups.add(JsonObject.of(
                    "offset", new JsonNumber(start == -1 ? -1 : input.codePointCount(0, result.start(i))),
                    "length", new JsonNumber(start == -1 ? 0 : input.codePointCount(result.start(i), result.end(i))),
                    "string", JsonString.valueOf(result.group(i)),
                    "name", JsonString.valueOf(groupNames.get(i))
            ));
        }

        return JsonObject.of(
                "offset", new JsonNumber(input.codePointCount(0, result.start())),
                "length", new JsonNumber(input.codePointCount(result.start(), result.end())),
                "string", new JsonString(result.group()),
                "captures", new JsonArray(groups)
        );
    }

    private static @NotNull String regexInput(@NotNull JQExpression.Context context) {
        if (context.root() instanceof JsonString(var input)) return input;
        throw error(context.root(), "cannot be matched, as it is not a string");
    }

    public static @NotNull Generator<@NotNull JsonValue> test(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return Generator.of(() -> JsonBoolean.valueOf(!isEmpty0(match(context, args, false))));
    }

    public static @NotNull Generator<@NotNull JsonValue> capture(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return match(context, args, false).flatMap(Function.identity()).map(JsonMath::capture0);
    }

    private static @NotNull JsonValue capture0(@NotNull MatchResult result) {
        var out = new LinkedHashMap<String, JsonValue>();
        result.namedGroups().forEach((name, number) -> out.put(name, JsonString.valueOf(result.group(number))));
        return new JsonObject(out);
    }

    public static @NotNull Generator<@NotNull JsonValue> scan(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return match(context, args, true).flatMap(Function.identity()).map(MatchResult::group).map(JsonString::new);
    }

    public static @NotNull Generator<@NotNull JsonValue> split(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return Generator.of(() -> new JsonArray(splits(context, args).toList()));
    }

    public static @NotNull Generator<@NotNull JsonValue> splits(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        if (!(context.root() instanceof JsonString(var inputString))) throw error(context.root(), "cannot be matched, as it is not a string");
        return match(context, args, true).flatMap(it -> new Generator<>() {
            private boolean finished;
            private int offset;

            @Override
            public boolean hasNext() {
                return !finished;
            }

            @Override
            public JsonValue next() throws NoSuchElementException {
                if (it.hasNext()) {
                    var match = it.next();
                    var out = new JsonString(inputString.substring(offset, match.start()));
                    offset = match.end();
                    return out;
                } else if (!finished) {
                    finished = true;
                    return new JsonString(inputString.substring(offset));
                } else {
                    throw new NoSuchElementException();
                }
            }
        });
    }

    public static @NotNull Generator<@NotNull JsonValue> sub(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return sub0(context, args, false);
    }

    public static @NotNull Generator<@NotNull JsonValue> gsub(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args
    ) {
        return sub0(context, args, true);
    }

    private static @NotNull Generator<@NotNull JsonValue> sub0(
            @NotNull JQExpression.Context context, @NotNull List<@NotNull JQExpression> args, boolean global
    ) {
        var input = regexInput(context);
        var matchArgs = switch (args.size()) {
            case 2 -> List.of(args.get(0), new JQConstant(null));
            case 3 -> List.of(args.get(0), args.get(2));
            default -> throw new IllegalArgumentException();
        };
        var subst = args.get(1);

        return match(context, matchArgs, global).flatMap(it -> {
            var offset = 0;
            var fragments = new ArrayList<String>();
            var values = new ArrayList<JQExpression>();

            while (it.hasNext()) {
                var result = it.next();
                var capture = capture0(result);
                fragments.add(input.substring(offset, result.start()));
                values.add(new JQPipeExpression(new JQConstant(capture), subst));
                offset = result.end();
            }
            fragments.add(input.substring(offset));
            return new JQStringInterpolation(null, fragments, values).evaluate(context);
        });
    }
    //</editor-fold>

    //<editor-fold desc="conversions" defaultstate="collapsed">
    public static @NotNull JsonString type(@Nullable JsonValue value) {
        return new JsonString(Util.type(value));
    }

    public static @NotNull JsonString tojson(@Nullable JsonValue value) {
        return new JsonString(JsonValue.toJsonString(value));
    }

    public static @NotNull JsonString tostring(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString string -> string;
            case null, default -> new JsonString(JsonValue.toJsonString(value));
        };
    }

    public static @NotNull JsonNumber tonumber(@Nullable JsonValue value) {
        return switch (value) {
            case JsonNumber number -> number;
            case JsonString(var string) -> {
                try {
                    var number = Double.parseDouble(string);
                    yield new JsonNumber(number);
                } catch (IllegalArgumentException _) {
                    throw error(value, "cannot be parsed as a number");
                }
            }
            case null, default -> throw error(value, "cannot be parsed as a number");
        };
    }

    public static @Nullable JsonValue fromjson(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> {
                try {
                    yield JsonValue.parse(string);
                } catch (JsonParserException ex) {
                    throw new JsonQueryException(ex.getMessage(), ex);
                }
            }
            case null, default -> throw error(value, "only strings can be parsed");
        };
    }
    //</editor-fold>

    //<editor-fold desc="time" defaultstate="collapsed">
    private static final DateTimeFormatter DATE_TIME_FORMATTER = DateTimeFormatter.ofPattern("yyyy-MM-dd'T'HH:mm:ss'Z'")
            .withZone(ZoneId.of("UTC"));

    public static @NotNull JsonNumber fromdateiso8601(@Nullable JsonValue value) {
        return JsonNumber.valueOf(switch (value) {
            case JsonString(var string) -> DATE_TIME_FORMATTER.parse(string, Instant::from).getEpochSecond();
            case null, default -> throw new JsonQueryException("fromdateiso8601/1 requires string inputs");
        });
    }

    public static @NotNull JsonNumber fromdate(@Nullable JsonValue value) {
        return fromdateiso8601(value);
    }

    public static @NotNull JsonString todateiso8601(@Nullable JsonValue value) {
        return new JsonString(switch (value) {
            case JsonNumber(var number) -> DATE_TIME_FORMATTER.format(Instant.ofEpochSecond((long) number));
            case null, default -> throw new JsonQueryException("todateiso8601/1 requires number inputs");
        });
    }

    public static @NotNull JsonString todate(@Nullable JsonValue value) {
        return todateiso8601(value);
    }
    //</editor-fold>

    //<editor-fold desc="misc" defaultstate="collapsed">
    public static @NotNull JsonNumber length(@Nullable JsonValue value) {
        return new JsonNumber(length0(value));
    }

    private static double length0(@Nullable JsonValue value) {
        return switch (value) {
            case JsonString(var string) -> string.length();
            case JsonNumber(double number) -> Math.abs(number);
            case JsonArray array -> array.size();
            case JsonObject object -> object.size();
            case JsonBoolean b -> throw error(b, "has no length");
            case null -> 0;
        };
    }

    public static @NotNull Generator<@Nullable JsonValue> repeat(@NotNull Supplier<@NotNull Generator<@Nullable JsonValue>> delegate) {
        return new Generator<>() {
            private @NotNull Generator<JsonValue> current = Generator.empty();

            @Override
            public boolean hasNext() { return true; }

            @Override
            public JsonValue next() throws NoSuchElementException {
                while (!current.hasNext()) current = delegate.get();
                return current.next();
            }
        };
    }
    //</editor-fold>

    //<editor-fold desc="math library" defaultstate="collapsed">
    private static @Nullable JsonNumber math(@Nullable JsonValue operand, @NotNull DoubleUnaryOperator operator) {
        return switch (operand) {
            case JsonNumber(double number) -> {
                var out = operator.applyAsDouble(number);
                yield Double.isNaN(out) ? null : new JsonNumber(out);
            }
            case null, default -> throw error(operand, "number required");
        };
    }

    private static @Nullable JsonNumber math(@Nullable JsonValue left, @Nullable JsonValue right, @NotNull DoubleBinaryOperator operator) {
        if (!(left instanceof JsonNumber(var number1))) {
            throw error(left, "number required");
        }
        if (!(right instanceof JsonNumber(var number2))) {
            throw error(right, "number required");
        }
        var out = operator.applyAsDouble(number1, number2);
        return Double.isNaN(out) ? null : new JsonNumber(out);
    }

    public static @Nullable JsonNumber acos(@Nullable JsonValue a) {
        return math(a, Math::acos);
    }

    public static @Nullable JsonNumber acosh(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("acosh");
        // return math(a, Math::acosh);
    }

    public static @Nullable JsonNumber asin(@Nullable JsonValue a) {
        return math(a, Math::asin);
    }

    public static @Nullable JsonNumber asinh(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("asinh");
        // return math(a, Math::asinh);
    }

    public static @Nullable JsonNumber atan(@Nullable JsonValue a) {
        return math(a, Math::atan);
    }

    public static @Nullable JsonNumber atanh(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("atanh");
        // return math(a, Math::atanh);
    }

    public static @Nullable JsonNumber cbrt(@Nullable JsonValue a) {
        return math(a, Math::cbrt);
    }

    public static @Nullable JsonNumber ceil(@Nullable JsonValue a) {
        return math(a, Math::ceil);
    }

    public static @Nullable JsonNumber cos(@Nullable JsonValue a) {
        return math(a, Math::cos);
    }

    public static @Nullable JsonNumber cosh(@Nullable JsonValue x) {
        return math(x, Math::cosh);
    }

    public static @Nullable JsonNumber erf(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("erf");
        // return math(a, Math::erf);
    }

    public static @Nullable JsonNumber erfc(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("erfc");
        // return math(a, Math::erfc);
    }

    public static @Nullable JsonNumber exp(@Nullable JsonValue a) {
        return math(a, Math::exp);
    }

    public static @Nullable JsonNumber exp10(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("exp10");
        // return math(a, Math::exp10);
    }

    public static @Nullable JsonNumber exp2(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("exp2");
        // return math(a, Math::exp2);
    }

    public static @Nullable JsonNumber expm1(@Nullable JsonValue x) {
        return math(x, Math::expm1);
    }

    public static @Nullable JsonNumber fabs(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("fabs");
        // return math(a, Math::fabs);
    }

    public static @Nullable JsonNumber floor(@Nullable JsonValue a) {
        return math(a, Math::floor);
    }

    public static @Nullable JsonNumber gamma(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("gamma");
        // return math(a, Math::gamma);
    }

    public static @Nullable JsonNumber j0(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("j0");
        // return math(a, Math::j0);
    }

    public static @Nullable JsonNumber j1(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("j1");
        // return math(a, Math::j1);
    }

    public static @Nullable JsonNumber lgamma(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("lgamma");
        // return math(a, Math::lgamma);
    }

    public static @Nullable JsonNumber log(@Nullable JsonValue a) {
        return math(a, Math::log);
    }

    public static @Nullable JsonNumber log10(@Nullable JsonValue a) {
        return math(a, Math::log10);
    }

    public static @Nullable JsonNumber log1p(@Nullable JsonValue x) {
        return math(x, Math::log1p);
    }

    public static @Nullable JsonNumber log2(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("log2");
        // return math(a, Math::log2);
    }

    public static @Nullable JsonNumber logb(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("logb");
        // return math(a, Math::logb);
    }

    public static @Nullable JsonNumber nearbyint(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("nearbyint");
        // return math(a, Math::nearbyint);
    }

    public static @Nullable JsonNumber rint(@Nullable JsonValue a) {
        return math(a, Math::rint);
    }

    public static @Nullable JsonNumber round(@Nullable JsonValue a) {
        return math(a, Math::round);
    }

    public static @Nullable JsonNumber significand(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("significand");
        // return math(a, Math::significand);
    }

    public static @Nullable JsonNumber sin(@Nullable JsonValue a) {
        return math(a, Math::sin);
    }

    public static @Nullable JsonNumber sinh(@Nullable JsonValue x) {
        return math(x, Math::sinh);
    }

    public static @Nullable JsonNumber sqrt(@Nullable JsonValue a) {
        return math(a, Math::sqrt);
    }

    public static @Nullable JsonNumber tan(@Nullable JsonValue a) {
        return math(a, Math::tan);
    }

    public static @Nullable JsonNumber tanh(@Nullable JsonValue x) {
        return math(x, Math::tanh);
    }

    public static @Nullable JsonNumber tgamma(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("tgamma");
        // return math(a, Math::tgamma);
    }

    public static @Nullable JsonNumber trunc(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("trunc");
        // return math(a, Math::trunc);
    }

    public static @Nullable JsonNumber y0(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("y0");
        // return math(a, Math::y0);
    }

    public static @Nullable JsonNumber y1(@Nullable JsonValue a) {
        throw new UnsupportedOperationException("y1");
        // return math(a, Math::y1);
    }

    public static @Nullable JsonNumber atan2(@Nullable JsonValue y, @Nullable JsonValue x) {
        return math(y, x, Math::atan2);
    }

    public static @Nullable JsonNumber copysign(@Nullable JsonValue magnitude, @Nullable JsonValue sign) {
        return math(magnitude, sign, Math::copySign);
    }

    public static @Nullable JsonNumber drem(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("drem");
        // return math(a, b, Math::drem);
    }

    public static @Nullable JsonNumber fdim(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("fdim");
        // return math(a, b, Math::fdim);
    }

    public static @Nullable JsonNumber fmax(@Nullable JsonValue a, @Nullable JsonValue b) {
        return math(a, b, Math::max);
    }

    public static @Nullable JsonNumber fmin(@Nullable JsonValue a, @Nullable JsonValue b) {
        return math(a, b, Math::min);
    }

    public static @Nullable JsonNumber fmod(@Nullable JsonValue a, @Nullable JsonValue b) {
        return math(a, b, (l, r) -> l % r);
    }

    public static @Nullable JsonNumber frexp(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("frexp");
        // return math(a, b, Math::frexp);
    }

    public static @Nullable JsonNumber hypot(@Nullable JsonValue x, @Nullable JsonValue y) {
        return math(x, y, Math::hypot);
    }

    public static @Nullable JsonNumber jn(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("jn");
        // return math(a, b, Math::jn);
    }

    public static @Nullable JsonNumber ldexp(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("ldexp");
        // return math(a, b, Math::ldexp);
    }

    public static @Nullable JsonNumber modf(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("modf");
        // return math(a, b, Math::modf);
    }

    public static @Nullable JsonNumber nextafter(@Nullable JsonValue start, @Nullable JsonValue direction) {
        return math(start, direction, Math::nextAfter);
    }

    public static @Nullable JsonNumber nexttoward(@Nullable JsonValue start, @Nullable JsonValue direction) {
        return math(start, direction, Math::nextAfter);
    }

    public static @Nullable JsonNumber pow(@Nullable JsonValue a, @Nullable JsonValue b) {
        return math(a, b, Math::pow);
    }

    public static @Nullable JsonNumber remainder(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("remainder");
        // return math(a, b, Math::remainder);
    }

    public static @Nullable JsonNumber scalb(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("scalb");
        // return math(a, b, Math::scalb);
    }

    public static @Nullable JsonNumber scalbln(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("scalbln");
        // return math(a, b, Math::scalbln);
    }

    public static @Nullable JsonNumber yn(@Nullable JsonValue a, @Nullable JsonValue b) {
        throw new UnsupportedOperationException("yn");
        // return math(a, b, Math::yn);
    }

    public static @Nullable JsonNumber fma(@Nullable JsonValue a, @Nullable JsonValue b, @Nullable JsonValue c) {
        if (!(a instanceof JsonNumber(var number1))) {
            throw error(a, "number required");
        }
        if (!(b instanceof JsonNumber(var number2))) {
            throw error(b, "number required");
        }
        if (!(c instanceof JsonNumber(var number3))) {
            throw error(c, "number required");
        }
        var out = Math.fma(number1, number2, number3);
        return Double.isNaN(out) ? null : new JsonNumber(out);
    }
    //</editor-fold>

    //<editor-fold desc="helpers" defaultstate="collapsed">
    private static @NotNull JsonQueryException error(@Nullable JsonValue value, @NotNull String message) {
        return new JsonQueryException(STR."\{Util.type(value)} (\{Util.value(value)}) \{message}.");
    }

    private static @NotNull JsonQueryException error(@Nullable JsonValue first, @Nullable JsonValue second, @NotNull String message) {
        return new JsonQueryException(STR."\{Util.type(first)} (\{Util.value(first)}) and \{Util.type(second)} (\{Util.value(second)}) \{message}.");
    }
    //</editor-fold>

}
