package eu.jonahbauer.json.query.impl;

import org.jetbrains.annotations.NotNull;

import java.util.*;
import java.util.function.*;
import java.util.stream.Stream;

public interface Generator<T> {
    boolean hasNext();
    T next() throws NoSuchElementException;

    default void forEach(@NotNull Consumer<? super T> action) {
        while (hasNext()) {
            action.accept(next());
        }
    }

    default T reduce(T seed, @NotNull BinaryOperator<T> operator) {
        while (hasNext()) {
            seed = operator.apply(seed, next());
        }
        return seed;
    }

    default boolean anyMatch(@NotNull Predicate<? super T> predicate) {
        while (hasNext()) {
            if (predicate.test(next())) return true;
        }
        return false;
    }

    default boolean allMatch(@NotNull Predicate<? super T> predicate) {
        while (hasNext()) {
            if (!predicate.test(next())) return false;
        }
        return true;
    }

    default @NotNull List<T> toList() {
        var out = new ArrayList<T>();
        while (hasNext()) {
            out.add(next());
        }
        return Collections.unmodifiableList(out);
    }

    default @NotNull Generator<T> filter(@NotNull Predicate<? super T> predicate) {
        return new FilterGenerator<>(this, predicate);
    }

    default @NotNull Generator<T> limit(int count) {
        return new LimitGenerator<>(this, count);
    }

    default <S> @NotNull Generator<S> map(@NotNull Function<? super T, ? extends S> function) {
        return new MappingGenerator<>(this, function);
    }

    default <S> @NotNull Generator<S> flatMap(@NotNull Function<? super T, ? extends Generator<? extends S>> function) {
        return new FlatMappingGenerator<>(this, function);
    }

    default <S> @NotNull Generator<S> mapMulti(@NotNull BiConsumer<? super T, ? super Consumer<S>> function) {
        return new MapMultiGenerator<>(this, function);
    }

    @SuppressWarnings("unchecked")
    static <T> @NotNull Generator<T> empty() {
        return (Generator<T>) EmptyGenerator.INSTANCE;
    }

    static <T> @NotNull Generator<T> concat(@NotNull Generator<? extends T> first, @NotNull Generator<? extends T> second) {
        return new ConcatGenerator<>(List.of(first, second));
    }

    static <T> @NotNull Generator<T> of(T value) {
        return new SingletonGenerator<>(value);
    }

    static <T> @NotNull Generator<T> of(@NotNull Supplier<T> value) {
        return new SupplierGenerator<>(value);
    }

    static <T> @NotNull Generator<T> from(@NotNull Iterable<T> value) {
        return new IteratorGenerator<>(value.iterator());
    }

    static <T> @NotNull Generator<T> from(@NotNull Iterator<T> value) {
        return new IteratorGenerator<>(value);
    }

    static <T> @NotNull Generator<T> from(@NotNull Stream<T> value) {
        return new IteratorGenerator<>(value.iterator());
    }
}
