package eu.jonahbauer.json.query.parser;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.query.JsonQueryParserException;
import eu.jonahbauer.json.query.parser.ast.*;
import eu.jonahbauer.json.query.parser.tokenizer.JQToken;
import eu.jonahbauer.json.query.parser.tokenizer.JQTokenKind;
import eu.jonahbauer.json.query.parser.tokenizer.JQTokenizer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.io.Reader;
import java.util.*;
import java.util.function.BinaryOperator;

public class JQParser {
    private final @NotNull JQTokenizer tokenizer;
    private final @NotNull Queue<JQToken> pushback = new LinkedList<>();

    public JQParser(@NotNull String string) {
        this.tokenizer = new JQTokenizer(string);
    }

    public JQParser(@NotNull Reader reader) {
        this.tokenizer = new JQTokenizer(reader);
    }

    public @NotNull JQProgram parseTopLevel() throws IOException {
        var module = parseModule();
        var imports = parseImports();
        var functions = parseFuncDefs();
        var expr = peek() == null ? null : parseExpression();

        return new JQProgram(module, imports, functions, expr);
    }

    private @Nullable JQModule parseModule() throws IOException {
        if (!tryConsume(JQTokenKind.MODULE)) return null;
        var metadata = parseExpression();
        consume(JQTokenKind.SEMICOLON);
        return new JQModule(metadata);
    }

    private @NotNull List<@NotNull JQImport> parseImports() throws IOException {
        var out = new ArrayList<JQImport>();

        while (peek(JQTokenKind.IMPORT) || peek(JQTokenKind.INCLUDE)) {
            out.add(parseImport());
        }

        return out;
    }

    private @NotNull JQImport parseImport() throws IOException {
        var include = tryConsume(JQTokenKind.INCLUDE);
        if (!include) consume(JQTokenKind.IMPORT);

        var path = parseString();

        String as;
        if (include) {
            as = null;
        } else {
            consume(JQTokenKind.AS);
            var dollar = tryConsume(JQTokenKind.DOLLAR);
            var ident = consume(JQTokenKind.IDENT);
            as = (dollar ? "$" : "") + ident.text();
        }

        if (tryConsume(JQTokenKind.SEMICOLON)) {
            return new JQImport(path, as, null);
        } else {
            var metadata = parseExpression();
            consume(JQTokenKind.SEMICOLON);
            return new JQImport(path, as, metadata);
        }
    }

    private @NotNull List<@NotNull JQFunction> parseFuncDefs() throws IOException {
        var out = new ArrayList<JQFunction>();
        while (peek(JQTokenKind.DEF)) {
            out.add(parseFuncDef());
        }
        return out;
    }

    private @NotNull JQFunction parseFuncDef() throws IOException {
        consume(JQTokenKind.DEF);
        var name = consume(JQTokenKind.IDENT).text();

        var params = new ArrayList<String>();
        if (tryConsume(JQTokenKind.LPAREN)) {
            do {
                var dollar = tryConsume(JQTokenKind.DOLLAR);
                var param = consume(JQTokenKind.IDENT).text();
                params.add((dollar ? "$" : "") + param);
            } while (tryConsume(JQTokenKind.SEMICOLON));

            consume(JQTokenKind.RPAREN);
        }

        consume(JQTokenKind.COLON);
        var body = parseExpression();
        consume(JQTokenKind.SEMICOLON);

        return new JQFunction(name, params, body);
    }

    public @NotNull JQExpression parseExpression() throws IOException {
        if (peek(JQTokenKind.DEF)) {
            var function = parseFuncDef();
            var expr = parsePipeExpression();
            return new JQFunctionDefinition(function, expr);
        } else if (peek(JQTokenKind.LABEL)) {

        } else {
            return parsePipeExpression();
        }
        throw new UnsupportedOperationException("not yet implemented");
    }

    private @NotNull JQExpression parseNonAssocBinaryExpression(
            @NotNull Map<@NotNull JQTokenKind, @NotNull BinaryOperator<@NotNull JQExpression>> operations,
            @NotNull Downstream next
    ) throws IOException {
        var first = next.get();

        var operator = peek();
        var operation = operator == null ? null : operations.get(operator.kind());
        if (operation != null) {
            next();
            var second = next.get();
            first = operation.apply(first, second);
        }

        return first;
    }

    private @NotNull JQExpression parseLeftAssocBinaryExpression(
            @NotNull Map<@NotNull JQTokenKind, @NotNull BinaryOperator<@NotNull JQExpression>> operations,
            @NotNull Downstream next
    ) throws IOException {
        var first = next.get();

        while (true) {
            var operator = peek();
            var operation = operator == null ? null : operations.get(operator.kind());
            if (operation == null) break;

            next();
            var second = next.get();
            first = operation.apply(first, second);
        }

        return first;
    }

    private interface Downstream {
        @NotNull
        JQExpression get() throws IOException;
    }

    private @NotNull JQExpression parsePipeExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.PIPE, JQPipeExpression::new
        ), this::parseCommaExpression);
    }

    private @NotNull JQExpression parseCommaExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.COMMA, JQCommaExpression::new
        ), this::parseAlternativeExpression);
    }

    private @NotNull JQExpression parseAlternativeExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.DEFINEDOR, JQAlternativeExpression::new
        ), this::parseAssignment);
    }

    private @NotNull JQExpression parseAssignment() throws IOException {
        return parseNonAssocBinaryExpression(Map.of(
                JQTokenKind.ASSIGN, JQAssignment::new,
                JQTokenKind.SETPLUS, JQAssignment::add,
                JQTokenKind.SETMINUS, JQAssignment::sub,
                JQTokenKind.SETMULT, JQAssignment::mul,
                JQTokenKind.SETDIV, JQAssignment::div,
                JQTokenKind.SETMOD, JQAssignment::mod,
                JQTokenKind.SETPIPE, JQAssignmentPipe::new,
                JQTokenKind.SETDEFINEDOR, JQAssignmentCoerce::new
        ), this::parseBooleanOrExpression);
    }

    private @NotNull JQExpression parseBooleanOrExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.OR, JQBooleanOrExpression::new
        ), this::parseBooleanAndExpression);
    }

    private @NotNull JQExpression parseBooleanAndExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.AND, JQBooleanAndExpression::new
        ), this::parseComparisonExpression);
    }

    private @NotNull JQExpression parseComparisonExpression() throws IOException {
        return parseNonAssocBinaryExpression(Map.of(
                JQTokenKind.EQ, JQBinaryExpression::eq,
                JQTokenKind.NEQ, JQBinaryExpression::neq,
                JQTokenKind.LESS, JQBinaryExpression::lt,
                JQTokenKind.GREATER, JQBinaryExpression::gt,
                JQTokenKind.LESSEQ, JQBinaryExpression::leq,
                JQTokenKind.GREATEREQ, JQBinaryExpression::geq
        ), this::parseAdditiveExpression);
    }

    private @NotNull JQExpression parseAdditiveExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.PLUS, JQBinaryExpression::add,
                JQTokenKind.MINUS, JQBinaryExpression::sub
        ), this::parseMultiplicativeExpression);
    }

    private @NotNull JQExpression parseMultiplicativeExpression() throws IOException {
        return parseLeftAssocBinaryExpression(Map.of(
                JQTokenKind.MULT, JQBinaryExpression::mul,
                JQTokenKind.DIV, JQBinaryExpression::div,
                JQTokenKind.MOD, JQBinaryExpression::mod
        ), this::parseTryCatch);
    }

    private @NotNull JQExpression parseTryCatch() throws IOException {
        if (tryConsume(JQTokenKind.TRY)) {
            var expr = parseNegation();
            var fallback = tryConsume(JQTokenKind.CATCH) ? parseNegation() : null;
            return new JQTryExpression(expr, fallback);
        } else {
            return parseNegation();
        }
    }

    private @NotNull JQExpression parseNegation() throws IOException {
        if (tryConsume(JQTokenKind.MINUS)) {
            var expr = parseErrorSuppression();
            return new JQNegation(expr);
        } else {
            return parseErrorSuppression();
        }
    }

    private @NotNull JQExpression parseErrorSuppression() throws IOException {
        var expression = parseControlFlow();
        if (tryConsume(JQTokenKind.QUESTION_MARK)) {
            return new JQTryExpression(expression);
        } else {
            return expression;
        }
    }

    private @NotNull JQExpression parseControlFlow() throws IOException {
        if (tryConsume(JQTokenKind.REDUCE)) {
            var expr = parseTerm();
            consume(JQTokenKind.AS);
            var patterns = parsePatterns();
            consume(JQTokenKind.LPAREN);
            var init = parseExpression();
            consume(JQTokenKind.SEMICOLON);
            var update = parseExpression();
            consume(JQTokenKind.RPAREN);
            return new JQReduceExpression(expr, patterns, init, update);
        } else if (tryConsume(JQTokenKind.FOREACH)) {
            var expr = parseTerm();
            consume(JQTokenKind.AS);
            var patterns = parsePatterns();
            consume(JQTokenKind.LPAREN);
            var init = parseExpression();
            consume(JQTokenKind.SEMICOLON);
            var update = parseExpression();
            var extract = tryConsume(JQTokenKind.SEMICOLON) ? parseExpression() : new JQRootExpression();
            consume(JQTokenKind.RPAREN);
            return new JQForEachExpression(expr, patterns, init, update, extract);
        } else if (tryConsume(JQTokenKind.IF)) {;
            var conds = new ArrayList<JQExpression>();
            var thens = new ArrayList<JQExpression>();

            do {
                conds.add(parseExpression());
                consume(JQTokenKind.THEN);
                thens.add(parseExpression());
            } while (tryConsume(JQTokenKind.ELSE_IF));

            var otherwise = tryConsume(JQTokenKind.ELSE) ? parseExpression() : null;
            consume(JQTokenKind.END);

            var out = new JQIfExpression(conds.removeLast(), thens.removeLast(), otherwise);
            while (!conds.isEmpty()) {
                out = new JQIfExpression(conds.removeLast(), thens.removeLast(), out);
            }

            return out;
        } else {
            var term = parseTerm();
            if (tryConsume(JQTokenKind.AS)) {
                var patterns = parsePatterns();
                consume(JQTokenKind.PIPE);
                var expr = parseExpression();
                return new JQAsExpression(term, patterns, expr);
            }
            return term;
        }
    }

    private @NotNull JQPatterns parsePatterns() throws IOException {
        var patterns = new ArrayList<JQPatterns.Pattern>();
        patterns.add(parsePattern());
        while (tryConsume(JQTokenKind.ALTERNATION)) {
            patterns.add(parsePattern());
        }
        return new JQPatterns(patterns);
    }

    private @NotNull JQPatterns.Pattern parsePattern() throws IOException {
        if (tryConsume(JQTokenKind.LBRACKET)) {
            var patterns = new ArrayList<JQPatterns.Pattern>();
            do {
                patterns.add(parsePattern());
            } while (tryConsume(JQTokenKind.COMMA));
            consume(JQTokenKind.RBRACKET);
            return new JQPatterns.Pattern.ArrayPattern(patterns);
        } else if (tryConsume(JQTokenKind.LBRACE)) {
            var patterns = new LinkedHashMap<JQExpression, JQPatterns.Pattern>();
            do {
                if (tryConsume(JQTokenKind.DOLLAR)) {
                    var ident = consume(JQTokenKind.IDENT).text();
                    var pattern = new JQPatterns.Pattern.ValuePattern("$" + ident);
                    patterns.put(new JQConstant(JsonString.valueOf(ident)), pattern);
                } else if (peek(JQTokenKind.IDENT)) {
                    var ident = consume(JQTokenKind.IDENT).text();
                    consume(JQTokenKind.COLON);
                    var pattern = parsePattern();
                    patterns.put(new JQConstant(JsonString.valueOf(ident)), pattern);
                } else if (tryConsume(JQTokenKind.LPAREN)) {
                    var expr = parseExpression();
                    consume(JQTokenKind.RPAREN);
                    consume(JQTokenKind.COLON);
                    var pattern = parsePattern();
                    patterns.put(new JQParenthesizedExpression(expr), pattern);
                } else {
                    var key = parseString();
                    consume(JQTokenKind.COLON);
                    var pattern = parsePattern();
                    patterns.put(key, pattern);
                }
            } while (tryConsume(JQTokenKind.COMMA));
            consume(JQTokenKind.RBRACE);
            return new JQPatterns.Pattern.ObjectPattern(patterns);
        } else {
            consume(JQTokenKind.DOLLAR);
            var ident = consume(JQTokenKind.IDENT).text();
            return new JQPatterns.Pattern.ValuePattern("$" + ident);
        }
    }

    private @NotNull JQExpression parseTerm() throws IOException {
        var next = peek();
        var term = switch (next == null ? null : next.kind()) {
            case NUMBER -> new JQConstant(new JsonNumber(Objects.requireNonNull(next().nval())));
            case LPAREN -> parseParenthesizedExpression();
            case LBRACKET -> parseArrayConstructionExpression();
            case LBRACE -> parseObjectConstructionExpression();
            case QQSTRING_START -> parseString();
            case FORMAT -> throw new UnsupportedOperationException("not yet implemented");
            case BREAK -> {
                consume(JQTokenKind.IDENT);
                consume(JQTokenKind.DOLLAR);
                var label = consume(JQTokenKind.IDENT).text();
                throw new UnsupportedOperationException("not yet implemented");
            }
            case IDENT -> switch (Objects.requireNonNull(peek()).text()) {
                case "null" -> {
                    consume(JQTokenKind.IDENT);
                    yield new JQConstant(null);
                }
                case "true" -> {
                    consume(JQTokenKind.IDENT);
                    yield new JQConstant(JsonBoolean.TRUE);
                }
                case "false" -> {
                    consume(JQTokenKind.IDENT);
                    yield new JQConstant(JsonBoolean.FALSE);
                }
                default -> parseFunctionInvocation();
            };
            case LOC -> {
                var loc = consume(JQTokenKind.LOC);
                yield new JQLocExpression("<top-level>", loc.line());
            }
            case DOLLAR -> parseVariableExpression();
            case FIELD -> {
                var name = consume(JQTokenKind.FIELD).text().substring(1);
                var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                yield new JQIndexExpression(new JQRootExpression(), name, optional);
            }
            case DOT -> {
                consume(JQTokenKind.DOT);
                if (peek(JQTokenKind.QQSTRING_START)) {
                    var name = parseString();
                    var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                    yield new JQIndexExpression(new JQRootExpression(), name, optional);
                } else {
                    yield new JQRootExpression();
                }
            }
            case REC -> {
                consume(JQTokenKind.REC);
                yield new JQRecursionExpression();
            }
            case null -> throw new JsonQueryParserException(-1, -1, "unexpected end of file");
            default -> {
                var token = Objects.requireNonNull(peek());
                throw new JsonQueryParserException(token.line(), token.column(), "unexpected token " + token.kind());
            }
        };

        while (peek(JQTokenKind.LBRACKET) || peek(JQTokenKind.DOT) || peek(JQTokenKind.FIELD)) {
            term = parseIndexingExpression(term);
        }

        return term;
    }

    private @NotNull JQExpression parseFunctionInvocation() throws IOException {
        var name = consume(JQTokenKind.IDENT).text();
        var args = new ArrayList<JQExpression>();
        if (tryConsume(JQTokenKind.LPAREN)) {
            do {
                args.add(parseExpression());
            } while (tryConsume(JQTokenKind.SEMICOLON));
            consume(JQTokenKind.RPAREN);
        }
        return new JQFunctionInvocation(name, args);
    }

    private @NotNull JQExpression parseVariableExpression() throws IOException {
        consume(JQTokenKind.DOLLAR);
        var ident = consume(JQTokenKind.IDENT).text();
        return new JQVariableExpression("$" + ident);
    }

    private @NotNull JQExpression parseParenthesizedExpression() throws IOException {
        consume(JQTokenKind.LPAREN);
        var expression = parseExpression();
        consume(JQTokenKind.RPAREN);
        return new JQParenthesizedExpression(expression);
    }

    private @NotNull JQExpression parseArrayConstructionExpression() throws IOException {
        consume(JQTokenKind.LBRACKET);
        if (tryConsume(JQTokenKind.RBRACKET)) {
            return new JQConstant(JsonArray.EMPTY);
        } else {
            var expression = parseExpression();
            consume(JQTokenKind.RBRACKET);
            return new JQArrayConstructionExpression(expression);
        }
    }

    private @NotNull JQExpression parseObjectConstructionExpression() throws IOException {
        consume(JQTokenKind.LBRACE);
        if (tryConsume(JQTokenKind.RBRACE)) {
            return new JQConstant(JsonObject.EMPTY);
        } else {
            throw new UnsupportedOperationException("not yet implemented");
        }
    }

    private @NotNull JQExpression parseIndexingExpression(@NotNull JQExpression root) throws IOException {
        if (tryConsume(JQTokenKind.LBRACKET)) {
            if (tryConsume(JQTokenKind.RBRACKET)) {
                var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                return new JQIterateExpression(root, optional);
            } else if (tryConsume(JQTokenKind.COLON)) {
                var end = parseExpression();
                consume(JQTokenKind.RBRACKET);
                var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                return new JQSliceExpression(root, null, end, optional);
            } else {
                var expr = parseExpression();
                if (tryConsume(JQTokenKind.COLON)) {
                    var end = peek(JQTokenKind.RBRACKET) ? null : parseExpression();
                    consume(JQTokenKind.RBRACKET);
                    var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                    return new JQSliceExpression(root, expr, end, optional);
                } else {
                    consume(JQTokenKind.RBRACKET);
                    var optional = tryConsume(JQTokenKind.QUESTION_MARK);
                    return new JQIndexExpression(root, expr, optional);
                }
            }
        } else if (peek(JQTokenKind.FIELD)) {
            var name = consume(JQTokenKind.FIELD).text().substring(1);
            var optional = tryConsume(JQTokenKind.QUESTION_MARK);
            return new JQIndexExpression(root, name, optional);
        } else {
            consume(JQTokenKind.DOT);
            var name = parseString();
            var optional = tryConsume(JQTokenKind.QUESTION_MARK);
            return new JQIndexExpression(root, name, optional);
        }
    }

    private @NotNull JQStringInterpolation parseString() throws IOException {
        var format = peek(JQTokenKind.FORMAT) ? next().text() : null;

        var current = new StringBuilder();
        var fragments = new ArrayList<String>();
        var values = new ArrayList<JQExpression>();

        consume(JQTokenKind.QQSTRING_START);

        current.setLength(0);
        while (peek(JQTokenKind.QQSTRING_TEXT)) {
            current.append(next().sval());
        }
        fragments.add(current.toString());

        while (tryConsume(JQTokenKind.QQSTRING_INTERP_START)) {
            values.add(parseExpression());
            consume(JQTokenKind.QQSTRING_INTERP_END);

            current.setLength(0);
            while (peek(JQTokenKind.QQSTRING_TEXT)) {
                current.append(next().sval());
            }
            fragments.add(current.toString());
        }

        consume(JQTokenKind.QQSTRING_END);
        return new JQStringInterpolation(format, fragments, values);
    }






    private void consume(@NotNull String ident) throws IOException {
        var token = next();
        if (token.kind() != JQTokenKind.IDENT || !Objects.equals(ident, token.text())) {
            throw new JsonQueryParserException(0, 0, "");
        }
    }

    private @NotNull JQToken consume(@NotNull JQTokenKind kind) throws IOException {
        var token = next();
        if (token.kind() != kind) throw new JsonQueryParserException(token.line(), token.column(), "unexpected " + token.kind() + " expected " + kind);
        return token;
    }

    private boolean tryConsume(@NotNull String ident) throws IOException {
        var out = peek(ident);
        if (out) next();
        return out;
    }

    private boolean tryConsume(@NotNull JQTokenKind kind) throws IOException {
        var out = peek(kind);
        if (out) next();
        return out;
    }

    private boolean peek(@NotNull String ident) throws IOException {
        var token = peek();
        return token != null && token.kind() == JQTokenKind.IDENT && Objects.equals(ident, token.text());
    }

    private boolean peek(@NotNull JQTokenKind kind) throws IOException {
        var token = peek();
        return token != null && token.kind() == kind;
    }

    private @Nullable JQToken peek() throws IOException {
        if (pushback.isEmpty()) {
            var next = tokenizer.next();
            pushback(next);
            return next;
        } else {
            return pushback.peek();
        }
    }

    private @NotNull JQToken next() throws IOException {
        if (pushback.isEmpty()) {
            var next = tokenizer.next();
            if (next == null) throw new JsonQueryParserException(0, 0, "unexpected $end");
            return next;
        } else {
            return pushback.remove();
        }
    }

    private void pushback(@Nullable JQToken token) {
        pushback.add(token);
    }
}
