package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.*;
import eu.jonahbauer.json.query.JsonMath;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.impl.Generator;
import eu.jonahbauer.json.query.parser.JQParser;
import eu.jonahbauer.json.query.parser.ast.JQExpression.Context;
import eu.jonahbauer.json.query.util.Util;
import lombok.SneakyThrows;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.util.*;
import java.util.function.*;
import java.util.stream.Collectors;

public enum JQBuiltIn implements JQInvocable {
    ABS(0, Implementation.map$V(JsonMath::abs)),
    ADD(0, (context, _) -> context.stream().map(value -> JsonMath.values(value).reduce(null, JsonMath::add))),
    NOT(0, Implementation.map$V(JsonMath::not)),

    // error handling
    ERROR$0(0, (context, _) -> context.stream().flatMap(JsonMath::error)),
    ERROR$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(JsonMath::error)),
    HALT(0, (_, _) -> Generator.of((JsonValue) null).flatMap(_ -> JsonMath.halt())),
    HALT_ERROR$0(0, (context, _) -> context.stream().flatMap(JsonMath::halt)),
    HALT_ERROR$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(value -> JsonMath.halt(context.root(), value))),

    // stream operations
    EMPTY(0, (_, _) -> Generator.empty()),
    RANGE$1(1, (context, args) -> args.getFirst().evaluate(context).flatMap(JsonMath::range)),
    RANGE$2(2, (context, args) -> Util.cross(args, context).flatMap(bounds -> JsonMath.range(bounds.get(0), bounds.get(1)))),
    RANGE$3(3, (context, args) -> Util.cross(args, context).flatMap(bounds -> JsonMath.range(bounds.get(0), bounds.get(1), bounds.get(2)))),
    LIMIT(2, (context, args) -> args.getFirst().evaluate(context).flatMap(limit -> JsonMath.limit(args.getLast().evaluate(context), limit))),
    FIRST$1(1, (context, args) -> JsonMath.first(args.getFirst().evaluate(context))),
    LAST$1(1, (context, args) -> JsonMath.last(args.getFirst().evaluate(context))),
    NTH$2(2, (context, args) -> args.getFirst().evaluate(context).flatMap(n -> JsonMath.nth(args.getLast().evaluate(context), n))),
    SELECT(1, Implementation.mapF$F(JsonMath::select)),

    // iterable operations
    MAP(1, Implementation.mapF$V(JsonMath::map)),
    MAP_VALUES(1, Implementation.mapF$V(JsonMath::mapValues)),
    KEYS(0, Implementation.map$V(JsonMath::keys)),
    KEYS_UNSORTED(0, Implementation.map$V(JsonMath::keysUnsorted)),
    HAS(1, Implementation.mapV$V(JsonMath::has)),
    IN(1, Implementation.mapV$V(JsonMath::in)),
    FIRST$0(0, Implementation.map$V(JsonMath::first)),
    LAST$0(0, Implementation.map$V(JsonMath::last)),
    NTH$1(1, Implementation.mapV$V(JsonMath::index)),
    ANY$0(0, Implementation.map$V(JsonMath::any)),
    ANY$1(1, Implementation.mapF$V(JsonMath::any)),
    ANY$2(2, Implementation.mapFF$V((root, gen, filter) -> JsonMath.any(gen.apply(root), filter))),
    ALL$0(0, Implementation.map$V(JsonMath::all)),
    ALL$1(1, Implementation.mapF$V(JsonMath::all)),
    ALL$2(2, Implementation.mapFF$V((root, gen, filter) -> JsonMath.all(gen.apply(root), filter))),
    FLATTEN$0(0, Implementation.map$V(JsonMath::flatten)),
    FLATTEN$1(1, Implementation.mapV$V(JsonMath::flatten)),
    SORT(0, Implementation.map$V(JsonMath::sort)),
    SORT_BY(1, Implementation.mapF$V(JsonMath::sort)),
    MIN(0, Implementation.map$V(JsonMath::min)),
    MIN_BY(1, Implementation.mapF$V(JsonMath::min)),
    MAX(0, Implementation.map$V(JsonMath::max)),
    MAX_BY(1, Implementation.mapF$V(JsonMath::max)),
    UNIQUE(0, Implementation.map$V(JsonMath::unique)),
    UNIQUE_BY(1, Implementation.mapF$V(JsonMath::unique)),
    GROUP_BY(1, Implementation.mapF$V(JsonMath::group)),
    REVERSE(0, Implementation.map$V(JsonMath::reverse)),
    CONTAINS(1, Implementation.mapV$V(JsonMath::contains)),
    INDICES(1, Implementation.mapV$V(JsonMath::indices)),
    INDEX(1, Implementation.mapV$V(JsonMath::firstindex)),
    RINDEX(1, Implementation.mapV$V(JsonMath::lastindex)),
    INSIDE(1, Implementation.mapV$V(JsonMath::inside)),
    COMBINATIONS$0(0, Implementation.map$F(JsonMath::combinations)),
    COMBINATIONS$1(1, Implementation.mapV$F(JsonMath::combinations)),
    BSEARCH(1, Implementation.mapV$V(JsonMath::bsearch)),
    TRANSPOSE(0, Implementation.map$V(JsonMath::transpose)),
    TO_ENTRIES(0, Implementation.map$V(JsonMath::toEntries)),
    FROM_ENTRIES(0, Implementation.map$V(JsonMath::fromEntries)),
    WITH_ENTRIES(1, Implementation.mapF$V(JsonMath::withEntries)),

    // filters
    ARRAYS(0, Implementation.filter(JsonMath::isArray0)),
    OBJECTS(0, Implementation.filter(JsonMath::isObject0)),
    ITERABLES(0, Implementation.filter(JsonMath::isIterable0)),
    BOOLEANS(0, Implementation.filter(JsonMath::isBoolean0)),
    NUMBERS(0, Implementation.filter(JsonMath::isNumber0)),
    NORMALS(0, Implementation.filter(JsonMath::isNormal0)),
    FINITES(0, Implementation.filter(JsonMath::isFinite0)),
    STRINGS(0, Implementation.filter(JsonMath::isString0)),
    NULLS(0, Implementation.filter(JsonMath::isNull0)),
    VALUES(0, Implementation.filter(JsonMath::isValue0)),
    SCALARS(0, Implementation.filter(JsonMath::isScalar0)),

    // checks
    ISINFINITE(0, Implementation.map$V(JsonMath::isInfinite)),
    ISNAN(0, Implementation.map$V(JsonMath::isNan)),
    ISFINITE(0, Implementation.map$V(JsonMath::isFinite)),
    ISNORMAL(0, Implementation.map$V(JsonMath::isNormal)),
    ISEMPTY(1, (context, args) -> Generator.of(() -> JsonMath.isEmpty(args.getFirst().evaluate(context)))),

    // string operations
    TRIM(0, Implementation.map$V(JsonMath::trim)),
    LTRIM(0, Implementation.map$V(JsonMath::ltrim)),
    RTRIM(0, Implementation.map$V(JsonMath::rtrim)),
    LTRIMSTR(1, Implementation.mapV$V(JsonMath::ltrimstr)),
    RTRIMSTR(1, Implementation.mapV$V(JsonMath::rtrimstr)),
    SPLIT$1(1, Implementation.mapV$V(JsonMath::split)),
    JOIN(1, Implementation.mapV$V(JsonMath::join)),
    IMPLODE(0, Implementation.map$V(JsonMath::implode)),
    EXPLODE(0, Implementation.map$V(JsonMath::explode)),
    ASCII_UPCASE(0, Implementation.map$V(JsonMath::asciiUpcase)),
    ASCII_DOWNCASE(0, Implementation.map$V(JsonMath::asciiDowncase)),
    UTF8BYTELENGTH(0, Implementation.map$V(JsonMath::utf8ByteLength)),
    STARTSWITH(1, Implementation.mapV$V(JsonMath::startswith)),
    ENDSWITH(1, Implementation.mapV$V(JsonMath::endswith)),

    // regex
    TEST$1(1, JsonMath::test),
    TEST$2(2, JsonMath::test),
    MATCH$1(1, JsonMath::match),
    MATCH$2(2, JsonMath::match),
    CAPTURE$1(1, JsonMath::capture),
    CAPTURE$2(2, JsonMath::capture),
    SCAN$1(1, JsonMath::scan),
    SCAN$2(2, JsonMath::scan),
    SPLIT$2(2, JsonMath::split),
    SPLITS$1(1, JsonMath::splits),
    SPLITS$2(2, JsonMath::splits),
    SUB$2(2, JsonMath::sub),
    SUB$3(3, JsonMath::sub),
    GSUB$2(2, JsonMath::gsub),
    GSUB$3(3, JsonMath::gsub),

    // conversions
    TYPE(0, Implementation.map$V(JsonMath::type)),
    TOJSON(0, Implementation.map$V(JsonMath::tojson)),
    TOSTRING(0, Implementation.map$V(JsonMath::tostring)),
    TONUMBER(0, Implementation.map$V(JsonMath::tonumber)),
    FROMJSON(0, Implementation.map$V(JsonMath::fromjson)),

    // time
    FROMDATEISO8601(0, Implementation.map$V(JsonMath::fromdateiso8601)),
    FROMDATE(0, Implementation.map$V(JsonMath::fromdate)),
    TODATEISO8601(0, Implementation.map$V(JsonMath::todateiso8601)),
    TODATE(0, Implementation.map$V(JsonMath::todate)),
    NOW(0, Implementation.map$V(_ -> JsonNumber.valueOf(System.currentTimeMillis()))),

    // misc
    LENGTH(0, Implementation.map$V(JsonMath::length)),
    REPEAT(1, (context, args) -> JsonMath.repeat(() -> args.getFirst().evaluate(context))),

    // math library
    ACOS(0, Implementation.map$V(JsonMath::acos)),
    ACOSH(0, Implementation.map$V(JsonMath::acosh)),
    ASIN(0, Implementation.map$V(JsonMath::asin)),
    ASINH(0, Implementation.map$V(JsonMath::asinh)),
    ATAN(0, Implementation.map$V(JsonMath::atan)),
    ATANH(0, Implementation.map$V(JsonMath::atanh)),
    CBRT(0, Implementation.map$V(JsonMath::cbrt)),
    CEIL(0, Implementation.map$V(JsonMath::ceil)),
    COS(0, Implementation.map$V(JsonMath::cos)),
    COSH(0, Implementation.map$V(JsonMath::cosh)),
    ERF(0, Implementation.map$V(JsonMath::erf)),
    ERFC(0, Implementation.map$V(JsonMath::erfc)),
    EXP(0, Implementation.map$V(JsonMath::exp)),
    EXP10(0, Implementation.map$V(JsonMath::exp10)),
    EXP2(0, Implementation.map$V(JsonMath::exp2)),
    EXPM1(0, Implementation.map$V(JsonMath::expm1)),
    FABS(0, Implementation.map$V(JsonMath::fabs)),
    FLOOR(0, Implementation.map$V(JsonMath::floor)),
    GAMMA(0, Implementation.map$V(JsonMath::gamma)),
    J0(0, Implementation.map$V(JsonMath::j0)),
    J1(0, Implementation.map$V(JsonMath::j1)),
    LGAMMA(0, Implementation.map$V(JsonMath::lgamma)),
    LOG(0, Implementation.map$V(JsonMath::log)),
    LOG10(0, Implementation.map$V(JsonMath::log10)),
    LOG1P(0, Implementation.map$V(JsonMath::log1p)),
    LOG2(0, Implementation.map$V(JsonMath::log2)),
    LOGB(0, Implementation.map$V(JsonMath::logb)),
    NEARBYINT(0, Implementation.map$V(JsonMath::nearbyint)),
    RINT(0, Implementation.map$V(JsonMath::rint)),
    ROUND(0, Implementation.map$V(JsonMath::round)),
    SIGNIFICAND(0, Implementation.map$V(JsonMath::significand)),
    SIN(0, Implementation.map$V(JsonMath::sin)),
    SINH(0, Implementation.map$V(JsonMath::sinh)),
    SQRT(0, Implementation.map$V(JsonMath::sqrt)),
    TAN(0, Implementation.map$V(JsonMath::tan)),
    TANH(0, Implementation.map$V(JsonMath::tanh)),
    TGAMMA(0, Implementation.map$V(JsonMath::tgamma)),
    TRUNC(0, Implementation.map$V(JsonMath::trunc)),
    Y0(0, Implementation.map$V(JsonMath::y0)),
    Y1(0, Implementation.map$V(JsonMath::y1)),
    ATAN2(2, Implementation.mapVV$V((_, a, b) -> JsonMath.atan2(a, b))),
    COPYSIGN(2, Implementation.mapVV$V((_, a, b) -> JsonMath.copysign(a, b))),
    DREM(2, Implementation.mapVV$V((_, a, b) -> JsonMath.drem(a, b))),
    FDIM(2, Implementation.mapVV$V((_, a, b) -> JsonMath.fdim(a, b))),
    FMAX(2, Implementation.mapVV$V((_, a, b) -> JsonMath.fmax(a, b))),
    FMIN(2, Implementation.mapVV$V((_, a, b) -> JsonMath.fmin(a, b))),
    FMOD(2, Implementation.mapVV$V((_, a, b) -> JsonMath.fmod(a, b))),
    FREXP(2, Implementation.mapVV$V((_, a, b) -> JsonMath.frexp(a, b))),
    HYPOT(2, Implementation.mapVV$V((_, a, b) -> JsonMath.hypot(a, b))),
    JN(2, Implementation.mapVV$V((_, a, b) -> JsonMath.jn(a, b))),
    LDEXP(2, Implementation.mapVV$V((_, a, b) -> JsonMath.ldexp(a, b))),
    MODF(2, Implementation.mapVV$V((_, a, b) -> JsonMath.modf(a, b))),
    NEXTAFTER(2, Implementation.mapVV$V((_, a, b) -> JsonMath.nextafter(a, b))),
    NEXTTOWARD(2, Implementation.mapVV$V((_, a, b) -> JsonMath.nexttoward(a, b))),
    POW(2, Implementation.mapVV$V((_, a, b) -> JsonMath.pow(a, b))),
    REMAINDER(2, Implementation.mapVV$V((_, a, b) -> JsonMath.remainder(a, b))),
    SCALB(2, Implementation.mapVV$V((_, a, b) -> JsonMath.scalb(a, b))),
    SCALBLN(2, Implementation.mapVV$V((_, a, b) -> JsonMath.scalbln(a, b))),
    YN(2, Implementation.mapVV$V((_, a, b) -> JsonMath.yn(a, b))),
    FMA(3, (context, args) -> Util.cross(args, context).map(values -> JsonMath.fma(values.get(0), values.get(1), values.get(2))))
    ;

    public static final @NotNull Map<@NotNull String, @NotNull JQInvocable> ALL_BUILTINS = Arrays.stream(JQBuiltIn.values())
            .collect(Collectors.toMap(
                    JQInvocable::reference,
                    Function.identity()
            ));

    private final @NotNull JQInvocable implementation;

    JQBuiltIn(@NotNull String body) {
        this(List.of(), parse(body));
    }

    JQBuiltIn(@NotNull String arg0, @NotNull String body) {
        this(List.of(arg0), parse(body));
    }

    JQBuiltIn(@NotNull String arg0, @NotNull String arg1, @NotNull String body) {
        this(List.of(arg0, arg1), parse(body));
    }

    JQBuiltIn(@NotNull String arg0, @NotNull String arg1, @NotNull String arg2, @NotNull String body) {
        this(List.of(arg0, arg1, arg2), parse(body));
    }

    JQBuiltIn(@NotNull List<@NotNull String> params, @NotNull JQExpression body) {
        this.implementation = new JQFunction(getIdentifier(name()), params, body);
    }
    
    JQBuiltIn(int arity, @NotNull Implementation implementation) {
        this.implementation = new InvocableDelegate(getIdentifier(name()), arity, implementation);
    }

    @SneakyThrows(IOException.class)
    private static @NotNull JQExpression parse(@NotNull String expression) {
        return new JQParser(expression).parseExpression();
    }

    private static @NotNull String getIdentifier(@NotNull String name) {
        var identifier = name.toLowerCase(Locale.ROOT);
        var idx = identifier.lastIndexOf("$");
        if (idx != -1) identifier = identifier.substring(0, idx);
        return identifier;
    }

    @Override
    public @NotNull Generator<@Nullable JsonValue> invoke(@NotNull Context context, @NotNull List<@NotNull JQExpression> args) {
        return implementation.invoke(context, args);
    }

    @Override
    public @NotNull String identifier() {
        return implementation.identifier();
    }

    @Override
    public int arity() {
        return implementation.arity();
    }
    
    private record InvocableDelegate(
            @NotNull String identifier, int arity,
            @NotNull Implementation delegate
    ) implements JQInvocable {
        
        @Override
        public @NotNull Generator<@Nullable JsonValue> invoke(@NotNull Context context, @NotNull List<@NotNull JQExpression> args) { 
            if (args.size() != arity) throw new JsonQueryException("invalid argument count");
            return delegate.invoke(context, args);
        }
    }

    @FunctionalInterface
    private interface Implementation {
        @NotNull Generator<@Nullable JsonValue> invoke(@NotNull Context context, @NotNull List<@NotNull JQExpression> args);

        // format: map<args>$<return>
        // where <args> and <return> use V for value, F for filter

        /**
         * Produces a zero-arg implementation applying the function to the current value.
         */
        static @NotNull Implementation map$F(@NotNull Function<? super @Nullable JsonValue, ? extends @NotNull Generator<? extends @Nullable JsonValue>> operator) {
            return (context, _) -> context.stream().flatMap(operator);
        }

        /**
         * Produces a zero-arg implementation applying the function to the current value.
         */
        static @NotNull Implementation map$V(@NotNull Function<? super @Nullable JsonValue, ? extends @Nullable JsonValue> operator) {
            return (context, _) -> context.stream().map(operator);
        }

        /**
         * Produces a one-arg implementation applying the function to the current value and the first argument
         * (passed by value).
         */
        static @NotNull Implementation mapV$V(@NotNull BiFunction<? super @Nullable JsonValue, ? super @Nullable JsonValue, ? extends @Nullable JsonValue> function) {
            return (context, args) -> {
                var arg = args.getFirst();
                return context.stream().flatMap(root -> arg.evaluate(context).map(value -> function.apply(root, value)));
            };
        }

        /**
         * Produces a one-arg implementation applying the function to the current value and the first argument
         * (passed by value).
         */
        static @NotNull Implementation mapV$F(@NotNull BiFunction<? super @Nullable JsonValue, ? super @Nullable JsonValue, ? extends @NotNull Generator<? extends @Nullable JsonValue>> function) {
            return (context, args) -> {
                var arg = args.getFirst();
                return context.stream().flatMap(root -> arg.evaluate(context).flatMap(value -> function.apply(root, value)));
            };
        }

        /**
         * Produces a one-arg implementation applying the function to the current value and the first argument
         * (passed as filter).
         */
        static @NotNull Implementation mapF$F(@NotNull BiFunction<? super @Nullable JsonValue, ? super @NotNull JQFilter, ? extends @NotNull Generator<? extends @Nullable JsonValue>> function) {
            return (context, args) -> {
                var arg = args.getFirst().bind(context);
                return context.stream().flatMap(root -> function.apply(root, arg));
            };
        }

        /**
         * Produces a one-arg implementation applying the function to the current value and the first argument
         * (passed as filter).
         */
        static @NotNull Implementation mapF$V(@NotNull BiFunction<? super @Nullable JsonValue, ? super @NotNull JQFilter, ? extends @Nullable JsonValue> function) {
            return (context, args) -> {
                var arg = args.getFirst().bind(context);
                return context.stream().map(root -> function.apply(root, arg));
            };
        }

        static @NotNull Implementation mapVV$V(@NotNull TriFunction<? super @Nullable JsonValue, ? super @Nullable JsonValue, ? super @Nullable JsonValue, ? extends @Nullable JsonValue> function) {
            return (context, args) -> {
                var arg0 = args.get(0);
                var arg1 = args.get(1);
                return context.stream().flatMap(root -> arg0.evaluate(context).flatMap(value0 -> arg1.evaluate(context).map(value1 -> function.apply(root, value0, value1))));
            };
        }

        static @NotNull Implementation mapFF$V(@NotNull TriFunction<? super @Nullable JsonValue, ? super @NotNull JQFilter, ? super @NotNull JQFilter, ? extends @Nullable JsonValue> function) {
            return (context, args) -> {
                var arg0 = args.get(0).bind(context);
                var arg1 = args.get(1).bind(context);
                return context.stream().map(root -> function.apply(root, arg0, arg1));
            };
        }

        static @NotNull Implementation filter(@NotNull Predicate<@Nullable JsonValue> predicate) {
            return (context, _) -> context.stream().filter(predicate);
        }
    }

    @FunctionalInterface
    private interface TriFunction<S, T, U, R> {
        R apply(S s, T t, U u);
    }
}
