package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.impl.Generator;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.stream.Collectors;

public interface JQExpression {

    @NotNull Generator<@Nullable JsonValue> evaluate(@NotNull Context context);

    boolean isConstant();

    @NotNull String toString();

    default @NotNull JQFilter bind(@NotNull Context context) {
        return value -> evaluate(context.withRoot(value));
    }

    record Context(
            @Nullable JsonValue root,
            @NotNull Map<@NotNull String, @Nullable JsonValue> variables,
            @NotNull Map<@NotNull String, @NotNull JQInvocable> functions
    ) {

        public Context(@Nullable JsonValue root) {
            this(root, Map.of(), JQBuiltIn.ALL_BUILTINS);
        }

        public Context {
            variables = variables.entrySet().stream().collect(Collectors.toMap(
                    entry -> {
                        Objects.requireNonNull(entry.getKey());
                        if (!entry.getKey().startsWith("$")) throw new IllegalArgumentException();
                        return entry.getKey();
                    },
                    Map.Entry::getValue
            ));
            functions = Map.copyOf(functions);
        }

        public @NotNull JQInvocable function(@NotNull String name, int arity) {
            var out = functions.get(name + "/" + arity);
            if (out == null) throw new JsonQueryException(name + "/" + arity + " is not defined.");
            return out;
        }

        public @Nullable JsonValue variable(@NotNull String name) {
            if (!variables.containsKey(name)) throw new JsonQueryException(name + " is not defined.");
            return variables.get(name);
        }

        public @NotNull Context withRoot(@Nullable JsonValue root) {
            return new Context(root, variables, functions);
        }

        public @NotNull Context withFunction(@NotNull JQInvocable function) {
            var f = new HashMap<>(functions);
            f.put(function.reference(), function);
            return new Context(root, variables, f);
        }

        public @NotNull Context withFunctions(@NotNull List<? extends @NotNull JQInvocable> functions) {
            var f = new HashMap<>(this.functions);
            functions.forEach(func -> f.put(func.reference(), func));
            return new Context(root, variables, f);
        }

        public @NotNull Context withVariable(@NotNull String name, @Nullable JsonValue variable) {
            var v = new HashMap<>(variables);
            v.put(name, variable);
            return new Context(root, v, functions);
        }

        public @NotNull Context withVariables(@NotNull Map<@NotNull String, @Nullable JsonValue> variables) {
            var v = new HashMap<>(this.variables);
            v.putAll(variables);
            return new Context(root, v, functions);
        }

        public @NotNull Generator<@Nullable JsonValue> stream() {
            return Generator.of(root());
        }
    }
}
