package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonMath;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.impl.Generator;
import eu.jonahbauer.json.query.util.Util;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Supplier;
import java.util.stream.Collectors;

public record JQPatterns(@NotNull List<@NotNull Pattern> patterns) {

    public JQPatterns {
        patterns = List.copyOf(patterns);
        if (patterns.isEmpty()) throw new IllegalArgumentException();
    }

    public @NotNull Generator<@NotNull Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull JQExpression.Context context, @Nullable JsonValue v) {
        var variables = new HashMap<String, JsonValue>();
        Generator.from(patterns).map(Pattern::variables).flatMap(Generator::from).forEach(key -> variables.put(key, null));

        var result = new Generator<@NotNull Map<@NotNull String, @Nullable JsonValue>>() {
            private final Generator<@NotNull Generator<@NotNull Map<@NotNull String, @Nullable JsonValue>>> source
                    = Generator.from(patterns).map(pattern -> pattern.bind(context, v));
            private @NotNull Generator<@NotNull Map<@NotNull String, @Nullable JsonValue>> current = Generator.of(() -> { throw new JsonQueryException("empty"); });

            private boolean hasValue;
            private @Nullable Map<@NotNull String, @Nullable JsonValue> value;

            private boolean advance() {
                if (hasValue) return true;
                while (true) {
                    try {
                        hasValue = current.hasNext();
                        value = hasValue ? current.next() : null;
                        return hasValue;
                    } catch (JsonQueryException _) {
                        // switch to next pattern on error
                        if (source.hasNext()) {
                            current = source.next();
                        } else {
                            value = null;
                            hasValue = false;
                            return false;
                        }
                    }
                }
            }

            @Override
            public boolean hasNext() {
                return advance();
            }

            @Override
            public @NotNull Map<@NotNull String, @Nullable JsonValue> next() {
                if (!advance()) throw new NoSuchElementException();
                assert value != null;

                var out = value;
                value = null;
                hasValue = false;
                return out;
            }
        };

        return result.map(vars -> {
            var out = new HashMap<>(variables);
            out.putAll(vars);
            return out;
        });
    }

    @Override
    public @NotNull String toString() {
        return patterns.stream().map(Objects::toString).collect(Collectors.joining(" ?// "));
    }

    public sealed interface Pattern {
        @NotNull
        Set<@NotNull String> variables();
        @NotNull
        Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull JQExpression.Context context, @Nullable JsonValue value);

        record ValuePattern(@NotNull String name) implements Pattern {
            public ValuePattern {
                Objects.requireNonNull(name);
                if (!name.startsWith("$")) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                return Set.of(name);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull JQExpression.Context context, @Nullable JsonValue value) {
                var map = new HashMap<String, JsonValue>();
                map.put(name, value);
                return Generator.of(Collections.unmodifiableMap(map));
            }

            @Override
            public @NotNull String toString() {
                return name;
            }
        }

        record ArrayPattern(@NotNull List<@NotNull Pattern> patterns) implements Pattern {
            public ArrayPattern {
                patterns = List.copyOf(patterns);
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull JQExpression.Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Generator<Map<String, JsonValue>>>>();

                for (int i = 0; i < patterns.size(); i++) {
                    var k = new JsonNumber(i);
                    var v = JsonMath.index(value, k);

                    var pattern = patterns.get(i);
                    streams.add(() -> pattern.bind(context, v));
                }

                return Util.crossReversed(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                return patterns.stream().map(Objects::toString).collect(Collectors.joining(", ", "[", "]"));
            }
        }

        record ObjectPattern(@NotNull SequencedMap<@NotNull JQExpression, @NotNull Pattern> patterns) implements Pattern {

            public ObjectPattern {
                patterns = Collections.unmodifiableSequencedMap(new LinkedHashMap<>(patterns));
                if (patterns.isEmpty()) throw new IllegalArgumentException();
            }

            @Override
            public @NotNull Set<@NotNull String> variables() {
                var out = new HashSet<String>();
                patterns.values().forEach(p -> out.addAll(p.variables()));
                return Collections.unmodifiableSet(out);
            }

            @Override
            public @NotNull Generator<Map<@NotNull String, @Nullable JsonValue>> bind(@NotNull JQExpression.Context context, @Nullable JsonValue value) {
                var streams = new ArrayList<Supplier<Generator<Map<String, JsonValue>>>>();

                this.patterns.reversed().forEach((keyExpression, pattern) -> streams.add(() -> {
                    var keyStream = keyExpression.evaluate(context);
                    return keyStream.flatMap(key -> pattern.bind(context, JsonMath.index(value, key)));
                }));

                return Util.cross(streams).map(list -> {
                    var map = new HashMap<String, JsonValue>();
                    list.reversed().forEach(map::putAll);
                    return map;
                });
            }

            @Override
            public @NotNull String toString() {
                var out = new StringJoiner(", ", "{", "}");
                patterns.forEach((key, pattern) -> out.add(key + ": " + pattern));
                return out.toString();
            }
        }
    }
}
