package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonArray;
import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonString;
import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.impl.Generator;
import eu.jonahbauer.json.query.util.Util;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public record JQSliceExpression(@NotNull JQExpression expression, @Nullable JQExpression start, @Nullable JQExpression end, boolean optional) implements JQExpression {
    public JQSliceExpression {
        Objects.requireNonNull(expression);
        if (start == null && end == null) throw new IllegalArgumentException();
    }

    @Override
    public @NotNull Generator<@Nullable JsonValue> evaluate(@NotNull Context context) {
        return expression.evaluate(context).flatMap(value -> switch (value) {
            case JsonArray array -> slice(context, "an array slice", array.size(), array::subList);
            case JsonString string -> slice(context, "a string slice", string.length(), string::subSequence);
            case null -> Generator.of((JsonValue) null);
            default -> {
                if (optional) yield Generator.empty();
                throw new JsonQueryException(STR."Cannot index \{Util.type(value)} with object.");
            }
        });
    }

    private @NotNull Generator<@Nullable JsonValue> slice(@NotNull Context context, @NotNull String type, int length, @NotNull BiFunction<Integer, Integer, JsonValue> slice) {
        return getIndices(start, context, type, length, 0)
                .map(start -> getIndices(end, context, type, length, length)
                        .map(end -> start > end ? slice.apply(0, 0) : slice.apply(start, end))
                )
                .flatMap(Function.identity());
    }

    private @NotNull Generator<Integer> getIndices(@Nullable JQExpression expression, @NotNull Context context, @NotNull String type, int length, int fallback) {
        if (expression == null) return Generator.of(fallback);
        return expression.evaluate(context).map(value -> getIndex(value, type, length));
    }

    private int getIndex(@Nullable JsonValue value, @NotNull String type, int length) {
        if (!(value instanceof JsonNumber(double d))) {
            throw new JsonQueryException(STR."Start and end indices of \{type} must be numbers.");
        }
        var i = (int) Math.floor(d);
        return i < 0 ? Math.max(0, length + i) : Math.min(length, i);
    }

    @Override
    public boolean isConstant() {
        return expression.isConstant() && (start == null || start.isConstant()) && (end == null || end.isConstant());
    }

    @Override
    public @NotNull String toString() {
        return expression + "[" + (start == null ? "" : start) + ":" + (end == null ? "" : end) + "]";
    }
}
