package eu.jonahbauer.json.query.parser.ast;

import eu.jonahbauer.json.JsonString;
import eu.jonahbauer.json.JsonValue;
import eu.jonahbauer.json.query.JsonQueryException;
import eu.jonahbauer.json.query.JsonQueryHaltException;
import eu.jonahbauer.json.query.impl.Generator;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.function.Function;

public record JQTryExpression(@NotNull JQExpression expression, @Nullable JQExpression fallback) implements JQExpression {
    public JQTryExpression {
        Objects.requireNonNull(expression);
    }

    public JQTryExpression(@NotNull JQExpression expression) {
        this(expression, null);
    }

    @Override
    public @NotNull Generator<@Nullable JsonValue> evaluate(@NotNull Context context) {
        return new QuietIterator(
                expression.evaluate(context),
                fallback == null ? null : ex -> fallback.evaluate(context.withRoot(new JsonString(ex.getMessage())))
        );
    }

    @Override
    public boolean isConstant() {
        return expression.isConstant() && (fallback == null || fallback.isConstant());
    }

    @Override
    public @NotNull String toString() {
        return "try " + expression + (fallback == null ? "" : " catch " + fallback);
    }

    private static class QuietIterator implements Generator<@Nullable JsonValue> {
        private @Nullable Generator<@Nullable JsonValue> delegate;
        private @Nullable Generator<@Nullable JsonValue> fallback;
        private @Nullable Function<@NotNull JsonQueryException, @NotNull Generator<@Nullable JsonValue>> fallbackSupplier;

        private JsonValue value;
        private boolean hasValue;

        private QuietIterator(
                @NotNull Generator<@Nullable JsonValue> delegate,
                @Nullable Function<@NotNull JsonQueryException, @NotNull Generator<@Nullable JsonValue>> fallback
        ) {
            this.delegate = delegate;
            this.fallbackSupplier = fallback;
        }

        private boolean advance() {
            if (hasValue) return true;
            assert fallback == null && delegate != null;
            try {
                hasValue = delegate.hasNext();
                value = hasValue ? delegate.next() : null;
                return hasValue;
            } catch (JsonQueryHaltException ex) {
                throw ex;
            } catch (JsonQueryException ex) {
                delegate = null;
                if (fallbackSupplier != null) {
                    fallback = fallbackSupplier.apply(ex);
                    hasValue = false;
                    value = null;
                    return fallback.hasNext();
                } else {
                    hasValue = false;
                    value = null;
                    return false;
                }
            }
        }

        @Override
        public boolean hasNext() {
            if (fallback != null) return fallback.hasNext();
            if (delegate == null) return false;
            return advance();
        }

        @Override
        public @Nullable JsonValue next() {
            if (fallback != null) return fallback.next();
            if (delegate == null) throw new NoSuchElementException();
            if (!advance()) throw new NoSuchElementException();
            var out = value;
            hasValue = false;
            value = null;
            return out;
        }
    }
}
