package eu.jonahbauer.json.query.parser.tokenizer;

import eu.jonahbauer.json.query.JsonQueryTokenizerException;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.*;
import java.util.*;
import java.util.function.IntPredicate;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public final class JQTokenizer implements Iterable<JQToken> {
    private final TrackingReader reader;

    private final Queue<State> stack = Collections.asLifoQueue(new ArrayDeque<>());

    public JQTokenizer(@NotNull String string) {
        this.reader = new TrackingReader(new StringReader(string));
    }

    public JQTokenizer(@NotNull Reader reader) {
        this.reader = new TrackingReader(new BufferedReader(reader));
    }

    public @Nullable JQToken next() throws IOException {
        int chr;
        int line;
        int column;

        if (stack.peek() == State.IN_QQSTRING) {
            line = reader.getLineNumber();
            column = reader.getColumnNumber();
            chr = reader.read();
            if (chr == -1) throw new JsonQueryTokenizerException(line, column, "unexpected $end");

            var text = new StringBuilder().append((char) chr);
            if (chr == '"') {
                stack.remove(State.IN_QQSTRING);
                return new JQToken(JQTokenKind.QQSTRING_END, text.toString(), line, column);
            } else if (chr == '\\') {
                int chr2 = reader.read();
                if (chr2 != -1) text.append((char) chr2);

                return switch (chr2) {
                    case '(' -> {
                        stack.add(State.IN_QQINTERP);
                        yield new JQToken(JQTokenKind.QQSTRING_INTERP_START, text.toString(), line, column);
                    }
                    case 'u' -> {
                        if (tryRead(text, this::isHexDigit, 4) != 4) {
                            throw new JsonQueryTokenizerException(line, column, "invalid \\uXXXX escape");
                        }
                        var code = Integer.parseInt(text.substring(2), 16);
                        yield new JQToken(JQTokenKind.QQSTRING_TEXT, text.toString(), String.valueOf((char) code), line, column);
                    }
                    case -1 -> throw new JsonQueryTokenizerException(line, column, "invalid character");
                    default -> new JQToken(JQTokenKind.QQSTRING_TEXT, text.toString(), switch (chr2) {
                        case '"' -> "\"";
                        case '\\' -> "\\";
                        case '/' -> "/";
                        case 'b' -> "\b";
                        case 'f' -> "\f";
                        case 'n' -> "\n";
                        case 'r' -> "\r";
                        case 't' -> "\t";
                        default -> throw new JsonQueryTokenizerException(line, column, "invalid escape");
                    }, line, column);
                };
            } else {
                tryRead(text, c -> c != '\\' && c != '"');
                return new JQToken(JQTokenKind.QQSTRING_TEXT, text.toString(), text.toString(), line, column);
            }
        }

        while (true) {
            line = reader.getLineNumber();
            column = reader.getColumnNumber();
            chr = reader.read();

            // rules producing no token
            if (isWhitespace(chr)) continue;
            if (chr == '#') {
                readEndOfLineComment();
                continue;
            }

            break;
        }

        // EOF
        if (chr == -1) return null;

        var text = new StringBuilder().append((char) chr);
        assert chr >= 0;
        var result = switch ((Character) (char) chr) {
            case Character c when c == '!' && tryRead(text, "=") -> JQTokenKind.NEQ;
            case Character c when c == '=' && tryRead(text, "=") -> JQTokenKind.EQ;
            case Character c when c == '/' && tryRead(text, "/") -> JQTokenKind.DEFINEDOR;
            case Character c when c == '|' && tryRead(text, "=") -> JQTokenKind.SETPIPE;
            case Character c when c == '+' && tryRead(text, "=") -> JQTokenKind.SETPLUS;
            case Character c when c == '-' && tryRead(text, "=") -> JQTokenKind.SETMINUS;
            case Character c when c == '*' && tryRead(text, "=") -> JQTokenKind.SETMULT;
            case Character c when c == '/' && tryRead(text, "=") -> JQTokenKind.SETDIV;
            case Character c when c == '%' && tryRead(text, "=") -> JQTokenKind.SETMOD;
            case Character c when c == '/' && tryRead(text, "/=") -> JQTokenKind.SETDEFINEDOR;
            case Character c when c == '<' && tryRead(text, "=") -> JQTokenKind.LESSEQ;
            case Character c when c == '>' && tryRead(text, "=") -> JQTokenKind.GREATEREQ;
            case Character c when c == '.' && tryRead(text, ".") -> JQTokenKind.REC;
            case Character c when c == '?' && tryRead(text, "//") -> JQTokenKind.ALTERNATION;
            case Character c when c == '@' && tryRead(text, this::isIdentifierPart) -> JQTokenKind.FORMAT;
            case Character c when c == '$' && tryRead(text, "__loc__") -> JQTokenKind.LOC;
            case '?' -> JQTokenKind.QUESTION_MARK;
            case '=' -> JQTokenKind.ASSIGN;
            case ';' -> JQTokenKind.SEMICOLON;
            case ',' -> JQTokenKind.COMMA;
            case ':' -> JQTokenKind.COLON;
            case '|' -> JQTokenKind.PIPE;
            case '+' -> JQTokenKind.PLUS;
            case '-' -> JQTokenKind.MINUS;
            case '*' -> JQTokenKind.MULT;
            case '/' -> JQTokenKind.DIV;
            case '%' -> JQTokenKind.MOD;
            case '$' -> JQTokenKind.DOLLAR;
            case '<' -> JQTokenKind.LESS;
            case '>' -> JQTokenKind.GREATER;

            case '[' -> {
                stack.add(State.IN_BRACKET);
                yield JQTokenKind.LBRACKET;
            }
            case '(' -> {
                stack.add(State.IN_PAREN);
                yield JQTokenKind.LPAREN;
            }
            case '{' -> {
                stack.add(State.IN_BRACE);
                yield JQTokenKind.LBRACE;
            }
            case Character c when c == ']' && stack.peek() == State.IN_BRACKET -> {
                stack.remove();
                yield JQTokenKind.RBRACKET;
            }
            case Character c when c == ')' && stack.peek() == State.IN_PAREN -> {
                stack.remove();
                yield JQTokenKind.RPAREN;
            }
            case Character c when c == ')' && stack.peek() == State.IN_QQINTERP -> {
                stack.remove();
                yield JQTokenKind.QQSTRING_INTERP_END;
            }
            case Character c when c == '}' && stack.peek() == State.IN_BRACE -> {
                stack.remove();
                yield JQTokenKind.RBRACE;
            }

            case '"' -> {
                stack.add(State.IN_QQSTRING);
                yield JQTokenKind.QQSTRING_START;
            }

            case Character c when c == '.' && tryRead(text, this::isDigit) -> {
                readExponential(text);
                yield new JQToken(JQTokenKind.NUMBER, text.toString(), Double.parseDouble(text.toString()), line, column);
            }
            case Character c when isDigit(c) -> {
                readDigits(text);
                if (tryRead(text, ".")) readDigits(text);
                readExponential(text);
                yield new JQToken(JQTokenKind.NUMBER, text.toString(), Double.parseDouble(text.toString()), line, column);
            }
            case Character c when c == '.' && tryRead(text, this::isIdentifierStart) -> {
                tryRead(text, this::isIdentifierPart);
                yield JQTokenKind.FIELD;
            }
            case '.' -> JQTokenKind.DOT;
            case Character c when isIdentifierStart(c) -> {
                tryRead(text, this::isIdentifierPart);
                while (true) {
                    reader.mark(3);
                    if (reader.read() != ':') { reader.reset(); break; }
                    if (reader.read() != ':') { reader.reset(); break; }
                    chr = reader.read();
                    if (!isIdentifierStart(chr)) { reader.reset(); break; }
                    text.append("::").append((char) chr);
                    tryRead(text, this::isIdentifierPart);
                }
                yield switch (text.toString()) {
                    case "as" -> JQTokenKind.AS;
                    case "import" -> JQTokenKind.IMPORT;
                    case "include" -> JQTokenKind.INCLUDE;
                    case "module" -> JQTokenKind.MODULE;
                    case "def" -> JQTokenKind.DEF;
                    case "if" -> JQTokenKind.IF;
                    case "then" -> JQTokenKind.THEN;
                    case "else" -> JQTokenKind.ELSE;
                    case "elif" -> JQTokenKind.ELSE_IF;
                    case "and" -> JQTokenKind.AND;
                    case "or" -> JQTokenKind.OR;
                    case "end" -> JQTokenKind.END;
                    case "reduce" -> JQTokenKind.REDUCE;
                    case "foreach" -> JQTokenKind.FOREACH;
                    case "try" -> JQTokenKind.TRY;
                    case "catch" -> JQTokenKind.CATCH;
                    case "label" -> JQTokenKind.LABEL;
                    case "break" -> JQTokenKind.BREAK;
                    default -> JQTokenKind.IDENT;
                };
            }
            default -> throw new JsonQueryTokenizerException(line, column, "invalid character");
        };

        if (result instanceof JQToken token) {
            return token;
        } else {
            return new JQToken((JQTokenKind) result, text.toString(), line, column);
        }
    }

    private void readEndOfLineComment() throws IOException {
        int chr;
        do {
            chr = reader.read();
        } while (chr != -1 && chr != '\r' && chr != '\n');
    }

    private void readDigits(@NotNull StringBuilder text) throws IOException {
        tryRead(text, this::isDigit);
    }

    private void readExponential(@NotNull StringBuilder text) throws IOException {
        reader.mark(3);

        // [eE]
        int e = reader.read();
        if (e != 'e' && e != 'E') {
            reader.reset();
            return;
        }

        // [+-]?
        int sign = reader.read();
        int digit;
        if (sign == '+' || sign == '-') {
            digit = reader.read();
        } else {
            digit = sign;
            sign = -1;
        }

        if (!isDigit(digit)) {
            reader.reset();
            return;
        }

        text.append((char) e);
        if (sign != -1) text.append((char) sign);
        text.append((char) digit);
        readDigits(text);
    }

    private boolean tryRead(@NotNull StringBuilder text, @NotNull String expected) throws IOException {
        int length = expected.length();
        reader.mark(length);
        for (int i = 0; i < length; i++) {
            if (reader.read() != expected.charAt(i)) {
                reader.reset();
                return false;
            }
        }

        text.append(expected);
        return true;
    }

    private boolean tryRead(@NotNull StringBuilder text, @NotNull IntPredicate predicate) throws IOException {
        return tryRead(text, predicate, Integer.MAX_VALUE) != 0;
    }

    private int tryRead(@NotNull StringBuilder text, @NotNull IntPredicate predicate, int limit) throws IOException {
        int i = 0;
        for (; i < limit; i++) {
            reader.mark(1);
            int chr = reader.read();
            if (chr == -1 || !predicate.test(chr)) {
                reader.reset();
                break;
            }
            text.append((char) chr);
        }
        return i;
    }

    private boolean isWhitespace(int chr) {
        return chr == ' ' || chr == '\t' || chr == '\r' || chr == '\n';
    }

    private boolean isDigit(int chr) {
        return '0' <= chr && chr <= '9';
    }

    private boolean isHexDigit(int chr) {
        return '0' <= chr && chr <= '9' || 'a' <= chr && chr <= 'f' || 'A' <= chr && chr <= 'F';
    }

    private boolean isIdentifierStart(int chr) {
        return 'a' <= chr && chr <= 'z' || 'A' <= chr && chr <= 'Z' || chr == '_';
    }

    private boolean isIdentifierPart(int chr) {
        return isIdentifierStart(chr) || isDigit(chr);
    }

    public @NotNull Iterator<JQToken> iterator() {
        return new JQTokenizerIterator();
    }

    public @NotNull Stream<@NotNull JQToken> stream() {
        return StreamSupport.stream(this.spliterator(), false);
    }

    private class JQTokenizerIterator implements Iterator<JQToken> {
        private JQToken next;
        private boolean valid = false;

        private void ensureValid() {
            try {
                if (!valid) {
                    next = JQTokenizer.this.next();
                    valid = true;
                }
            } catch (IOException ex) {
                throw new UncheckedIOException(ex);
            }
        }

        @Override
        public boolean hasNext() {
            ensureValid();
            return next != null;
        }

        @Override
        public @NotNull JQToken next() {
            ensureValid();
            if (next == null) {
                throw new NoSuchElementException();
            } else {
                valid = false;
                return next;
            }
        }
    }

    private enum State {
        IN_PAREN,
        IN_BRACKET,
        IN_BRACE,
        IN_QQSTRING,
        IN_QQINTERP,
        ;
    }

    @RequiredArgsConstructor
    private static class TrackingReader {
        private final Reader delegate;

        private boolean skipLF;
        @Getter
        private int lineNumber = 1;
        @Getter
        private int columnNumber = 1;

        private boolean markedSkipLF;
        private int markedLineNumber;
        private int markedColumnNumber;

        public int read() throws IOException {
            int c = delegate.read();

            // handle line feed
            if (skipLF) {
                skipLF = false;
            } else if (c == '\n') {
                lineNumber++;
                columnNumber = 1;
            }

            // handle carriage return
            if (c == '\r') {
                lineNumber++;
                columnNumber = 1;
                skipLF = true;
            }

            if (c != '\n' && c != '\r') {
                columnNumber++;
            }

            return c;
        }

        public void mark(int readAheadLimit) throws IOException {
            delegate.mark(readAheadLimit);
            markedSkipLF = skipLF;
            markedLineNumber = lineNumber;
            markedColumnNumber = columnNumber;
        }

        public void reset() throws IOException {
            delegate.reset();
            skipLF = markedSkipLF;
            lineNumber = markedLineNumber;
            columnNumber = markedColumnNumber;
        }
    }
}
