package eu.jonahbauer.json.query;

import eu.jonahbauer.json.JsonArray;
import eu.jonahbauer.json.JsonBoolean;
import eu.jonahbauer.json.JsonNumber;
import eu.jonahbauer.json.JsonValue;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;

class JsonMathTest {

    @ParameterizedTest
    @MethodSource("addArguments")
    void add(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.add(first, second));
    }

    static Stream<Arguments> addArguments() {
        return arguments(
                Arguments.of("foo", "bar", "foobar"),
                Arguments.of(7, 1, 8),
                Arguments.of(List.of(1, 2), List.of(3, 4), List.of(1, 2, 3, 4)),
                Arguments.of(1, null, 1),
                Arguments.of(Map.of("a", 1), Map.of("b", 1), Map.of("a", 1, "b", 1)),
                Arguments.of(Map.of("a", 1), Map.of("a", 42), Map.of("a", 42))
        );
    }

    @ParameterizedTest
    @MethodSource("subArguments")
    void sub(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.sub(first, second));
    }

    static Stream<Arguments> subArguments() {
        return arguments(
                Arguments.of(4, 1, 3),
                Arguments.of(List.of("xml", "yaml", "json"), List.of("xml", "yaml"), List.of("json"))
        );
    }

    @ParameterizedTest
    @MethodSource("mulArguments")
    void mul(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.mul(first, second));
    }

    static Stream<Arguments> mulArguments() {
        return arguments(
                Arguments.of("foo", 0, ""),
                Arguments.of(0, "foo", ""),
                Arguments.of(1.5, "foo", "foofoo"),
                Arguments.of("foo", 1.5, "foofoo"),
                Arguments.of(-1, "foo", null),
                Arguments.of("foo", -1, null),
                Arguments.of("foo", 1E300, null),
                Arguments.of(1E300, "foo", null),
                Arguments.of(2, 3, 6),
                Arguments.of(
                        Map.of("k", Map.of("a", 1, "b", 2)),
                        Map.of("k", Map.of("a", 0, "c", 3)),
                        Map.of("k", Map.of("a", 0, "b", 2, "c", 3))
                )
        );
    }

    @ParameterizedTest
    @MethodSource("divArguments")
    void div(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.div(first, second));
    }

    static Stream<Arguments> divArguments() {
        return arguments(
                Arguments.of(10, 5, 2),
                Arguments.of("a, b,c,d, e", ", ", List.of("a", "b,c,d", "e"))
        );
    }

    @ParameterizedTest
    @MethodSource("eqArguments")
    void eq(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.eq(first, second));
    }

    static Stream<Arguments> eqArguments() {
        return arguments(
                Arguments.of(null, false, false),
                Arguments.of(1, 1.0, true),
                Arguments.of(1, "1", false),
                Arguments.of(1, "banana", false)
        );
    }

    @ParameterizedTest
    @MethodSource("sortArguments")
    void sort(@NotNull JsonArray value, @NotNull JsonArray expected) {
        assertEquals(expected, JsonMath.sort(value));
    }

    static Stream<Arguments> sortArguments() {
        return arguments(
                Arguments.of(Arrays.asList(8, 3, null, 6, true, false), Arrays.asList(null, false, true, 3, 6, 8)),
                Arguments.of(List.of(List.of("a", "b", "c"), List.of("a", "b")), List.of(List.of("a", "b"), List.of("a", "b", "c"))),
                Arguments.of(List.of("foo", "bar", "baz"), List.of("bar", "baz", "foo")),
                Arguments.of(List.of(Map.of("a", 1), Map.of("b", 1)), List.of(Map.of("a", 1), Map.of("b", 1))),
                Arguments.of(List.of(Map.of("b", 1), Map.of("a", 1)), List.of(Map.of("a", 1), Map.of("b", 1))),
                Arguments.of(List.of(Map.of("a", 1), Map.of("a", 2)), List.of(Map.of("a", 1), Map.of("a", 2))),
                Arguments.of(List.of(Map.of("a", 2), Map.of("a", 1)), List.of(Map.of("a", 1), Map.of("a", 2))),
                Arguments.of(List.of(Map.of("a", 2), Map.of("a", 1, "c", 1)), List.of(Map.of("a", 2), Map.of("a", 1, "c", 1))),
                Arguments.of(List.of(Map.of("a", 1, "c", 1), Map.of("a", 2)), List.of(Map.of("a", 2), Map.of("a", 1, "c", 1)))
        );
    }

    @Test
    void lt() {
        assertEquals(JsonBoolean.FALSE, JsonMath.lt(JsonNumber.valueOf(1), JsonNumber.valueOf(1)));
        assertEquals(JsonBoolean.TRUE, JsonMath.lt(JsonNumber.valueOf(1), JsonNumber.valueOf(2)));
        assertEquals(JsonBoolean.FALSE, JsonMath.lt(JsonNumber.valueOf(2), JsonNumber.valueOf(1)));
    }

    @Test
    void gt() {
        assertEquals(JsonBoolean.FALSE, JsonMath.gt(JsonNumber.valueOf(1), JsonNumber.valueOf(1)));
        assertEquals(JsonBoolean.FALSE, JsonMath.gt(JsonNumber.valueOf(1), JsonNumber.valueOf(2)));
        assertEquals(JsonBoolean.TRUE, JsonMath.gt(JsonNumber.valueOf(2), JsonNumber.valueOf(1)));
    }

    @Test
    void leq() {
        assertEquals(JsonBoolean.TRUE, JsonMath.leq(JsonNumber.valueOf(1), JsonNumber.valueOf(1)));
        assertEquals(JsonBoolean.TRUE, JsonMath.leq(JsonNumber.valueOf(1), JsonNumber.valueOf(2)));
        assertEquals(JsonBoolean.FALSE, JsonMath.leq(JsonNumber.valueOf(2), JsonNumber.valueOf(1)));
    }

    @Test
    void geq() {
        assertEquals(JsonBoolean.TRUE, JsonMath.geq(JsonNumber.valueOf(1), JsonNumber.valueOf(1)));
        assertEquals(JsonBoolean.FALSE, JsonMath.geq(JsonNumber.valueOf(1), JsonNumber.valueOf(2)));
        assertEquals(JsonBoolean.TRUE, JsonMath.geq(JsonNumber.valueOf(2), JsonNumber.valueOf(1)));
    }

    @ParameterizedTest
    @MethodSource("andArguments")
    void and(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.and(first, second));
    }

    static Stream<Arguments> andArguments() {
        return arguments(
                Arguments.of(42, "a string", true),
                Arguments.of(true, true, true),
                Arguments.of(true, false, false)
        );
    }
    
    @ParameterizedTest
    @MethodSource("orArguments")
    void or(@Nullable JsonValue first, @Nullable JsonValue second, @Nullable JsonValue expected) {
        assertEquals(expected, JsonMath.or(first, second));
    }

    static Stream<Arguments> orArguments() {
        return arguments(
                Arguments.of(false, "a string", true),
                Arguments.of(false, null, false),
                Arguments.of(1, false, true)
        );
    }

    static Stream<Arguments> arguments(@NotNull Arguments @NotNull ... arguments) {
        return Stream.of(arguments).map(args -> Arguments.of(Arrays.stream(args.get()).map(JsonValue::valueOf).toArray()));
    }

}