package eu.jonahbauer.json.query;

import eu.jonahbauer.json.JsonValue;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.junit.jupiter.api.Assumptions;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.opentest4j.AssertionFailedError;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;

class JsonQueryTest {

    @ParameterizedTest
    @MethodSource("arguments")
    void examples(@NotNull String jq, @Nullable JsonValue input, @NotNull List<@Nullable JsonValue> output) {
        JsonQuery query;
        try {
            query = JsonQuery.parse(jq);
        } catch (UnsupportedOperationException _) {
            throw Assumptions.<RuntimeException>abort("not yet implemented");
        } catch (Throwable t) {
            throw new AssertionFailedError("Unexpected exception thrown: " + t.getMessage(), t);
        }
        var actual = query.run(input).toList();
        assertEquals(output, actual);
    }

    /**
     * Examples taken from the manual
     *
     * @see <a href="https://jqlang.github.io/jq/manual/">jq Manual (development version)</a>
     */
    @SuppressWarnings("ArraysAsListWithZeroOrOneArgument")
    static Stream<Arguments> arguments() {
        return Stream.of(
                Arguments.of(
                        ".",
                        JsonValue.parse("\"Hello, world!\""),
                        Arrays.asList(JsonValue.parse("\"Hello, world!\""))
                ),
                Arguments.of(
                        ".",
                        JsonValue.parse("0.12345678901234567890123456789"),
                        Arrays.asList(JsonValue.parse("0.12345678901234567890123456789"))
                ),
                Arguments.of(
                        "[., tojson]",
                        JsonValue.parse("12345678909876543212345"),
                        Arrays.asList(JsonValue.parse("[12345678909876543212345,\"12345678909876543212345\"]"))
                ),
                Arguments.of(
                        ". < 0.12345678901234567890123456788",
                        JsonValue.parse("0.12345678901234567890123456789"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "map([., . == 1]) | tojson",
                        JsonValue.parse("[1, 1.000, 1.0, 100e-2]"),
                        Arrays.asList(JsonValue.parse("\"[[1,true],[1.000,true],[1.0,true],[1.00,true]]\""))
                ),
                Arguments.of(
                        ". as $big | [$big, $big + 1] | map(. > 10000000000000000000000000000000)",
                        JsonValue.parse("10000000000000000000000000000001"),
                        Arrays.asList(JsonValue.parse("[true, false]"))
                ),
                Arguments.of(
                        ".foo",
                        JsonValue.parse("{\"foo\": 42, \"bar\": \"less interesting data\"}"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        ".foo",
                        JsonValue.parse("{\"notfoo\": true, \"alsonotfoo\": false}"),
                        Arrays.asList(JsonValue.parse("null"))
                ),
                Arguments.of(
                        ".[\"foo\"]",
                        JsonValue.parse("{\"foo\": 42}"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        ".foo?",
                        JsonValue.parse("{\"foo\": 42, \"bar\": \"less interesting data\"}"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        ".foo?",
                        JsonValue.parse("{\"notfoo\": true, \"alsonotfoo\": false}"),
                        Arrays.asList(JsonValue.parse("null"))
                ),
                Arguments.of(
                        ".[\"foo\"]?",
                        JsonValue.parse("{\"foo\": 42}"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        "[.foo?]",
                        JsonValue.parse("[1,2]"),
                        Arrays.asList(JsonValue.parse("[]"))
                ),
                Arguments.of(
                        ".[0]",
                        JsonValue.parse("[{\"name\":\"JSON\", \"good\":true}, {\"name\":\"XML\", \"good\":false}]"),
                        Arrays.asList(JsonValue.parse("{\"name\":\"JSON\", \"good\":true}"))
                ),
                Arguments.of(
                        ".[2]",
                        JsonValue.parse("[{\"name\":\"JSON\", \"good\":true}, {\"name\":\"XML\", \"good\":false}]"),
                        Arrays.asList(JsonValue.parse("null"))
                ),
                Arguments.of(
                        ".[-2]",
                        JsonValue.parse("[1,2,3]"),
                        Arrays.asList(JsonValue.parse("2"))
                ),
                Arguments.of(
                        ".[2:4]",
                        JsonValue.parse("[\"a\",\"b\",\"c\",\"d\",\"e\"]"),
                        Arrays.asList(JsonValue.parse("[\"c\", \"d\"]"))
                ),
                Arguments.of(
                        ".[2:4]",
                        JsonValue.parse("\"abcdefghi\""),
                        Arrays.asList(JsonValue.parse("\"cd\""))
                ),
                Arguments.of(
                        ".[:3]",
                        JsonValue.parse("[\"a\",\"b\",\"c\",\"d\",\"e\"]"),
                        Arrays.asList(JsonValue.parse("[\"a\", \"b\", \"c\"]"))
                ),
                Arguments.of(
                        ".[-2:]",
                        JsonValue.parse("[\"a\",\"b\",\"c\",\"d\",\"e\"]"),
                        Arrays.asList(JsonValue.parse("[\"d\", \"e\"]"))
                ),
                Arguments.of(
                        ".[]",
                        JsonValue.parse("[{\"name\":\"JSON\", \"good\":true}, {\"name\":\"XML\", \"good\":false}]"),
                        Arrays.asList(
                                JsonValue.parse("{\"name\":\"JSON\", \"good\":true}"),
                                JsonValue.parse("{\"name\":\"XML\", \"good\":false}")
                        )
                ),
                Arguments.of(
                        ".[]",
                        JsonValue.parse("[]"),
                        Arrays.asList()
                ),
                Arguments.of(
                        ".foo[]",
                        JsonValue.parse("{\"foo\":[1,2,3]}"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("2"),
                                JsonValue.parse("3")
                        )
                ),
                Arguments.of(
                        ".[]",
                        JsonValue.parse("{\"a\": 1, \"b\": 1}"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("1")
                        )
                ),
                Arguments.of(
                        ".foo, .bar",
                        JsonValue.parse("{\"foo\": 42, \"bar\": \"something else\", \"baz\": true}"),
                        Arrays.asList(
                                JsonValue.parse("42"),
                                JsonValue.parse("\"something else\"")
                        )
                ),
                Arguments.of(
                        ".user, .projects[]",
                        JsonValue.parse("{\"user\":\"stedolan\", \"projects\": [\"jq\", \"wikiflow\"]}"),
                        Arrays.asList(
                                JsonValue.parse("\"stedolan\""),
                                JsonValue.parse("\"jq\""),
                                JsonValue.parse("\"wikiflow\"")
                        )
                ),
                Arguments.of(
                        ".[4,2]",
                        JsonValue.parse("[\"a\",\"b\",\"c\",\"d\",\"e\"]"),
                        Arrays.asList(
                                JsonValue.parse("\"e\""),
                                JsonValue.parse("\"c\"")
                        )
                ),
                Arguments.of(
                        ".[] | .name",
                        JsonValue.parse("[{\"name\":\"JSON\", \"good\":true}, {\"name\":\"XML\", \"good\":false}]"),
                        Arrays.asList(
                                JsonValue.parse("\"JSON\""),
                                JsonValue.parse("\"XML\"")
                        )
                ),
                Arguments.of(
                        "(. + 2) * 5",
                        JsonValue.parse("1"),
                        Arrays.asList(JsonValue.parse("15"))
                ),
                Arguments.of(
                        "[.user, .projects[]]",
                        JsonValue.parse("{\"user\":\"stedolan\", \"projects\": [\"jq\", \"wikiflow\"]}"),
                        Arrays.asList(JsonValue.parse("[\"stedolan\", \"jq\", \"wikiflow\"]"))
                ),
                Arguments.of(
                        "[ .[] | . * 2]",
                        JsonValue.parse("[1, 2, 3]"),
                        Arrays.asList(JsonValue.parse("[2, 4, 6]"))
                ),
                Arguments.of(
                        "{user, title: .titles[]}",
                        JsonValue.parse("{\"user\":\"stedolan\",\"titles\":[\"JQ Primer\", \"More JQ\"]}"),
                        Arrays.asList(
                                JsonValue.parse("{\"user\":\"stedolan\", \"title\": \"JQ Primer\"}"),
                                JsonValue.parse("{\"user\":\"stedolan\", \"title\": \"More JQ\"}")
                        )
                ),
                Arguments.of(
                        "{(.user): .titles}",
                        JsonValue.parse("{\"user\":\"stedolan\",\"titles\":[\"JQ Primer\", \"More JQ\"]}"),
                        Arrays.asList(JsonValue.parse("{\"stedolan\": [\"JQ Primer\", \"More JQ\"]}"))
                ),
                Arguments.of(
                        ".. | .a?",
                        JsonValue.parse("[[{\"a\":1}]]"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        ".a + 1",
                        JsonValue.parse("{\"a\": 7}"),
                        Arrays.asList(JsonValue.parse("8"))
                ),
                Arguments.of(
                        ".a + .b",
                        JsonValue.parse("{\"a\": [1,2], \"b\": [3,4]}"),
                        Arrays.asList(JsonValue.parse("[1,2,3,4]"))
                ),
                Arguments.of(
                        ".a + null",
                        JsonValue.parse("{\"a\": 1}"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        ".a + 1",
                        JsonValue.parse("{}"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        "{a: 1} + {b: 2} + {c: 3} + {a: 42}",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("{\"a\": 42, \"b\": 2, \"c\": 3}"))
                ),
                Arguments.of(
                        "4 - .a",
                        JsonValue.parse("{\"a\":3}"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        ". - [\"xml\", \"yaml\"]",
                        JsonValue.parse("[\"xml\", \"yaml\", \"json\"]"),
                        Arrays.asList(JsonValue.parse("[\"json\"]"))
                ),
                Arguments.of(
                        "10 / . * 3",
                        JsonValue.parse("5"),
                        Arrays.asList(JsonValue.parse("6"))
                ),
                Arguments.of(
                        ". / \", \"",
                        JsonValue.parse("\"a, b,c,d, e\""),
                        Arrays.asList(JsonValue.parse("[\"a\",\"b,c,d\",\"e\"]"))
                ),
                Arguments.of(
                        "{\"k\": {\"a\": 1, \"b\": 2}} * {\"k\": {\"a\": 0,\"c\": 3}}",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("{\"k\": {\"a\": 0, \"b\": 2, \"c\": 3}}"))
                ),
                Arguments.of(
                        ".[] | (1 / .)?",
                        JsonValue.parse("[1,0,-1]"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("-1")
                        )
                ),
                Arguments.of(
                        "map(abs)",
                        JsonValue.parse("[-10, -1.1, -1e-1]"),
                        Arrays.asList(JsonValue.parse("[10,1.1,1e-1]"))
                ),
                Arguments.of(
                        ".[] | length",
                        JsonValue.parse("[[1,2], \"string\", {\"a\":2}, null, -5]"),
                        Arrays.asList(
                                JsonValue.parse("2"),
                                JsonValue.parse("6"),
                                JsonValue.parse("1"),
                                JsonValue.parse("0"),
                                JsonValue.parse("5")
                        )
                ),
                Arguments.of(
                        "utf8bytelength",
                        JsonValue.parse("\"\\u03bc\""),
                        Arrays.asList(JsonValue.parse("2"))
                ),
                Arguments.of(
                        "keys",
                        JsonValue.parse("{\"abc\": 1, \"abcd\": 2, \"Foo\": 3}"),
                        Arrays.asList(JsonValue.parse("[\"Foo\", \"abc\", \"abcd\"]"))
                ),
                Arguments.of(
                        "keys",
                        JsonValue.parse("[42,3,35]"),
                        Arrays.asList(JsonValue.parse("[0,1,2]"))
                ),
                Arguments.of(
                        "map(has(\"foo\"))",
                        JsonValue.parse("[{\"foo\": 42}, {}]"),
                        Arrays.asList(JsonValue.parse("[true, false]"))
                ),
                Arguments.of(
                        "map(has(2))",
                        JsonValue.parse("[[0,1], [\"a\",\"b\",\"c\"]]"),
                        Arrays.asList(JsonValue.parse("[false, true]"))
                ),
                Arguments.of(
                        ".[] | in({\"foo\": 42})",
                        JsonValue.parse("[\"foo\", \"bar\"]"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("false")
                        )
                ),
                Arguments.of(
                        "map(in([0,1]))",
                        JsonValue.parse("[2, 0]"),
                        Arrays.asList(JsonValue.parse("[false, true]"))
                ),
                Arguments.of(
                        "map(.+1)",
                        JsonValue.parse("[1,2,3]"),
                        Arrays.asList(JsonValue.parse("[2,3,4]"))
                ),
                Arguments.of(
                        "map_values(.+1)",
                        JsonValue.parse("{\"a\": 1, \"b\": 2, \"c\": 3}"),
                        Arrays.asList(JsonValue.parse("{\"a\": 2, \"b\": 3, \"c\": 4}"))
                ),
                Arguments.of(
                        "map(., .)",
                        JsonValue.parse("[1,2]"),
                        Arrays.asList(JsonValue.parse("[1,1,2,2]"))
                ),
                Arguments.of(
                        "map_values(. // empty)",
                        JsonValue.parse("{\"a\": null, \"b\": true, \"c\": false}"),
                        Arrays.asList(JsonValue.parse("{\"b\":true}"))
                ),
                Arguments.of(
                        "pick(.a, .b.c, .x)",
                        JsonValue.parse("{\"a\": 1, \"b\": {\"c\": 2, \"d\": 3}, \"e\": 4}"),
                        Arrays.asList(JsonValue.parse("{\"a\":1,\"b\":{\"c\":2},\"x\":null}"))
                ),
                Arguments.of(
                        "pick(.[2], .[0], .[0])",
                        JsonValue.parse("[1,2,3,4]"),
                        Arrays.asList(JsonValue.parse("[1,null,3]"))
                ),
                Arguments.of(
                        "path(.a[0].b)",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[\"a\",0,\"b\"]"))
                ),
                Arguments.of(
                        "[path(..)]",
                        JsonValue.parse("{\"a\":[{\"b\":1}]}"),
                        Arrays.asList(JsonValue.parse("[[],[\"a\"],[\"a\",0],[\"a\",0,\"b\"]]"))
                ),
                Arguments.of(
                        "del(.foo)",
                        JsonValue.parse("{\"foo\": 42, \"bar\": 9001, \"baz\": 42}"),
                        Arrays.asList(JsonValue.parse("{\"bar\": 9001, \"baz\": 42}"))
                ),
                Arguments.of(
                        "del(.[1, 2])",
                        JsonValue.parse("[\"foo\", \"bar\", \"baz\"]"),
                        Arrays.asList(JsonValue.parse("[\"foo\"]"))
                ),
                Arguments.of(
                        "getpath([\"a\",\"b\"])",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("null"))
                ),
                Arguments.of(
                        "[getpath([\"a\",\"b\"], [\"a\",\"c\"])]",
                        JsonValue.parse("{\"a\":{\"b\":0, \"c\":1}}"),
                        Arrays.asList(JsonValue.parse("[0, 1]"))
                ),
                Arguments.of(
                        "setpath([\"a\",\"b\"]; 1)",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("{\"a\": {\"b\": 1}}"))
                ),
                Arguments.of(
                        "setpath([\"a\",\"b\"]; 1)",
                        JsonValue.parse("{\"a\":{\"b\":0}}"),
                        Arrays.asList(JsonValue.parse("{\"a\": {\"b\": 1}}"))
                ),
                Arguments.of(
                        "setpath([0,\"a\"]; 1)",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[{\"a\":1}]"))
                ),
                Arguments.of(
                        "delpaths([[\"a\",\"b\"]])",
                        JsonValue.parse("{\"a\":{\"b\":1},\"x\":{\"y\":2}}"),
                        Arrays.asList(JsonValue.parse("{\"a\":{},\"x\":{\"y\":2}}"))
                ),
                Arguments.of(
                        "to_entries",
                        JsonValue.parse("{\"a\": 1, \"b\": 2}"),
                        Arrays.asList(JsonValue.parse("[{\"key\":\"a\", \"value\":1}, {\"key\":\"b\", \"value\":2}]"))
                ),
                Arguments.of(
                        "from_entries",
                        JsonValue.parse("[{\"key\":\"a\", \"value\":1}, {\"key\":\"b\", \"value\":2}]"),
                        Arrays.asList(JsonValue.parse("{\"a\": 1, \"b\": 2}"))
                ),
                Arguments.of(
                        "with_entries(.key |= \"KEY_\" + .)",
                        JsonValue.parse("{\"a\": 1, \"b\": 2}"),
                        Arrays.asList(JsonValue.parse("{\"KEY_a\": 1, \"KEY_b\": 2}"))
                ),
                Arguments.of(
                        "map(select(. >= 2))",
                        JsonValue.parse("[1,5,3,0,7]"),
                        Arrays.asList(JsonValue.parse("[5,3,7]"))
                ),
                Arguments.of(
                        ".[] | select(.id == \"second\")",
                        JsonValue.parse("[{\"id\": \"first\", \"val\": 1}, {\"id\": \"second\", \"val\": 2}]"),
                        Arrays.asList(JsonValue.parse("{\"id\": \"second\", \"val\": 2}"))
                ),
                Arguments.of(
                        ".[]|numbers",
                        JsonValue.parse("[[],{},1,\"foo\",null,true,false]"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        "1, empty, 2",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("2")
                        )
                ),
                Arguments.of(
                        "[1,2,empty,3]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[1,2,3]"))
                ),
                Arguments.of(
                        "try error catch .",
                        JsonValue.parse("\"error message\""),
                        Arrays.asList(JsonValue.parse("\"error message\""))
                ),
                Arguments.of(
                        "try error(\"invalid value: \\(.)\") catch .",
                        JsonValue.parse("42"),
                        Arrays.asList(JsonValue.parse("\"invalid value: 42\""))
                ),
                Arguments.of(
                        "try error(\"\\($__loc__)\") catch .",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("\"{\\\"file\\\":\\\"<top-level>\\\",\\\"line\\\":1}\""))
                ),
                Arguments.of(
                        "[paths]",
                        JsonValue.parse("[1,[[],{\"a\":2}]]"),
                        Arrays.asList(JsonValue.parse("[[0],[1],[1,0],[1,1],[1,1,\"a\"]]"))
                ),
                Arguments.of(
                        "[paths(type == \"number\")]",
                        JsonValue.parse("[1,[[],{\"a\":2}]]"),
                        Arrays.asList(JsonValue.parse("[[0],[1,1,\"a\"]]"))
                ),
                Arguments.of(
                        "add",
                        JsonValue.parse("[\"a\",\"b\",\"c\"]"),
                        Arrays.asList(JsonValue.parse("\"abc\""))
                ),
                Arguments.of(
                        "add",
                        JsonValue.parse("[1, 2, 3]"),
                        Arrays.asList(JsonValue.parse("6"))
                ),
                Arguments.of(
                        "add",
                        JsonValue.parse("[]"),
                        Arrays.asList(JsonValue.parse("null"))
                ),
                Arguments.of(
                        "any",
                        JsonValue.parse("[true, false]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "any",
                        JsonValue.parse("[false, false]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "any",
                        JsonValue.parse("[]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "all",
                        JsonValue.parse("[true, false]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "all",
                        JsonValue.parse("[true, true]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "all",
                        JsonValue.parse("[]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "flatten",
                        JsonValue.parse("[1, [2], [[3]]]"),
                        Arrays.asList(JsonValue.parse("[1, 2, 3]"))
                ),
                Arguments.of(
                        "flatten(1)",
                        JsonValue.parse("[1, [2], [[3]]]"),
                        Arrays.asList(JsonValue.parse("[1, 2, [3]]"))
                ),
                Arguments.of(
                        "flatten",
                        JsonValue.parse("[[]]"),
                        Arrays.asList(JsonValue.parse("[]"))
                ),
                Arguments.of(
                        "flatten",
                        JsonValue.parse("[{\"foo\": \"bar\"}, [{\"foo\": \"baz\"}]]"),
                        Arrays.asList(JsonValue.parse("[{\"foo\": \"bar\"}, {\"foo\": \"baz\"}]"))
                ),
                Arguments.of(
                        "range(2; 4)",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("2"),
                                JsonValue.parse("3")
                        )
                ),
                Arguments.of(
                        "[range(2; 4)]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[2,3]"))
                ),
                Arguments.of(
                        "[range(4)]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[0,1,2,3]"))
                ),
                Arguments.of(
                        "[range(0; 10; 3)]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[0,3,6,9]"))
                ),
                Arguments.of(
                        "[range(0; 10; -1)]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[]"))
                ),
                Arguments.of(
                        "[range(0; -5; -1)]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[0,-1,-2,-3,-4]"))
                ),
                Arguments.of(
                        "floor",
                        JsonValue.parse("3.14159"),
                        Arrays.asList(JsonValue.parse("3"))
                ),
                Arguments.of(
                        "sqrt",
                        JsonValue.parse("9"),
                        Arrays.asList(JsonValue.parse("3"))
                ),
                Arguments.of(
                        ".[] | tonumber",
                        JsonValue.parse("[1, \"1\"]"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("1")
                        )
                ),
                Arguments.of(
                        ".[] | tostring",
                        JsonValue.parse("[1, \"1\", [1]]"),
                        Arrays.asList(
                                JsonValue.parse("\"1\""),
                                JsonValue.parse("\"1\""),
                                JsonValue.parse("\"[1]\"")
                        )
                ),
                Arguments.of(
                        "map(type)",
                        JsonValue.parse("[0, false, [], {}, null, \"hello\"]"),
                        Arrays.asList(JsonValue.parse(
                                "[\"number\", \"boolean\", \"array\", \"object\", \"null\", \"string\"]"))
                ),
                Arguments.of(
                        ".[] | (infinite * .) < 0",
                        JsonValue.parse("[-1, 1]"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("false")
                        )
                ),
                Arguments.of(
                        "infinite, nan | type",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("\"number\""),
                                JsonValue.parse("\"number\"")
                        )
                ),
                Arguments.of(
                        "sort",
                        JsonValue.parse("[8,3,null,6]"),
                        Arrays.asList(JsonValue.parse("[null,3,6,8]"))
                ),
                Arguments.of(
                        "sort_by(.foo)",
                        JsonValue.parse("[{\"foo\":4, \"bar\":10}, {\"foo\":3, \"bar\":10}, {\"foo\":2, \"bar\":1}]"),
                        Arrays.asList(
                                JsonValue.parse(
                                        "[{\"foo\":2, \"bar\":1}, {\"foo\":3, \"bar\":10}, {\"foo\":4, \"bar\":10}]")
                        )
                ),
                Arguments.of(
                        "sort_by(.foo, .bar)",
                        JsonValue.parse(
                                "[{\"foo\":4, \"bar\":10}, {\"foo\":3, \"bar\":20}, {\"foo\":2, \"bar\":1}, {\"foo\":3, \"bar\":10}]"),
                        Arrays.asList(
                                JsonValue.parse(
                                        "[{\"foo\":2, \"bar\":1}, {\"foo\":3, \"bar\":10}, {\"foo\":3, \"bar\":20}, {\"foo\":4, \"bar\":10}]")
                        )
                ),
                Arguments.of(
                        "group_by(.foo)",
                        JsonValue.parse("[{\"foo\":1, \"bar\":10}, {\"foo\":3, \"bar\":100}, {\"foo\":1, \"bar\":1}]"),
                        Arrays.asList(
                                JsonValue.parse(
                                        "[[{\"foo\":1, \"bar\":10}, {\"foo\":1, \"bar\":1}], [{\"foo\":3, \"bar\":100}]]")
                        )
                ),
                Arguments.of(
                        "min",
                        JsonValue.parse("[5,4,2,7]"),
                        Arrays.asList(JsonValue.parse("2"))
                ),
                Arguments.of(
                        "max_by(.foo)",
                        JsonValue.parse("[{\"foo\":1, \"bar\":14}, {\"foo\":2, \"bar\":3}]"),
                        Arrays.asList(JsonValue.parse("{\"foo\":2, \"bar\":3}"))
                ),
                Arguments.of(
                        "unique",
                        JsonValue.parse("[1,2,5,3,5,3,1,3]"),
                        Arrays.asList(JsonValue.parse("[1,2,3,5]"))
                ),
                Arguments.of(
                        "unique_by(.foo)",
                        JsonValue.parse("[{\"foo\": 1, \"bar\": 2}, {\"foo\": 1, \"bar\": 3}, {\"foo\": 4, \"bar\": 5}]"),
                        Arrays.asList(JsonValue.parse("[{\"foo\": 1, \"bar\": 2}, {\"foo\": 4, \"bar\": 5}]"))
                ),
                Arguments.of(
                        "unique_by(length)",
                        JsonValue.parse("[\"chunky\", \"bacon\", \"kitten\", \"cicada\", \"asparagus\"]"),
                        Arrays.asList(JsonValue.parse("[\"bacon\", \"chunky\", \"asparagus\"]"))
                ),
                Arguments.of(
                        "reverse",
                        JsonValue.parse("[1,2,3,4]"),
                        Arrays.asList(JsonValue.parse("[4,3,2,1]"))
                ),
                Arguments.of(
                        "contains(\"bar\")",
                        JsonValue.parse("\"foobar\""),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "contains([\"baz\", \"bar\"])",
                        JsonValue.parse("[\"foobar\", \"foobaz\", \"blarp\"]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "contains([\"bazzzzz\", \"bar\"])",
                        JsonValue.parse("[\"foobar\", \"foobaz\", \"blarp\"]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "contains({foo: 12, bar: [{barp: 12}]})",
                        JsonValue.parse("{\"foo\": 12, \"bar\":[1,2,{\"barp\":12, \"blip\":13}]}"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "contains({foo: 12, bar: [{barp: 15}]})",
                        JsonValue.parse("{\"foo\": 12, \"bar\":[1,2,{\"barp\":12, \"blip\":13}]}"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "indices(\", \")",
                        JsonValue.parse("\"a,b, cd, efg, hijk\""),
                        Arrays.asList(JsonValue.parse("[3,7,12]"))
                ),
                Arguments.of(
                        "indices(1)",
                        JsonValue.parse("[0,1,2,1,3,1,4]"),
                        Arrays.asList(JsonValue.parse("[1,3,5]"))
                ),
                Arguments.of(
                        "indices([1,2])",
                        JsonValue.parse("[0,1,2,3,1,4,2,5,1,2,6,7]"),
                        Arrays.asList(JsonValue.parse("[1,8]"))
                ),
                Arguments.of(
                        "index(\", \")",
                        JsonValue.parse("\"a,b, cd, efg, hijk\""),
                        Arrays.asList(JsonValue.parse("3"))
                ),
                Arguments.of(
                        "index(1)",
                        JsonValue.parse("[0,1,2,1,3,1,4]"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        "index([1,2])",
                        JsonValue.parse("[0,1,2,3,1,4,2,5,1,2,6,7]"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        "rindex(\", \")",
                        JsonValue.parse("\"a,b, cd, efg, hijk\""),
                        Arrays.asList(JsonValue.parse("12"))
                ),
                Arguments.of(
                        "rindex(1)",
                        JsonValue.parse("[0,1,2,1,3,1,4]"),
                        Arrays.asList(JsonValue.parse("5"))
                ),
                Arguments.of(
                        "rindex([1,2])",
                        JsonValue.parse("[0,1,2,3,1,4,2,5,1,2,6,7]"),
                        Arrays.asList(JsonValue.parse("8"))
                ),
                Arguments.of(
                        "inside(\"foobar\")",
                        JsonValue.parse("\"bar\""),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "inside([\"foobar\", \"foobaz\", \"blarp\"])",
                        JsonValue.parse("[\"baz\", \"bar\"]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "inside([\"foobar\", \"foobaz\", \"blarp\"])",
                        JsonValue.parse("[\"bazzzzz\", \"bar\"]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "inside({\"foo\": 12, \"bar\":[1,2,{\"barp\":12, \"blip\":13}]})",
                        JsonValue.parse("{\"foo\": 12, \"bar\": [{\"barp\": 12}]}"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "inside({\"foo\": 12, \"bar\":[1,2,{\"barp\":12, \"blip\":13}]})",
                        JsonValue.parse("{\"foo\": 12, \"bar\": [{\"barp\": 15}]}"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "[.[]|startswith(\"foo\")]",
                        JsonValue.parse("[\"fo\", \"foo\", \"barfoo\", \"foobar\", \"barfoob\"]"),
                        Arrays.asList(JsonValue.parse("[false, true, false, true, false]"))
                ),
                Arguments.of(
                        "[.[]|endswith(\"foo\")]",
                        JsonValue.parse("[\"foobar\", \"barfoo\"]"),
                        Arrays.asList(JsonValue.parse("[false, true]"))
                ),
                Arguments.of(
                        "combinations",
                        JsonValue.parse("[[1,2], [3, 4]]"),
                        Arrays.asList(
                                JsonValue.parse("[1, 3]"),
                                JsonValue.parse("[1, 4]"),
                                JsonValue.parse("[2, 3]"),
                                JsonValue.parse("[2, 4]")
                        )
                ),
                Arguments.of(
                        "combinations(2)",
                        JsonValue.parse("[0, 1]"),
                        Arrays.asList(
                                JsonValue.parse("[0, 0]"),
                                JsonValue.parse("[0, 1]"),
                                JsonValue.parse("[1, 0]"),
                                JsonValue.parse("[1, 1]")
                        )
                ),
                Arguments.of(
                        "[.[]|ltrimstr(\"foo\")]",
                        JsonValue.parse("[\"fo\", \"foo\", \"barfoo\", \"foobar\", \"afoo\"]"),
                        Arrays.asList(JsonValue.parse("[\"fo\",\"\",\"barfoo\",\"bar\",\"afoo\"]"))
                ),
                Arguments.of(
                        "[.[]|rtrimstr(\"foo\")]",
                        JsonValue.parse("[\"fo\", \"foo\", \"barfoo\", \"foobar\", \"foob\"]"),
                        Arrays.asList(JsonValue.parse("[\"fo\",\"\",\"bar\",\"foobar\",\"foob\"]"))
                ),
                Arguments.of(
                        "trim, ltrim, rtrim",
                        JsonValue.parse("\" abc \""),
                        Arrays.asList(
                                JsonValue.parse("\"abc\""),
                                JsonValue.parse("\"abc \""),
                                JsonValue.parse("\" abc\"")
                        )
                ),
                Arguments.of(
                        "explode",
                        JsonValue.parse("\"foobar\""),
                        Arrays.asList(JsonValue.parse("[102,111,111,98,97,114]"))
                ),
                Arguments.of(
                        "implode",
                        JsonValue.parse("[65, 66, 67]"),
                        Arrays.asList(JsonValue.parse("\"ABC\""))
                ),
                Arguments.of(
                        "split(\", \")",
                        JsonValue.parse("\"a, b,c,d, e, \""),
                        Arrays.asList(JsonValue.parse("[\"a\",\"b,c,d\",\"e\",\"\"]"))
                ),
                Arguments.of(
                        "join(\", \")",
                        JsonValue.parse("[\"a\",\"b,c,d\",\"e\"]"),
                        Arrays.asList(JsonValue.parse("\"a, b,c,d, e\""))
                ),
                Arguments.of(
                        "join(\" \")",
                        JsonValue.parse("[\"a\",1,2.3,true,null,false]"),
                        Arrays.asList(JsonValue.parse("\"a 1 2.3 true false\""))
                ),
                Arguments.of(
                        "ascii_upcase",
                        JsonValue.parse("\"useful but not for �\""),
                        Arrays.asList(JsonValue.parse("\"USEFUL BUT NOT FOR �\""))
                ),
                Arguments.of(
                        "[while(.<100; .*2)]",
                        JsonValue.parse("1"),
                        Arrays.asList(JsonValue.parse("[1,2,4,8,16,32,64]"))
                ),
                Arguments.of(
                        "[repeat(.*2, error)?]",
                        JsonValue.parse("1"),
                        Arrays.asList(JsonValue.parse("[2]"))
                ),
                Arguments.of(
                        "[.,1]|until(.[0] < 1; [.[0] - 1, .[1] * .[0]])|.[1]",
                        JsonValue.parse("4"),
                        Arrays.asList(JsonValue.parse("24"))
                ),
                Arguments.of(
                        "recurse(.foo[])",
                        JsonValue.parse("{\"foo\":[{\"foo\": []}, {\"foo\":[{\"foo\":[]}]}]}"),
                        Arrays.asList(
                                JsonValue.parse("{\"foo\":[{\"foo\":[]},{\"foo\":[{\"foo\":[]}]}]}"),
                                JsonValue.parse("{\"foo\":[]}"),
                                JsonValue.parse("{\"foo\":[{\"foo\":[]}]}"),
                                JsonValue.parse("{\"foo\":[]}")
                        )
                ),
                Arguments.of(
                        "recurse",
                        JsonValue.parse("{\"a\":0,\"b\":[1]}"),
                        Arrays.asList(
                                JsonValue.parse("{\"a\":0,\"b\":[1]}"),
                                JsonValue.parse("0"),
                                JsonValue.parse("[1]"),
                                JsonValue.parse("1")
                        )
                ),
                Arguments.of(
                        "recurse(. * .; . < 20)",
                        JsonValue.parse("2"),
                        Arrays.asList(
                                JsonValue.parse("2"),
                                JsonValue.parse("4"),
                                JsonValue.parse("16")
                        )
                ),
                Arguments.of(
                        "walk(if type == \"array\" then sort else . end)",
                        JsonValue.parse("[[4, 1, 7], [8, 5, 2], [3, 6, 9]]"),
                        Arrays.asList(JsonValue.parse("[[1,4,7],[2,5,8],[3,6,9]]"))
                ),
                Arguments.of(
                        "walk( if type == \"object\" then with_entries( .key |= sub( \"^_+\"; \"\") ) else . end )",
                        JsonValue.parse("[ { \"_a\": { \"__b\": 2 } } ]"),
                        Arrays.asList(JsonValue.parse("[{\"a\":{\"b\":2}}]"))
                ),
                Arguments.of(
                        "$ENV.PAGER",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("\"less\""))
                ),
                Arguments.of(
                        "env.PAGER",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("\"less\""))
                ),
                Arguments.of(
                        "transpose",
                        JsonValue.parse("[[1], [2,3]]"),
                        Arrays.asList(JsonValue.parse("[[1,2],[null,3]]"))
                ),
                Arguments.of(
                        "bsearch(0)",
                        JsonValue.parse("[0,1]"),
                        Arrays.asList(JsonValue.parse("0"))
                ),
                Arguments.of(
                        "bsearch(0)",
                        JsonValue.parse("[1,2,3]"),
                        Arrays.asList(JsonValue.parse("-1"))
                ),
                Arguments.of(
                        "bsearch(4) as $ix | if $ix < 0 then .[-(1+$ix)] = 4 else . end",
                        JsonValue.parse("[1,2,3]"),
                        Arrays.asList(JsonValue.parse("[1,2,3,4]"))
                ),
                Arguments.of(
                        "\"The input was \\(.), which is one less than \\(.+1)\"",
                        JsonValue.parse("42"),
                        Arrays.asList(JsonValue.parse("\"The input was 42, which is one less than 43\""))
                ),
                Arguments.of(
                        "[.[]|tostring]",
                        JsonValue.parse("[1, \"foo\", [\"foo\"]]"),
                        Arrays.asList(JsonValue.parse("[\"1\",\"foo\",\"[\\\"foo\\\"]\"]"))
                ),
                Arguments.of(
                        "[.[]|tojson]",
                        JsonValue.parse("[1, \"foo\", [\"foo\"]]"),
                        Arrays.asList(JsonValue.parse("[\"1\",\"\\\"foo\\\"\",\"[\\\"foo\\\"]\"]"))
                ),
                Arguments.of(
                        "[.[]|tojson|fromjson]",
                        JsonValue.parse("[1, \"foo\", [\"foo\"]]"),
                        Arrays.asList(JsonValue.parse("[1,\"foo\",[\"foo\"]]"))
                ),
                Arguments.of(
                        "@html",
                        JsonValue.parse("\"This works if x < y\""),
                        Arrays.asList(JsonValue.parse("\"This works if x &lt; y\""))
                ),
                Arguments.of(
                        "@sh \"echo \\(.)\"",
                        JsonValue.parse("\"O'Hara's Ale\""),
                        Arrays.asList(JsonValue.parse("\"echo 'O'\\\\''Hara'\\\\''s Ale'\""))
                ),
                Arguments.of(
                        "@base64",
                        JsonValue.parse("\"This is a message\""),
                        Arrays.asList(JsonValue.parse("\"VGhpcyBpcyBhIG1lc3NhZ2U=\""))
                ),
                Arguments.of(
                        "@base64d",
                        JsonValue.parse("\"VGhpcyBpcyBhIG1lc3NhZ2U=\""),
                        Arrays.asList(JsonValue.parse("\"This is a message\""))
                ),
                Arguments.of(
                        "fromdate",
                        JsonValue.parse("\"2015-03-05T23:51:47Z\""),
                        Arrays.asList(JsonValue.parse("1425599507"))
                ),
                Arguments.of(
                        "strptime(\"%Y-%m-%dT%H:%M:%SZ\")",
                        JsonValue.parse("\"2015-03-05T23:51:47Z\""),
                        Arrays.asList(JsonValue.parse("[2015,2,5,23,51,47,4,63]"))
                ),
                Arguments.of(
                        "strptime(\"%Y-%m-%dT%H:%M:%SZ\")|mktime",
                        JsonValue.parse("\"2015-03-05T23:51:47Z\""),
                        Arrays.asList(JsonValue.parse("1425599507"))
                ),
                Arguments.of(
                        ". == false",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        ". == {\"b\": {\"d\": (4 + 1e-20), \"c\": 3}, \"a\":1}",
                        JsonValue.parse("{\"a\":1, \"b\": {\"c\": 3, \"d\": 4}}"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        ".[] == 1",
                        JsonValue.parse("[1, 1.0, \"1\", \"banana\"]"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("true"),
                                JsonValue.parse("false"),
                                JsonValue.parse("false")
                        )
                ),
                Arguments.of(
                        "if . == 0 then \"zero\" elif . == 1 then \"one\" else \"many\" end",
                        JsonValue.parse("2"),
                        Arrays.asList(JsonValue.parse("\"many\""))
                ),
                Arguments.of(
                        ". < 5",
                        JsonValue.parse("2"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "42 and \"a string\"",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "(true, false) or false",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("false")
                        )
                ),
                Arguments.of(
                        "(true, true) and (true, false)",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("false"),
                                JsonValue.parse("true"),
                                JsonValue.parse("false")
                        )
                ),
                Arguments.of(
                        "[true, false | not]",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[false, true]"))
                ),
                Arguments.of(
                        "empty // 42",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        ".foo // 42",
                        JsonValue.parse("{\"foo\": 19}"),
                        Arrays.asList(JsonValue.parse("19"))
                ),
                Arguments.of(
                        ".foo // 42",
                        JsonValue.parse("{}"),
                        Arrays.asList(JsonValue.parse("42"))
                ),
                Arguments.of(
                        "(false, null, 1) // 42",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("1"))
                ),
                Arguments.of(
                        "(false, null, 1) | . // 42",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("42"),
                                JsonValue.parse("42"),
                                JsonValue.parse("1")
                        )
                ),
                Arguments.of(
                        "try .a catch \". is not an object\"",
                        JsonValue.parse("true"),
                        Arrays.asList(JsonValue.parse("\". is not an object\""))
                ),
                Arguments.of(
                        "[.[]|try .a]",
                        JsonValue.parse("[{}, true, {\"a\":1}]"),
                        Arrays.asList(JsonValue.parse("[null, 1]"))
                ),
                Arguments.of(
                        "try error(\"some exception\") catch .",
                        JsonValue.parse("true"),
                        Arrays.asList(JsonValue.parse("\"some exception\""))
                ),
                Arguments.of(
                        "[.[] | .a?]",
                        JsonValue.parse("[{}, true, {\"a\":1}]"),
                        Arrays.asList(JsonValue.parse("[null, 1]"))
                ),
                Arguments.of(
                        "[.[] | tonumber?]",
                        JsonValue.parse("[\"1\", \"invalid\", \"3\", 4]"),
                        Arrays.asList(JsonValue.parse("[1, 3, 4]"))
                ),
                Arguments.of(
                        "test(\"foo\")",
                        JsonValue.parse("\"foo\""),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        ".[] | test(\"a b c # spaces are ignored\"; \"ix\")",
                        JsonValue.parse("[\"xabcd\", \"ABC\"]"),
                        Arrays.asList(
                                JsonValue.parse("true"),
                                JsonValue.parse("true")
                        )
                ),
                Arguments.of(
                        "match(\"(abc)+\"; \"g\")",
                        JsonValue.parse("\"abc abc\""),
                        Arrays.asList(
                                JsonValue.parse(
                                        "{\"offset\": 0, \"length\": 3, \"string\": \"abc\", \"captures\": [{\"offset\": 0, \"length\": 3, \"string\": \"abc\", \"name\": null}]}"),
                                JsonValue.parse(
                                        "{\"offset\": 4, \"length\": 3, \"string\": \"abc\", \"captures\": [{\"offset\": 4, \"length\": 3, \"string\": \"abc\", \"name\": null}]}")
                        )
                ),
                Arguments.of(
                        "match(\"foo\")",
                        JsonValue.parse("\"foo bar foo\""),
                        Arrays.asList(JsonValue.parse(
                                "{\"offset\": 0, \"length\": 3, \"string\": \"foo\", \"captures\": []}"))
                ),
                Arguments.of(
                        "match([\"foo\", \"ig\"])",
                        JsonValue.parse("\"foo bar FOO\""),
                        Arrays.asList(
                                JsonValue.parse("{\"offset\": 0, \"length\": 3, \"string\": \"foo\", \"captures\": []}"),
                                JsonValue.parse("{\"offset\": 8, \"length\": 3, \"string\": \"FOO\", \"captures\": []}")
                        )
                ),
                Arguments.of(
                        "match(\"foo (?<bar123>bar)? foo\"; \"ig\")",
                        JsonValue.parse("\"foo bar foo foo  foo\""),
                        Arrays.asList(
                                JsonValue.parse(
                                        "{\"offset\": 0, \"length\": 11, \"string\": \"foo bar foo\", \"captures\": [{\"offset\": 4, \"length\": 3, \"string\": \"bar\", \"name\": \"bar123\"}]}"),
                                JsonValue.parse(
                                        "{\"offset\": 12, \"length\": 8, \"string\": \"foo  foo\", \"captures\": [{\"offset\": -1, \"length\": 0, \"string\": null, \"name\": \"bar123\"}]}")
                        )
                ),
                Arguments.of(
                        "[ match(\".\"; \"g\")] | length",
                        JsonValue.parse("\"abc\""),
                        Arrays.asList(JsonValue.parse("3"))
                ),
                Arguments.of(
                        "capture(\"(?<a>[a-z]+)-(?<n>[0-9]+)\")",
                        JsonValue.parse("\"xyzzy-14\""),
                        Arrays.asList(JsonValue.parse("{ \"a\": \"xyzzy\", \"n\": \"14\" }"))
                ),
                Arguments.of(
                        "scan(\"c\")",
                        JsonValue.parse("\"abcdefabc\""),
                        Arrays.asList(
                                JsonValue.parse("\"c\""),
                                JsonValue.parse("\"c\"")
                        )
                ),
                Arguments.of(
                        "split(\", *\"; null)",
                        JsonValue.parse("\"ab,cd, ef\""),
                        Arrays.asList(JsonValue.parse("[\"ab\",\"cd\",\"ef\"]"))
                ),
                Arguments.of(
                        "splits(\", *\")",
                        JsonValue.parse("\"ab,cd, ef, gh\""),
                        Arrays.asList(
                                JsonValue.parse("\"ab\""),
                                JsonValue.parse("\"cd\""),
                                JsonValue.parse("\"ef\""),
                                JsonValue.parse("\"gh\"")
                        )
                ),
                Arguments.of(
                        "sub(\"[^a-z]*(?<x>[a-z]+)\"; \"Z\\(.x)\"; \"g\")",
                        JsonValue.parse("\"123abc456def\""),
                        Arrays.asList(JsonValue.parse("\"ZabcZdef\""))
                ),
                Arguments.of(
                        "[sub(\"(?<a>.)\"; \"\\(.a|ascii_upcase)\", \"\\(.a|ascii_downcase)\")]",
                        JsonValue.parse("\"aB\""),
                        Arrays.asList(JsonValue.parse("[\"AB\",\"aB\"]"))
                ),
                Arguments.of(
                        "gsub(\"(?<x>.)[^a]*\"; \"+\\(.x)-\")",
                        JsonValue.parse("\"Abcabc\""),
                        Arrays.asList(JsonValue.parse("\"+A-+a-\""))
                ),
                Arguments.of(
                        "[gsub(\"p\"; \"a\", \"b\")]",
                        JsonValue.parse("\"p\""),
                        Arrays.asList(JsonValue.parse("[\"a\",\"b\"]"))
                ),
                Arguments.of(
                        ".bar as $x | .foo | . + $x",
                        JsonValue.parse("{\"foo\":10, \"bar\":200}"),
                        Arrays.asList(JsonValue.parse("210"))
                ),
                Arguments.of(
                        ". as $i|[(.*2|. as $i| $i), $i]",
                        JsonValue.parse("5"),
                        Arrays.asList(JsonValue.parse("[10,5]"))
                ),
                Arguments.of(
                        ". as [$a, $b, {c: $c}] | $a + $b + $c",
                        JsonValue.parse("[2, 3, {\"c\": 4, \"d\": 5}]"),
                        Arrays.asList(JsonValue.parse("9"))
                ),
                Arguments.of(
                        ".[] as [$a, $b] | {a: $a, b: $b}",
                        JsonValue.parse("[[0], [0, 1], [2, 1, 0]]"),
                        Arrays.asList(
                                JsonValue.parse("{\"a\":0,\"b\":null}"),
                                JsonValue.parse("{\"a\":0,\"b\":1}"),
                                JsonValue.parse("{\"a\":2,\"b\":1}")
                        )
                ),
                Arguments.of(
                        ".[] as {$a, $b, c: {$d, $e}} ?// {$a, $b, c: [{$d, $e}]} | {$a, $b, $d, $e}",
                        JsonValue.parse(
                                "[{\"a\": 1, \"b\": 2, \"c\": {\"d\": 3, \"e\": 4}}, {\"a\": 1, \"b\": 2, \"c\": [{\"d\": 3, \"e\": 4}]}]"),
                        Arrays.asList(
                                JsonValue.parse("{\"a\":1,\"b\":2,\"d\":3,\"e\":4}"),
                                JsonValue.parse("{\"a\":1,\"b\":2,\"d\":3,\"e\":4}")
                        )
                ),
                Arguments.of(
                        ".[] as {$a, $b, c: {$d}} ?// {$a, $b, c: [{$e}]} | {$a, $b, $d, $e}",
                        JsonValue.parse(
                                "[{\"a\": 1, \"b\": 2, \"c\": {\"d\": 3, \"e\": 4}}, {\"a\": 1, \"b\": 2, \"c\": [{\"d\": 3, \"e\": 4}]}]"),
                        Arrays.asList(
                                JsonValue.parse("{\"a\":1,\"b\":2,\"d\":3,\"e\":null}"),
                                JsonValue.parse("{\"a\":1,\"b\":2,\"d\":null,\"e\":4}")
                        )
                ),
                Arguments.of(
                        ".[] as [$a] ?// [$b] | if $a != null then error(\"err: \\($a)\") else {$a,$b} end",
                        JsonValue.parse("[[3]]"),
                        Arrays.asList(JsonValue.parse("{\"a\":null,\"b\":3}"))
                ),
                Arguments.of(
                        "def addvalue(f): . + [f]; map(addvalue(.[0]))",
                        JsonValue.parse("[[1,2],[10,20]]"),
                        Arrays.asList(JsonValue.parse("[[1,2,1], [10,20,10]]"))
                ),
                Arguments.of(
                        "def addvalue(f): f as $x | map(. + $x); addvalue(.[0])",
                        JsonValue.parse("[[1,2],[10,20]]"),
                        Arrays.asList(JsonValue.parse("[[1,2,1,2], [10,20,1,2]]"))
                ),
                Arguments.of(
                        "isempty(empty)",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "isempty(.[])",
                        JsonValue.parse("[]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "isempty(.[])",
                        JsonValue.parse("[1,2,3]"),
                        Arrays.asList(JsonValue.parse("false"))
                ),
                Arguments.of(
                        "[limit(3;.[])]",
                        JsonValue.parse("[0,1,2,3,4,5,6,7,8,9]"),
                        Arrays.asList(JsonValue.parse("[0,1,2]"))
                ),
                Arguments.of(
                        "[first(range(.)), last(range(.)), nth(./2; range(.))]",
                        JsonValue.parse("10"),
                        Arrays.asList(JsonValue.parse("[0,9,5]"))
                ),
                Arguments.of(
                        "[range(.)]|[first, last, nth(5)]",
                        JsonValue.parse("10"),
                        Arrays.asList(JsonValue.parse("[0,9,5]"))
                ),
                Arguments.of(
                        "reduce .[] as $item (0; . + $item)",
                        JsonValue.parse("[1,2,3,4,5]"),
                        Arrays.asList(JsonValue.parse("15"))
                ),
                Arguments.of(
                        "reduce .[] as [$i,$j] (0; . + $i * $j)",
                        JsonValue.parse("[[1,2],[3,4],[5,6]]"),
                        Arrays.asList(JsonValue.parse("44"))
                ),
                Arguments.of(
                        "reduce .[] as {$x,$y} (null; .x += $x | .y += [$y])",
                        JsonValue.parse("[{\"x\":\"a\",\"y\":1},{\"x\":\"b\",\"y\":2},{\"x\":\"c\",\"y\":3}]"),
                        Arrays.asList(JsonValue.parse("{\"x\":\"abc\",\"y\":[1,2,3]}"))
                ),
                Arguments.of(
                        "foreach .[] as $item (0; . + $item)",
                        JsonValue.parse("[1,2,3,4,5]"),
                        Arrays.asList(
                                JsonValue.parse("1"),
                                JsonValue.parse("3"),
                                JsonValue.parse("6"),
                                JsonValue.parse("10"),
                                JsonValue.parse("15")
                        )
                ),
                Arguments.of(
                        "foreach .[] as $item (0; . + $item; [$item, . * 2])",
                        JsonValue.parse("[1,2,3,4,5]"),
                        Arrays.asList(
                                JsonValue.parse("[1,2]"),
                                JsonValue.parse("[2,6]"),
                                JsonValue.parse("[3,12]"),
                                JsonValue.parse("[4,20]"),
                                JsonValue.parse("[5,30]")
                        )
                ),
                Arguments.of(
                        "foreach .[] as $item (0; . + 1; {index: ., $item})",
                        JsonValue.parse("[\"foo\", \"bar\", \"baz\"]"),
                        Arrays.asList(
                                JsonValue.parse("{\"index\":1,\"item\":\"foo\"}"),
                                JsonValue.parse("{\"index\":2,\"item\":\"bar\"}"),
                                JsonValue.parse("{\"index\":3,\"item\":\"baz\"}")
                        )
                ),
                Arguments.of(
                        "def range(init; upto; by): def _range: if (by > 0 and . < upto) or (by < 0 and . > upto) then ., ((.+by)|_range) else . end; if by == 0 then init else init|_range end | select((by > 0 and . < upto) or (by < 0 and . > upto)); range(0; 10; 3)",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("0"),
                                JsonValue.parse("3"),
                                JsonValue.parse("6"),
                                JsonValue.parse("9")
                        )
                ),
                Arguments.of(
                        "def while(cond; update): def _while: if cond then ., (update | _while) else empty end; _while; [while(.<100; .*2)]",
                        JsonValue.parse("1"),
                        Arrays.asList(JsonValue.parse("[1,2,4,8,16,32,64]"))
                ),
                Arguments.of(
                        "truncate_stream([[0],1],[[1,0],2],[[1,0]],[[1]])",
                        JsonValue.parse("1"),
                        Arrays.asList(
                                JsonValue.parse("[[0],2]"),
                                JsonValue.parse("[[0]]")
                        )
                ),
                Arguments.of(
                        "fromstream(1|truncate_stream([[0],1],[[1,0],2],[[1,0]],[[1]]))",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("[2]"))
                ),
                Arguments.of(
                        ". as $dot|fromstream($dot|tostream)|.==$dot",
                        JsonValue.parse("[0,[1,{\"a\":1},{\"b\":2}]]"),
                        Arrays.asList(JsonValue.parse("true"))
                ),
                Arguments.of(
                        "(..|select(type==\"boolean\")) |= if . then 1 else 0 end",
                        JsonValue.parse("[true,false,[5,true,[true,[false]],false]]"),
                        Arrays.asList(JsonValue.parse("[1,0,[5,1,[1,[0]],0]]"))
                ),
                Arguments.of(
                        ".foo += 1",
                        JsonValue.parse("{\"foo\": 42}"),
                        Arrays.asList(JsonValue.parse("{\"foo\": 43}"))
                ),
                Arguments.of(
                        ".a = .b",
                        JsonValue.parse("{\"a\": {\"b\": 10}, \"b\": 20}"),
                        Arrays.asList(JsonValue.parse("{\"a\":20,\"b\":20}"))
                ),
                Arguments.of(
                        ".a |= .b",
                        JsonValue.parse("{\"a\": {\"b\": 10}, \"b\": 20}"),
                        Arrays.asList(JsonValue.parse("{\"a\":10,\"b\":20}"))
                ),
                Arguments.of(
                        "(.a, .b) = range(3)",
                        JsonValue.parse("null"),
                        Arrays.asList(
                                JsonValue.parse("{\"a\":0,\"b\":0}"),
                                JsonValue.parse("{\"a\":1,\"b\":1}"),
                                JsonValue.parse("{\"a\":2,\"b\":2}")
                        )
                ),
                Arguments.of(
                        "(.a, .b) |= range(3)",
                        JsonValue.parse("null"),
                        Arrays.asList(JsonValue.parse("{\"a\":0,\"b\":0}"))
                )
        );
    }
}
